COMMENT #   
                                                                            
                               Lasting Child                              
                         Intel(R) Chipset version                         
                                                                          
                Copyright (c) 2001-2002, Alessandro Ghignola              

     #

; 
;  L.C. System Volume Map  - mass storage addresses are in bytes           
; 
;  Bootstrap code and Volume Parameters Block      +00000000-000001FF # 0.5K
;  NanoKernel                                      +00000200-00007DFF  31.0K
;  VTOC                                            +00007E00-????????      *
;  Available for contiguous data slots             +????????-????????     **
; 
;   All volumes contain the system kernel and the system shell:
;    there is no difference between data disks and bootable disks.
; +  Sign is here to remember there may be further most-significant digits.
; *  The size of the VTOC is calculated as 1/512th of the whole volume size.
;    The VTOC holds entries of 128 bits, so the number of file slots is given
;    by VTOC size / 16. For short...
;           VTOC size = disk size SHRD 9
;           VTOC slots = disk size SHRD 13
;    (SHRD = SHift Right in double-precision, intended for 64-bit values).
; ** The available volume space is given by the following formula:
;    free space = volume size - VTOC size - 31.5 Kb
; #  The bootstrap code's size may not vary, accordingly to the capacity of
;    a standard sector. That's because the bootstrap takes exactly 1 sector.
;    Today, many BIOSes, and related software, are hard-wired to 512 bytes.
;    The Lasting Child nanokernel, for itself, is NOT hard-wired to 512 bytes
;    per sector. Supported sectors sizes are 512, 1024, 2048, 4096, 8192,
;    16384 and 32768 (maximum, 32 Kb per sector).
; 

; 
;  L.C. Volume Parameters Block (VPB) - mass storage addresses are in bytes 
; 
;  BID (BIOS ID) - can be 0,1,2,3, or 80h ... 87h   00000002-00000005     4b
;  NTK (Number of TracKs) - always 1 on LBA discs   00000006-00000009     4b
;  NHD (Number of HeaDs) - always 1 on LBA discs    0000000A-0000000D     4b
;  SPT (Sectors per Track) - total sectors on LBA   0000000E-00000011     4b
;  BPS (2's logarythm of Bytes per Sector)          00000012-00000012     1b
;  STL (Sectors To Load, initialized on bootstrap)  00000013-00000013     1b
;  vtoc (VTOC base addr, giving system code size)   00000014-00000017     4b
;  reserved for possible future use, set to zero    00000018-0000001B     4b
;  cksum (nanokernel code checksum)                 0000001C-0000001F     4b
;  sgntr (signature 'VPBX' denotes presence of VPB) 00000020-00000023     4b
;  reserved for possible future use, set to zero    00000024-00000027     4b
;  vlabel (volume label string, 64 characters)      00000028-00000067    64b
; 
; BID: (32 bit)
;   It's directly the BIOS ID of current volume.
;   It is used by the bootstrap loader to determine from which drive to read
;   the nanokernel. It is also used by the SystemReset function upon reloading
;   the boot sector (if SystemReset is entered with a parameter of EAX = 0).
;   The BID is a 32-bit field, even if BIOSes actually see the first 8 bits,
;   but it's intended for possible future remapping of media IDs...
; NTK: (32 bit)
;   Number of tracks, or cylinders, the disk is divided into.
;   Note they're called cylinders only if there is more than 1 head.
; NHD: (32 bit)
;   Number of heads.
;   This field, multiplied by SPT, gives the number of sectors per cylinder.
; SPT: (32 bit)
;   Number of sectors per *physical* track.
;   Physical tracks can form cylinders, if NHD is greater than 1.
;   For instance, a common 1.44Mb-floppy has 18 sectors per track and 2 heads,
;   so it's got 18 * 2 = 36 sectors per cylinder. The number of sectors per
;   cylinder is not stored in the VBP, exactly because it's to be calculated.
; BPS: (8 bit)
;   This is actually the log of 2 of the number of bytes in a single sector.
;   Many volumes are hard-wired to a "standard" value of BPS, which is 9,
;   and which means 512 bytes per sector, because 2 ^ 9 gives 512.
;   The bootstrap loader is also fitting 512 bytes, but this doesn't mean
;   it can't boot from drives having more than 512 bytes per sector.
;   Any value for BPS which is a power of 2 and upto 2 ^ 15 will be supported,
;   so in the end the BPS field can hold values between 9 and 15, the last
;   one meaning there are 32Kb per sector.
;   Values above 15 (more than 32Kb per sector) might result in DMA transfer
;   failure, since the nanokernel is aligned on a 32Kb boundary (at physical
;   address 8000h, leaving for another 32Kb before hitting the 64Kb boundary
;   which is the limit for DMA transfers). And in any cases, values outside
;   the range 9..15 are strongly not suggested for use by Lasting Child.
; STL: (8 bit)
;   This field is NOT INITIALIZED. It is treaten as a scratchpad by the
;   bootstrap loader. It is totally ignored by other applications or system
;   modules, and it doesn't need to be initialized (it can hold any values).
;   Technically, the STL field will hold the number of sectors forming the
;   nanokernel, during the bootstrap, and will be calculated from the value
;   of the "vtoc" field.
; vtoc: (32 bit)
;   This field is very important. Not only it informs the nanokernel's data
;   management system about where the VTOC begins: also, "vtoc" determines
;   the number of sectors that the bootstrap code will load into memory from
;   the absolute address 8000h. This field may be changed, but it has little
;   probability to change, because its actual value (32256) is the highest
;   possible value for maximum compatibility. This value gives upto 63 sectors
;   to be loaded by the bootstrap code. Since the bootstrap code uses legacy
;   ISA drive services to read the nanokernel, it can only "see" a maximum of
;   63 sectors per track, without crossing possible boundaries given by any
;   present and future devices allowing for more than 63 sectors per track,
;   or more than 256 heads, which are the final limits for ISA services.
;   Of course the system may be extended beyond the nanokernel, but I've
;   carefully targetted my efforts so that the nanokernel fits inside 63
;   sectors at 512 bytes per sector, and I'm trying hard to keep it there.
; cksum: (32 bit)
;   This is the checksum (the simple sum of all bytes) of the nanokernel code.
;   The checksum is here for security reasons, to prevent that accidental
;   damages occured to the kernel (or while reading the kernel from the disk)
;   should hang the machine. If the checksum of bytes 8000h .. FC00h as they
;   appear in memory, is not equal to the value of this field, the bootstrap
;   code will inform the user, wait for a keystroke and reboot.
;   The checksum can be temporarily disabled to allow for quick changes while
;   developing the nanokernel, by simply setting this field to zero. It is
;   pratically impossible that the 32-bit checksum should be zero, otherwise.
; sgntr: (32-bit)
;   Contains a 32-bit value which, if interpreted as a small-endian string,
;   translates to exactly 'VPBX'. This field should be used to see if a VPB
;   is present at all, before accessing any other data from the VPB itself.
;   Also, if the VPB signature is present, it means the disk is formatted
;   for use with Lasting Child, and also that the disk is bootable. All the
;   disks, no matter if they're dedicated to data or to be system disks, are
;   bootable under Lasting Child, due to the tiny size of the system kernel.
; vlabel: (text string)
;   It is a blank-padded string (I mean: remaining characters must be padded
;   using blank spaces, having ASCII code 32). It's the volume's name, and can
;   be upto 64 bytes, or ASCII characters, in size. It is ignored by the boot-
;   strap code, since it's considered to be just informative...
; reserved: (32-bit)
;   Reserved for future use, and must be zero...
; 

; 
;  L.C. VTOC entry format (16 bytes per slot)                               
; 
;  (TYPE) File type ID string, NULL = free slot     00000000-00000003     4b
;  (QOND) Quick Object Name Disambiguator           00000004-00000007     4b
;  (FPOS) Absolute position of file slot            00000008-0000000F     8b
; 
; TYPE: (32-bit)
;   Do you know what's the file's "extension" under MS-DOS and other systems?
;   Well, this 4-characters string represents the file's type in the same way
;   the extension does. It holds a blank-padded string of upto 4 bytes, formed
;   by plain ASCII codes (eg. 'TEXT'). And nothing more than this...
;   By the way, if this field is zero, it means it's a free slot in the VTOC.
;   For security reasons, the TYPE field is also echoed in the file datagram,
;   along with the file's name and other attributes.
; QOND: (32-bit)
;   The QOND code is an intelligent trick for speeding up the search of files
;   across the VTOC. Since the VTOC, actually, does NOT allow for directories,
;   all the files are indexed within the same, possibly huge, list. It would
;   be extremely slow to find and operate a certain file if it wasn't for the
;   QOND trick. The QOND is a 32-bit code made out of the target file's name,
;   using a special and pretty fast method involving bit-wise operations; it's
;   some sort of "checksum", but it isn't the simple sum of the name's bytes,
;   even though the QOND uses the file name's ASCII codes for the final value.
;   It is also important to understand that the QOND, on itself, is NOT an
;   UNIVOQUE IDENTIFICATOR for the file. Its role is that of indicating a file
;   that, at 99.999999% of probability, is the one we're looking for, but NOT
;   at 100%. So, once the QOND from the VTOC matches the file's name, the F/S
;   still needs to load up the file name from the disk and check if that file
;   is effectively the right one. But, as you can imagine, since there's only
;   1 case against over four billions in which the QOND may indicate the wrong
;   file, the physical scanning of the file's name will be pratically always
;   performed only one time, and almost always confirm it was the right file.
; FPOS: (64-bit)
;   This 64-bit (QWORD) field represents the PBN (Progressive Byte Number,
;   you can imagine the PBN to be like a FLAT address, in 64-bit, used for
;   addressing the volume's memory) of the corresponding file slot.
;   The values of those fields of the VTOC depend on the disk's configuration:
;   configuring a disk means reserving a certain number of slots of given size
;   until the whole free space has been allocated. Allocation is made BEFORE
;   storing files, and NOT while the files are physically stored.
;   A value of zero for FPOS means there is no slot allocated to this entry.
;   That said, the address given by FPOS points to the "File Info Block";
;   immediately after the File Info Block, there's the file data.
; 

; 
;  L.C. File Info Block (structure pointed by FPOS)                         
; 
;  File beacon (a dword always set to F17EDA7Ah)    00000000-00000003     4b
;  File type ID string (echo of TYPE field in VTOC) 00000004-00000007     4b
;  Attributes form:
;   Effective size of attributes form, bytes (=52)  00000008-0000000B     4b
;   Effective size of file name, bytes (=64)        0000000C-0000000F     4b
;   Effective size of file data, bytes              00000010-00000017     8b
;   Effective size of raw file data, bytes          00000018-0000001F     8b
;   Attributes bitmask                              00000020-00000027     8b
;                
;    63                     1312111009080706050403020100
;                
;                                            CTLG               
;                                               TEMP             
;                                                  BACK           
;                                                     SYST         
;                                                        WRIT       
;                                                           READ     
;                                                              EXEC   
;                                                                 DELE 
;   Creation time    (tts past midnight 12/31/1975) 00000028-0000002F     8b
;   Last write       (tts past midnight 12/31/1975) 00000030-00000037     8b
;  File name                                        ????????-???????? varies
;  File data                                        ????????-???????? varies
; 
; File beacon: (32-bit)
;   it's a constant dword, assuming the same value for any kind of file.
;   All the files are marked with a beacon, a facility for finding the files
;   again in case the VTOC should get unaccessible, damaged, or corrupted.
;   The value of "file beacon" is always F17EDA7A, as hexadecimal.
;   As you may have noticed, it approximates the string "FILEDATA".
;   Of course the file beacon, on itself, is insufficient for identifying a
;   file, since that hexadecimal pattern could also appear as plain data:
;   so, once the beacon is detected, further checks must be performed to
;   determine if a file was really there. Those checks can be done in the
;   attributes form, or searching for specific type of files (type ID).
; File type ID string: (32-bit)
;   it's the same string stored in the VTOC.
;   The only difference being that, here, it CANNOT be null.
; Effective size of attributes form: (32-bit)
;   the attributes form has some mandatory fields: the ones you can see in
;   the above scheme; however, further fields may become mandatory for future
;   versions of the file system, so the size of the attributes form may vary.
;   The value giving the size of the attributes form includes this field.
; Effective size of file name: (32-bit)
;   The file name is an ASCII string and must be padded with blanks upto the
;   number of characters reported by this field. Actually the maximum length
;   for a file name is 64 characters, also meaning 64 bytes.
;   I hope it's enough for a start, even though it isn't very much.
; Effective size of file data: (64-bit)
;   The length, in bytes, of the whole File Info Block structure.
; Effective size of raw file data: (64-bit)
;   The length, in bytes, of the <File data> field.
; Attributes bitmask: (64-bit)
;   Only the seven less significant bits (bits from zero to six) have been
;   defined; remaining bits are reserved for future use and must be zero.
;   Currently defined bits and their equates follow...
;   DELE (equ 01h) - file can be deleted.
;   EXEC (equ 02h) - file is executable (program, module or application);
;   READ (equ 04h) - applications are allowed to read from the file;
;   WRIT (equ 08h) - applications are allowed to write to the file;
;   SYST (equ 10h) - file is part of the system (a system module/archive);
;   BACK (equ 20h) - file is the backup copy of another file (informative);
;   TEMP (equ 40h) - used as temporary storage, a swap file (informative);
;   CTLG (equ 80h) - catalogue (not sure will be implemented in version A);
; Creation time: (64-bit)
;   Can be used to calculate date and time of file's creation.
;   It's a 64-bit SIGNED integer value. Negative values are valid
;   and would reflect creation date/time before the "epoch".
;   The "epoch" is the conventional point relatively to which the file's
;   creation date and time is referred. The epoch has been fixed to
;   December 31, 1975, exactly at midnight (HH:MM:SS = 00:00:00).
;   The time since the epoch is measured in "timer ticks" (tts):
;      1 tt equals to            ~0.0000201192114264 seconds;
;      1 second equals to        ~49704              tts;
;   YES, I KNOW, it all sounds very weird but there are precise reasons:
;      The PC has an internal oscillator linked to its hardware-triggered
;      interrupt gates. The oscillator is part of a chip, called the
;      "Programmable Interval Timer", or PIT for short. The oscillator
;      generates a frequency of 1,193,181 counts per second; now, there
;      is a channel of the PIT, channel zero (out of three channels),
;      which is used by the BIOS to synchronize the real-time clock:
;      for compatibility reasons (some peripherals NEED this channel to
;      pulse, when reading input-ouput port numbered 40h, exactly once
;      every 65,536 counts). If you divide 1,193,181 by 65,536, you get
;      18.20649719 pulses per second. As you can see, one "timer tick"
;      is set to 2730 times that value (=~49704). In facts, the ticks
;      counter (later I'm giving its address) is increased by 2730 units
;      once every those 65,536 counts from the PIT. This happens because
;      when the PIT's count reaches 65,536 counts, it is reset to zero
;      AND an hardware interrupt is generated. By monitoring the interrupt
;      associated with channel zero of the PIT, which is interrupt #8,
;      the system handlers can increase the timer ticks counter by those
;      2730 units. But there's one last thing to explain: WHY 2730?
;      Well, the timer ticks counter holds an INTEGER value, while the
;      number of PIT pulses per second is 18.20649719 (a floating-point
;      value). Because the counter must be divided by some constants to
;      calculate the date and time out of its value, there will be less
;      significant approximations in these constants; for instance, one
;      constant represents how many pulses happen in a minute. If the
;      counter was increasing by 1 unit (not 2730) per pulse, the said
;      constant would be 60 multiplied by 18.20649719, giving 1092.38983...
;      and as you can see, it is AGAIN a floating-point value. That value
;      would be necessarily approximated to 1092 (the nearest integer) in
;      order to perform the (integer) division. Instead, being the timer
;      ticks counter ALREADY multiplied by 2730, the integer constant can
;      approximate its real value much better, and rather than by 1092,
;      the counter can be divided by 1092.38983 * 2730 = 2,982,224.
;      To conclude, why exactly 2730? Because the biggest constant to
;      calculate by the datetime conversion functions is the amount of
;      timer ticks per day. This is (using 2730) 2730 x 86400 x 18.206...
;      which gives 4,294,402,905: you know? That's ALMOST (2 ^ 32) - 1,
;      the greatest unsigned integer that is possible to store in 32 bit.
;      Since the registers used to convert date and time are 32 bit, the
;      multiplier 2730 is the largest (and most precise) multiplier that
;      we could use. With more than 2730, it wouldn't fit 32 bit anymore...
;   The actual (synched to the real-time clock) timer ticks count can be
;   gathered from absolute memory location number 7E00h (constant @tts);
;   of course that location is 64 bits in size, and updated in real time.
; Last write: (64-bit)
;   Date/time of the latest write access to the file.
;   Same notation as for the creation time.
; File name: (actually 64 bytes)
;  PBN of file name =
;                    = VTOC->FPOS
;                    + 8
;                    + <Effective size of attributes form>
;   Not very much to say...
;   A plain ASCII string, blank-padded upto <Effective size of file name>.
; File data:
;  PBN of file data
;                    = VTOC->FPOS
;                    + 8
;                    + <Effective size of attributes form>
;                    + <Effective size of file name>
; 

; 
;  Notes about file slots and the File System in general:                   
; 
;  A file slot is a contiguous fragment of the disk's storage containing the
;   corresponding file. Lasting Child's VTOC is always based on contiguous
;   storage of files; I know it presents a few disadvantages, but some special
;   techniques have been implemented in the file system to workarond the most
;   significant disadvantage, being the unability to automate the expansion of
;   files. It works this way: when files become too larger to fit in their
;   actual slot, they're automatically moved to a bigger slot. Columbus' egg.
;  Now, if the TYPE field of the entry is set to zero, it means that the VTOC
;   entry is free, and can be used to add a new file. Of course the VTOC has
;   a precise limit for the number of files that can be stored on the disk;
;   it's given by the size of the VTOC, which depends on the size of the disk.
;   The VTOC's average file size is 8 Kb, so the number of slots that can be
;   allocated on the disk is given by disk size / 8Kb, or "disk size" shifted
;   right (at binary level) by 13 places. On a simple 1.44Mb floppy, it gives
;   180 file slots. Of course, each slot is NOT limited to 8 Kb: the size of
;   the various slots varies depending on what you have to store on the disk,
;   or you may configure the disk for general use by providing many slots of
;   different size.
;   It may look like a very low limit, to have only upto 180 files on a disk.
;   But remember that it depends on the size of the disk, and 180 files are
;   those you can store on a floppy! If you have a 10 Gb hard drive, the nr.
;   of files you can store there raises to 10 Gb / 8 Kb, giving 1,310,720!
;  Now, let's talk of some advantages... (here comes the good).
;   The Lasting Child file system, managing contiguous datagrams (files),
;   and due to its many optimizations, is probably the world's fastest file
;   system. If used properly, the moves a drive's heads have to perform for
;   reading and writing file's contents are MINIMAL. Pratically it's like
;   directly accessing the disk at low-level, upto the theoretical maximum
;   data transfer rate of the drive. To conclude, the main advantage is
;   simply that: extreme file system performance.
;  Well, there is, however, one smaller advantage, and interesting anyway.
;   It's for data security: data can be found and recovered easily from the
;   disk, in case the VTOC should get destroyed, damaged or overwritten;
;   since the files are always stored as contiguous datagrams, if you find
;   the file you can also recover its contents. To find a file without the
;   VTOC is also very easy, and can be totally automatic: the first data in
;   a file's slot is the ASCII file name, a null-terminated string. Find the
;   name of the file by low-level scanning the disk (throught "VolumeIO"),
;   and you will ALSO find all the informations (including the file's size)
;   right after the string giving the name of the file.
;   If, at that point, you also calculate the QOND for that file, it's very
;   easy to rebuild the VTOC entirely. In practice, you will NEVER lose the
;   contents of a disk unless the same contents you're looking for have been
;   touched directly by damage, which is rather improbable, and becomes almost
;   impossible if you also make some backups...
;  One remaining disadvantage, connected to the choice of using contiguous
;   datagrams, is that the disk's storage space will most probably never get
;   ENTIRELY filled. Even though the structure of a disk can be re-organized
;   to fit for specific amounts of files of given size, some of the memory
;   will almost always get wasted. There's nothing to do about this, but the
;   decision to get back to contiguous storage allocation has been made in
;   consideration of the fact that today's devices allow for MASSIVE storage
;   of data; I mean: impressive amounts of memory. This partially compensates
;   the waste of space due to contiguous storage allocation, and I'm pretty
;   sure that many open-minded users will perfectly understand what I mean.
; 

; 
;  L.C. System Memory Map                                                   
; 
;  Real-mode Interrupt Vectors Table                00000000-000003FF     1K
;  ROM BIOS Communication area                      00000400-000004FF  0.25K
;  Page Zero (system data area)                     00000500-00007FFF 30.25K
;    rmstck   base of real-mode stack, used during interrupts redirection
;    rec      text memory recovery (4000 bytes used) at entry-exit from int.
;    sysidt   Interrupts Descriptors Table for flat-mode
;    irtab    interrupts redirection table
;    hist     hypertext interface history (1024 entries, circular buffer)
;    cp       Control Panel's resident hypertext contest
;    aa       Active Areas (links and TEDs) informations
;    hdb      Hard Data Backup, holds copies of original IVT and some of BDA
;    bootst   Bootstrap loader code, still here, never discarded...
;    gp       512 bytes reserved by the system for general-purpose data
;  Memory manager code, 16to32-bit memory extender  00008000-000088FF  2.25K
;  System code, 32-bit FLAT code, nanokernel        00008900-0000FBFF 28.75K
;  System code backup (first 31.5Kb of boot disk)   0000FC00-000179FF  31.5K
;  Global Access Table (GAT)                        00017A00-0004FFFF 225.5K
;  Drive and media I-O transfer buffer              00050000-0005FFFF  64.0K
;  Memory Tables                                    00060000-0006FFFF  64.0K
;  Hypertext contest buffer (user interface)        00070000-0008FFFF 128.0K
;  VTOC input-output buffer (optimizes file system) 00090000-00097FFF  32.0K
;  Machine stack (default placement of stack)       00098000-0009FBFF  31.0K
; 
;  End conventional memory - upper memory follows...
; 
;  Reserved to the BIOS for possible extensions     0009FC00-000A0000     1K
;  Legacy graphics video memory, and unused         000A0000-000AFFFF    64K
;  Monochrome text video memory, and unused         000B0000-000B7FFF    32K
;  Color text video memory (default screen mode)    000B8000-000BFFFF    32K
;  Video ROM BIOS                                   000C0000-000C7FFF    32K
;  Possible Hard Drive BIOS                         000C8000-000CFFFF    32K
;  Legacy Upper Memory Blocks (UMB), and unused     000D0000-000EFFFF   128K
;  System ROM BIOS                                  000F0000-000FFFF0 65520b
;  Warm boot handler                                000FFFF0-000FFFFF    16b
; 
;  End upper memory - expanded memory (above the first Mb) follows...
; 
;  Available for programs                           00100000-????????   REST
;  Video Linear Frame Buffer (in 32-bit addx space) (VARIES)-(VARIES)    N/A
;  APIC and other memory-mapped chipset registers   FEC00000-FFFF0000    N/A
;  Extended ROM BIOSes                              FFFF0000-FFFFFFFF    64K
; 
; 

.model tiny
.code
.486p

; 
;  MEMORY MAPPINGS IN DETAIL                                                
; 
;               @CONST  = ADDRESS ; N-R ;   SIZE  DESCRIPTION
; --------------------------------;-----;-------------------------------------
                @rmivt  = 000000h ;  *  ;   1.00K Real-mode IVT
                @bda    = 000400h ;  *  ;   0.25K BIOS Data area
                @simtx  = 000500h ;     ;   4.00K Simulated text memory
                @rec    = 001500h ;     ;   4.00K Exception recovery text mem.
                @rmst   = 002500h ;     ;   8.00K  Real-mode int. stack frame
                @rmstop = 004500h ;     ;   =next  R-M int. stack frame top
                @fmidt  = 004500h ;     ;   2.00K Flat-mode IDT
                @f2r_it = 004D00h ;     ;   1.00K FM-to-RM Redirection Table
                @hist   = 005100h ;     ;   4.00K  H-T Interface history
                @hrange = 0003FFh ;     ; x1024-1  History addressing range
                @cp     = 006100h ;     ;   1.50K Control Panel H-T contest
                @aa     = 006700h ;     ;   4.00K  Active Area info
                @aarang = 0002AAh ;     ;    x682  Max active areas (=682)
                @hdb    = 007700h ;  +  ;   1.25K System hard data backup
                @bootst = 007C00h ;  *  ;   0.50K Boot code as loaded by BIOS
                @discst = 007E00h ;  +  ; (31.50k) Disc Strap (bootstrap+krnl)
                @gp     = 007E00h ;  +  ;   0.50K System general-purpose data
                                  ;-----;------------------------------------
                @tts    = 007E00h ;  +  ;      8b Real-time timerticks counter
                @fmuth  = 007E08h ;  +  ;  9 x 4b Flat-mode user tick handlers
                @rmuth  = 007E30h ;  +  ;  9 x 4b Real-mode user tick handlers
                @fnstat = 007E58h ;  +  ;     64b FnStatusLog (upto 64 bytes)
                @rmds   = 007E98h ;  +  ;      2b DS segment in R-M interrupts
                @rmes   = 007E9Ah ;  +  ;      2b ES segment in R-M interrupts
                @rmfs   = 007E9Ch ;  +  ;      2b FS segment in R-M interrupts
                @rmgs   = 007E9Eh ;  +  ;      2b GS segment in R-M interrupts
                @app_st = 007EA0h ;  +  ;      4b Top byte of flat-mode stack
                @f2r_st = 007EA4h ;  +  ;      2b Top of int. flat2real stack
                                  ;-----;------------------------------------
                @mm     = 008000h ;  +  ;   2.25K Memory manager data and code
                @cb     = 008900h ;  +  ;  28.75K System32 resident code
                @krnl   = 00FC00h ;  +  ;  31.50K Bootstrap+kernel code backup
                @gat    = 017A00h ;     ; 209.50K  Global Access Table (GAT)
                @gatsz  = 00D180h ;     ;  x53632  GAT Size
                @phist  = 04C000h ;     ;  16.00K  Prompt history
                @phsize = 004000h ;     ; 16x1024  prompt history lines
                @io     = 050000h ;     ;  64.00K  Mass storage I-O buffer
                @seg_io = 005000h ;     ;  @io/16  (corresponding segment)
                @mt     = 060000h ;     ;  64.00K  Memory tables
                @mlocks = 002000h ;     ;   x8192  Maximum lockable blocks
                @ht     = 070000h ;     ; 128.00K Hypertext contest buffer
                @vtoc   = 090000h ;     ;  32.00K  VTOC input-output buffer
                @vtsize = 008000h ;     ;    x32K  buffer size in bytes
                @fmst   = 098000h ;     ;  31.00K  Def. flat-mode stack frame
                @fmstop = 09FC00h ;     ;   =next  Def. F-M stack frame top
                @lk     = 09FC00h ;  *  ;   1.00K Last Kilobyte (of base mem.)
                @vpf    = 0A0000h ;  *  ;  64.00K Video Page Frame
                @tx     = 0B8000h ;  *  ;  32.00K Text video memory
; --------------------------------;-----;-------------------------------------
; Meaning of N-R signs:
;  * = Non-relocable or non-resizeable because hard-wired
;  + = Non-relocable or non-resizeable for compatibility reasons
; ----------------------------------------------------------------------------

; 
;  BOOTSTRAP LOADER                                                         
; 
;  Mass memory frame: 0-1FFh (0.5 Kb)
;  Memory frame: 7C00h-7DFFh (0.5 Kb)
; 
; 

Bootstrap       segment para    public  use16   'Bootstrap'
                assume  CS:Bootstrap, DS:Bootstrap, SS:nothing
ADDRESS7C00:    jmp     @@sVPB          ; jmp to absolute offset 104 over VPB
BID             dd      00h             ; absolute offset: 02, size 4 bytes
NTK             dd      50h             ; absolute offset: 06, size 4 bytes
NHD             dd      02h             ; absolute offset: 10, size 4 bytes
SPT             dd      12h             ; absolute offset: 14, size 4 bytes
BPS             db      09h             ; absolute offset: 18, size 1 byte
STL             db      00h             ; absolute offset: 19, size 1 byte
vtoc            dd      32256           ; absolute offset: 20, size 4 bytes
reserved1       dd      00000000h       ; absolute offset: 24, size 4 bytes
cksum           dd      'TEST'          ; absolute offset: 28, size 4 bytes
sgntr           dd      'VPBX'          ; absolute offset: 32, size 4 bytes
reserved2       dd      00000000h       ; absolute offset: 36, size 4 bytes
vlabel          db      'Hactar'        ; absolute offset: 40, size 64 bytes
                db      ' System Disk'  ; .
                db      46 dup (32)     ; volume label padder
@@sVPB:         cli                     ; * IRQ off (stack not ready)
                xor     ax, ax          ;  ax = 0000h    
                mov     ss, ax          ;  ss = 0000h     real-mode stack
                mov     sp, @rmstop     ;  sp = 4500h    
                mov     ax, @bootst/16  ;  temporary data segment
                mov     ds, ax          ;  DS = 7C0h (BIOS loads me there)
                sti                     ; * IRQ on (real-mode stack now ready)
                cmp byte ptr ss:[449h],3; 
                je      @@mod3          ;  set video mode 80x25, color
                mov     ax, 3           ;  if not already set...
                int     10h             ; 
@@mod3:         lea     bp, intro       ;  write "Loading Hactar"
                mov     dx, 1800h       ;  at line 24, column zero,
                mov     bx, 0Eh         ;  intensified yellow, page 0,
                mov     cx, 16          ;  message is 16 characters,
                mov     ax, 1301h       ;  BIOS function (textout),
                push    ds              ;  ES must point to DS because the
                pop     es              ;  textout function needs ES:BP
                int     10h             ;  BIOS VIDEO SERVICES
                mov     ax, @discst/16  ;  system code segment (@discst/16)
                mov     es, ax          ;  ES = 7E0h, 31.5K from address zero
                mov     ax,word ptr vtoc; 
                mov     cl, BPS         ;  calculate number of sectors
                shr     ax, cl          ;  to be loaded as vtoc / BPS
                inc     al              ; ...+1 to avoid rounding problems...
                mov     STL, al         ; 
                sub     BPS, 4          ; * make BPS in paragraphs, not bytes
                movzx   si, byte ptr STL;  number of sectors to load
                movzx   dx, byte ptr BID;  from this drive (and head zero)
                xor     bp, bp          ;  from cylinder zero
                mov     di, 1           ;  from sector one
@@read:         mov     cx, di          ;  CL = current sector
                mov     bx, bp          ;  BX = current track
                mov     ch, bl          ;  CH = track number
                mov     ax, 0201h       ;  read 1 sector per call (AL = 01)
                xor     bx, bx          ;  offset zero of destination segment
                int     13h             ;  LEGACY ISA DRIVE SERVICES
                jnc     @@done          ;  check error status throught carry
                cmp     ah, 11h         ;  error code: corrected ECC/CRC
                je      @@done          ;  (go on, that wasn't a real error)
                cmp     ah, 06h         ;  error code: disk changed
                je      @@read          ;  (ok, we realized it's changed)
                jmp     @@crit          ;  else, there's nothing left to do
@@done:         pusha                   ; * save all regs for next read
                push    es              ; * save ES (it's the destination seg)
                movzx   dx, byte ptr STL;  adjust coordinates for dots on
                sub     dx, si          ;  screen to mean the number of
                add     dx, 16          ;  sectors loaded so far...
                or      dx, 1800h       ;  (each dot = 1 sector)
                lea     bp, dot         ;  write single dot
                mov     bx, 0Fh         ;  intensified white, page 0,
                mov     cx, 1           ;  1 character,
                mov     ax, 1301h       ;  BIOS function (textout),
                push    ds              ;  ES must point to DS because the
                pop     es              ;  textout function needs ES:BP
                int     10h             ;  BIOS VIDEO SERVICES
                pop     es              ; * restore ES (destination segment)
                popa                    ; * restore all regs
                dec     si              ;  subtract sector from counter
                jz      @@stop          ;  (all sectors read, so quitting)
                inc     di              ;  next sector to read
                cmp     di, word ptr SPT;  still within this head?
                jbe     @@smcy          ;  (yes)
                mov     di, 1           ;  (no, go to next head, sector 1)
                inc     dh              ;  next head
                cmp     dh, byte ptr NHD;  still within this cylinder?
                jb      @@smcy          ;  (yes)
                xor     dh, dh          ;  (no, so next cyl, sec. 1, head 0)
                inc     bp              ;  next cylinder
@@smcy:         mov     cl, BPS         ;  set bitwise shifter
                mov     ax, 1           ;  we've read 1 sector
                shl     ax, cl          ;  AX = n. of data paragraphs read
                mov     bx, es          ;  BX = above destination segment
                add     bx, ax          ;  BX = next destination segment
                mov     es, bx          ;  ES = new destination segment
                jmp     @@read          ;  restart the loop to next sector
@@crit:         lea     bp, error       ;  write "Load error! etc..."
                mov     dx, 1800h       ;  at line 24, column zero,
                mov     bx, 0Ch         ;  intensified red, page 0,
                mov     cx, 32          ;  message is 32 characters,
                mov     ax, 1301h       ;  BIOS function (textout),
                push    ds              ;  ES must point to DS because the
                pop     es              ;  textout function needs ES:BP
                int     10h             ;  BIOS VIDEO SERVICES
                xor     ax, ax          ;  that's the parameter for int 16h:
                int     16h             ;  halt system until next keypress,
                int     19h             ;  then reload the bootstrap code...
@@stop:         mov     cx,word ptr vtoc;  bytes in checksum (sys code size)
                sub     cx, 512         ;  (minus size of bootstrap loader)
                mov     bx, +400h       ;  +400h from 7C00h (DS) = 8000h
                xor     eax, eax        ;  zero accumulator
@@csum:         movzx edx, byte ptr [bx];  get byte
                add     eax, edx        ;  sum
                inc     bx              ;  next byte
                dec     cx              ;  decrease remaining bytes
                jnz     @@csum          ;  checksum loop
                cmp     cksum, 'TEST'   ;  mantainance/debug signal
                je      @@strt          ;  (skip if checksum value is 'TEST')
                cmp     eax, cksum      ;  checksum test
                je      @@strt          ;  (start system code, if successful)
                lea     bp, cserr       ;  write "Checksum error! etc..."
                mov     dx, 1800h       ;  at line 24, column zero,
                mov     bx, 0Ch         ;  intensified red, page 0,
                mov     cx, 36          ;  message is 36 characters,
                mov     ax, 1301h       ;  BIOS function (textout),
                push    ds              ;  ES must point to DS because the
                pop     es              ;  textout function needs ES:BP
                int     10h             ;  BIOS VIDEO SERVICES
                xor     ax, ax          ;  that's the parameter for int 16h:
                int     16h             ;  halt system until next keypress,
                int     19h             ;  then reload the bootstrap code...
@@strt:         xor     ax, ax          ; * preparing for page zero backups...
                mov     ds, ax          ;  DS = 0000h source segm. of R-M IVT
                mov     es, ax          ;  ES = 0000h backup segm. of R-M IVT
                mov     si, ax          ;  SI = 0000h source offs. of R-M IVT
                mov     di, @hdb        ;  DI = 7700h backup offs. of R-M IVT
                mov     cx, 41Ah        ;  CX = bytes to copy (IVT + BDA)
                cld                     ;  (copy forward)
                rep     movsb           ;  preserving original IVT and BDA
                mov     ax, @krnl / 16  ;  ES = 0FC0h backup segm. of system
                mov     es, ax          ;  kernel code + bootstrap loader
                mov     si, @discst     ;  SI = 7E00h source offs. of @discst
                xor     di, di          ;  DI = 0000h backup offs. of @discst
                mov     cx, 7E00h       ;  CX = bytes to copy (31.5 Kb)
                rep     movsb           ;  preserving all system code
                db      0EAh            ;  direct opcode: jump far, full ptr.
                dw      @mm             ;  destination offset (8000h)
                dw      0               ;  destination segment (0000h)
intro           db      '> Loading Hactar'
error           db      'Load error! Any key to reboot...'
cserr           db      'Checksum error! Any key to reboot..'
dot             db      '.'
Bootstrap       ends

; 
;  FLAT MEMORY AND INTERRUPTS MANAGER                                       
; 
;  Mass memory frame: 200h-AFFh (2.25 Kb)
;  Memory frame: 8000h-88FFh (2.25 Kb)
; 
; 

MemoryManager   segment para    public  use16   'MemoryManager'
                assume  CS:MemoryManager, DS:MemoryManager, SS:nothing
;
descriptor      struc
                limit   dw      ?
                base_1  dw      ?
                base_2  db      ?
                access  db      ?
                misc    db      ?
                base_3  db      ?
descriptor      ends
int_gate        struc
                offs_1  dw      ?
                ic_sel  dw      ?
                tword   dw      ?
                offs_2  dw      ?
int_gate        ends
;
SNULL           equ     00              ; 00) null selector on top of GDT
ICODE           equ     08              ; 08) int. code selector for real mode
IDATA           equ     16              ; 16) int. data selector for real mode
FCODE           equ     24              ; 24) code selector of entire memory
FDATA           equ     32              ; 32) data selector of entire memory
GDT_SYS         equ     5               ; number of system selectors in GDT
GDT_LIM         equ     GDT_SYS * 8 - 1 ; LIMIT FOR GDT = (SELECTORS * 8) - 1
int_code        equ     8E00h           ; access type word for interrupt gate
trap_code       equ     8F00h           ; access type word for trap gate
POSTresetflag   equ     472h            ; action to perform after reset (word)
PICmaster       equ     020h            ; PIC master port
PICslave        equ     0A0h            ; PIC slave port
d_EOI           equ     020h            ; default End-Of-Interrupt command
IRQ0            equ     8               ; interrupt number of IRQ line 0
;
ADDRESS8000:    cli                     ; Turn IRQs off: raw operations below!
                xor     dx, dx          ; (0000h for real-mode CS, DS, ES, SS)
                mov     ds, dx          ; (SS already zero/see bootstrap code)
                mov     es, dx
; ------------- Setup flat-to-real-mode interrupts redirection table ---------
                mov     si, @f2r_it     ; Compile interrupt redirection table
                mov     dx, offset f2r_int_handler[@mm]
                sub     dx, @f2r_it + 3 ; (call <rel16> is 3 bytes long)
                mov     cx, 256         ; (loop counter)
fill_f2r_it:    mov byte ptr [si+0],0E8h; E8 = Call intrasegment (relative)
                mov word ptr [si+1],  dx; dx = actual destination address
                mov byte ptr [si+3], 90h; 90 = NOP (padding call to 4 bytes)
                sub     dx, 4           ; dx is decremented 'cos IP is raising
                add     si, 4
                loop    fill_f2r_it
                mov     si, @fmidt      ; Compile IDT redirection vectors
                xor     cx, cx
fill_f2r_hooks: mov     word ptr [si+0], @f2r_it
                mov     word ptr [si+2], ICODE
                mov     word ptr [si+4], int_code
                mov     word ptr [si+6], 0000h
                add     word ptr [si+0], cx
                add     si, 8
                add     cx, 4
                cmp     cx, 4 * 00FFh
                jbe     fill_f2r_hooks
; ------------- End flat-to-real-mode interrupts redirection table setup -----
; ------------- Setup real-time timerticks counter ---------------------------
                ; link virtual8086-mode reset trap to flat-mode IDT
                mov word ptr ds:[@fmidt+ 32h*8+2], ICODE
                mov word ptr ds:[@fmidt+ 32h*8+6], 0000h
                mov word ptr ds:[@fmidt+ 32h*8+0], offset f2r_reset_trap[@mm]
                ; link flat-mode timer ticks handler to flat-mode IDT
                mov word ptr ds:[@fmidt+IRQ0*8+2], FCODE
                mov word ptr ds:[@fmidt+IRQ0*8+6], 0000h
                mov word ptr ds:[@fmidt+IRQ0*8+0], offset fmtthandler[@cb]
                ; link real-mode timer ticks handler to real-mode IVT
                mov word ptr ds:[@rmivt+IRQ0*4+2], 0000h
                mov word ptr ds:[@rmivt+IRQ0*4+0], offset rmtthandler[@mm]
                mov eax, ds:[@hdb + IRQ0*4]
                mov vectortoirq0[@mm], eax
                ; compile usertick handlers table
                mov     di, 8 * 4
fill_uthtable:  mov     dword ptr ds:[@fmuth + di], offset dummy_fmuth[@cb]
                mov     dword ptr ds:[@rmuth + di], offset dummy_rmuth[@mm]
                sub     di, 4
                jnc     fill_uthtable
; ------------- End real-time timerticks counter setup -----------------------
; ------------- Setup flat-mode environment ----------------------------------
                call    enablea20
                mov     dword ptr ds:[@app_st], @fmstop
                mov      word ptr ds:[@f2r_st], @rmstop
                lgdt    fword ptr ds:gdtr[@mm]
                lidt    fword ptr ds:idtr[@mm]
                mov     eax, cr0
                or      eax, 1
                mov     cr0, eax
                db      0EAh
                dw      $ + @mm + 4, ICODE
                mov     ax, FDATA
                mov     ss, ax
                mov     esp, ss:[@app_st]
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                sti
; ------------- End flat-mode environment setup ------------------------------
                db      0EAh            ; Far jmp to absolute FLAT address...
                dw      @cb             ; ...to this offset...
                dw      FCODE           ; ...in this segment...
;
gdtr            dw      GDT_LIM, sysgdt[@mm], 0000h
idtr            dw      256 * 8 - 1, @fmidt, 0000h
ivtr            dw      03FFh, 0000h, 0000h
sysgdt          descriptor <00000h, 0000, 00h,  00h, 000h, 00h> ; DUMMY (NULL)
                descriptor <0FFFFh, 0000, 00h,  9Eh, 000h, 00h> ; PAGE 0 CODE
                descriptor <0FFFFh, 0000, 00h,  92h, 000h, 00h> ; PAGE 0 DATA
                descriptor <0FFFFh, 0000, 00h,  9Eh, 0CFh, 00h> ; FLAT CODE
                descriptor <0FFFFh, 0000, 00h,  92h, 0CFh, 00h> ; FLAT DATA
;
f2r_eax         dd      0
f2r_ebx         dd      0
f2r_ecx         dd      0
f2r_edx         dd      0
f2r_esi         dd      0
f2r_edi         dd      0
f2r_esp         dd      0
f2r_ebp         dd      0
f2r_eflags      dd      0
f2r_sm          db      0
f2r_old_eip     dd      0
f2r_errorcode   dd      0
;
bin2asc         db      '0123456789ABCDEF'      ; Byte-to-ASCII conv. table
b2a_bfr         db      '????????'              ; Binary to ASCII conv. buffer
;
em_mm_title     db    10, 'HSP Hactar'
em_severe       db    12, 'Severe fault'
em_recover      db     9, 'Interrupt'
em_faultchoice  db    29, 'ENTER: resume  ESCAPE: reboot'
em_benignchoice db    44, 'ENTER: resume  SPACE: ignore  ESCAPE: reboot'
em_a20fail      db    22, 'Cannot enable gate A20'
em_reboot       db    24, 'Hit any key to reboot...'
em_codeindex    db    06, 'Error#'
em_eipindex     db    06, 'EIP/PC'
em_eflagsindex  db    06, 'EFlags'
em_eaxindex     db    03, 'EAX'
em_ebxindex     db    03, 'EBX'
em_ecxindex     db    03, 'ECX'
em_edxindex     db    03, 'EDX'
em_esiindex     db    03, 'ESI'
em_ediindex     db    03, 'EDI'
em_espindex     db    03, 'ESP'
em_ebpindex     db    03, 'EBP'
em_excepttbl    db    14, 'divide by zero', 00000011b ; CPU00
                db    14, 'debug trap    ', 00000010b ; CPU01
                db    14, 'NMI break call', 00000010b ; NMI02
                db    14, 'breakpoint    ', 00000010b ; S/W03
                db    14, 'overflow      ', 00000010b ; CPU04
                db    14, 'BOUND override', 00000010b ; CPU05
                db    14, 'invalid opcode', 00000011b ; CPU06
                db    14, 'FPU not found ', 00000011b ; CPU07
                db    14, 'double fault! ', 00000011b ; SYS08
                db    14, 'FPU overrun   ', 00000010b ; CPU09
                db    14, 'invalid TSS   ', 00000011b ; CPU10
                db    14, 'seg.tion fault', 00000011b ; CPU11
                db    14, 'stack fault   ', 00000011b ; CPU12
                db    14, 'gen. pr. fault', 00000011b ; CPU13
                db    14, 'page fault    ', 00000011b ; CPU14
                db    14, 'exception #15 ', 00000010b ; ???15
                db    14, 'FP calc. error', 00000010b ; CPU16
                db    14, 'unknown fault ', 00000010b ; ???17-31
;       flags are ->    bit 1 = check for it <ٳ
;                       bit 0 = severe error <
;
enablea20:      push    fs gs           ; save needed segment regs.
                xor     ax, ax          ; to test if it worked,
                mov     fs, ax          ; set FS = 0000, GS = FFFF
                dec     ax
                mov     gs, ax
                call    a20test         ; A20 already enabled?
                jz      a20enabled      ; (yes)
                in      al, 92h         ; A20 enable bit 1 of p92h
                or      al, 2           ; (this is valid for PS/2)
                jmp     $+2
                jmp     $+2
                out     92h, al         ; write to p92h
                call    a20test         ; A20 actually enabled?
                jz      a20enabled      ; (yes)
                call    a20wait         ; (no, so try with AT style)
                jnz     a20ATtry
                mov     al, 0d1h        ; output to i8042 STATUS PORT
                out     64h, al
                call    a20wait         ; wait for safe write
                jnz     a20ATtry        ; proceed
                mov     al, 0dfh        ; tell PORT A to enable A20
                out     60h, al
                call    a20wait         ; wait for safe write
a20ATtry:       mov     cx, 2048        ; try 2048 cycles
a20try_loop:    call    a20test         ; enabled?
                jz      a20enabled      ; (yes)
                in      al, 40h         ; current PIT ch.0 tick
                jmp     $+2
                jmp     $+2
                in      al, 40h
                mov     ah, al          ; save current tick number
a20try_loop_2:  in      al, 40h         ; wait next tick
                jmp     $+2
                jmp     $+2
                in      al, 40h
                cmp     al, ah          ; still same tick?
                je      a20try_loop_2   ; (yes)
                loop    a20try_loop     ; (no, go on with next try)
a20error:       mov     dx, 0000h       ; line 0, column 0
                mov     bl, 0Bh         ; color 0Bh (cyan)
                mov     bp, offset em_mm_title[@mm] + 1
                call    write_string_rm
                mov     dx, 0100h       ; line 1, column 0
                mov     bl, 0Ch         ; color 0Ch (red)
                mov     bp, offset em_a20fail[@mm] + 1
                call    write_string_rm
                mov     dx, 0303h       ; line 3, column 3
                mov     bl, 0Fh         ; color 0Fh (white)
                mov     bp, offset em_reboot[@mm] + 1
                call    write_string_rm
                sti
                xor     ax, ax          ; wait key
                int     16h
                int     19h             ; reboot
a20enabled:     pop     gs fs           ; A20..31 address lines ok.
                ret
a20wait:        xor     cx, cx          ; wait safe write to i8042
_dummy:         jmp     $+2             ; try this 65536 times
                jmp     $+2
                in      al, 64h         ; 8042 status
                test    al, 2           ; is buffer full?
                loopnz  _dummy          ; if yes, wait again
                ret
a20test:        mov     al, fs:[0]      ; al <- [0:0]
                mov     ah, al          ; preserve al
                not     al              ; change it and put it in...
                xchg    al, gs:[10h]    ; ...ffffh:10h (HMA start)
                cmp     ah, fs:[0]      ; set ZF if no echo occured
                mov     gs:[10h], al    ; restore the test byte
                ret                     ; ZF=1 if A20 is now enabled
;
rmtthandler:    pushad                                  ; RM timertick handler
                add     dword ptr ss:[@tts + 0], 2730   ; (increment tts/LSD)
                adc     dword ptr ss:[@tts + 4], 0      ; (increment tts/MSD)
                call    dword ptr ss:[@rmuth + 1 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 2 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 3 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 4 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 5 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 6 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 7 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ss:[@rmuth + 8 * 4]   ; (USER TICK HANDLER)
                popad                                   ; * restore registers
                db      0EAh                            ;  jump to original
vectortoirq0    dd      ?                               ;  handler of IRQ 0
dummy_rmuth:    retf                                    ; dummyusertickhandler
;
;  D  O       31              0                 31              0
;  I  F       ͹                 ͹
;  R              OLD              OLD
;  E  E       ͹   SS:ESP        ͹   SS:ESP
;  C  X                            
;  T  P       ͹           ͹
;  I  A    +10  OLD EFLAGS                 +14  OLD EFLAGS   
;  O  N       ͹                 ͹
;  N  S    +6 OLD CS               +10OLD CS 
;     I       ͹                 ͹
;    O    +2     OLD EIP        NEW       +6     OLD EIP    
;    N       ͹   SS:ESP        ͹
;         +0        CALL/IP             +2   ERROR CODE       NEW
;            ͹           ͹   SS:ESP
;                                          +0        CALL/IP     
;                                             ͹
;                                                            
f2r_int_handler:mov     ss:f2r_eax[@mm], eax
                mov     ss:f2r_esp[@mm], esp
                mov     ax, [esp]
                sub     ax, @f2r_it + 3
                shr     ax, 2
                mov     ss:f2r_ivector[@mm], al
                cmp     word ptr [esp + 6], FCODE
                je      f2r_normal_int
                mov     eax, [esp + 14]
                mov     ss:f2r_eflags[@mm], eax
                mov     eax, [esp + 6]
                mov     ss:f2r_old_eip[@mm], eax
                mov     eax, [esp + 2]
                mov     ss:f2r_errorcode[@mm], eax
                add     ss:f2r_esp[@mm], 2 + 4
                jmp     f2r_exception
f2r_normal_int: cmp     al, 8
                jnb     f2r_above_7
                mov     eax, [esp + 10]
                mov     ss:f2r_eflags[@mm], eax
                mov     eax, [esp + 2]
                mov     ss:f2r_old_eip[@mm], eax
                mov     ss:f2r_errorcode[@mm], 0
                add     ss:f2r_esp[@mm], 2
                jmp     f2r_exception
f2r_above_7:    mov     eax, [esp + 10]
                mov     ss:f2r_eflags[@mm], eax
                add     ss:f2r_esp[@mm], 2
                lidt    fword ptr ss:ivtr[@mm]
                mov     ax, IDATA
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                mov     ss, ax
                mov     eax, cr0
                and     eax, 0FFFFFFFEh
                mov     cr0, eax
                db      0EAh
                dw      $ + @mm + 4, 0
                xor     ax, ax
                mov     ss, ax
                movzx   esp, word ptr ss:[@f2r_st]
                mov     ax, word ptr ss:[@rmds]
                mov     ds, ax
                mov     ax, word ptr ss:[@rmes]
                mov     es, ax
                mov     ax, word ptr ss:[@rmfs]
                mov     fs, ax
                mov     ax, word ptr ss:[@rmgs]
                mov     gs, ax
                mov     eax, ss:f2r_eax[@mm]
                push    ss:f2r_eflags[@mm]
                popfd
                db      0CDh
f2r_ivector     db      004h
                pushfd
                cli
                pop     ss:f2r_eflags[@mm]
                mov     ss:f2r_eax[@mm], eax
                mov     ax, ds
                mov     word ptr ss:[@rmds], ax
                mov     ax, es
                mov     word ptr ss:[@rmes], ax
                mov     ax, fs
                mov     word ptr ss:[@rmfs], ax
                mov     ax, gs
                mov     word ptr ss:[@rmgs], ax
f2r_retfromexc: lidt    fword ptr ss:idtr[@mm]
                mov     eax, cr0
                or      eax, 1
                mov     cr0, eax
                db      0EAh
                dw      $ + @mm + 4, ICODE
                mov     ax, FDATA
                mov     ss, ax
                mov     esp, ss:f2r_esp[@mm]
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                mov     eax, ss:f2r_eflags[@mm]
                mov     [esp + 8], eax
                mov     eax, ss:f2r_eax[@mm]
                iretd
f2r_exception:  mov     ss:f2r_ebx[@mm], ebx
                mov     ss:f2r_ecx[@mm], ecx
                mov     ss:f2r_edx[@mm], edx
                mov     ss:f2r_esi[@mm], esi
                mov     ss:f2r_edi[@mm], edi
                mov     ss:f2r_ebp[@mm], ebp
                cmp     ss:f2r_ivector[@mm], 17
                jb      f2r_knownexc
                mov     ss:f2r_ivector[@mm], 17
f2r_knownexc:   movzx   bp, ss:f2r_ivector[@mm]
                shl     bp, 4
                add     bp, offset em_excepttbl[@mm]
                lidt    fword ptr ss:ivtr[@mm]
                mov     ax, IDATA
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                mov     ss, ax
                mov     eax, cr0
                and     eax, 0FFFFFFFEh
                mov     cr0, eax
                db      0EAh
                dw      $ + @mm + 4, 0
                xor     ax, ax
                mov     ss, ax
                movzx   esp, word ptr ss:[@f2r_st]
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                test    byte ptr [bp + 15], 2
                jz      f2r_ignore
                test    byte ptr [bp + 15], 1
                jz      f2r_recover
f2r_severe:     call    f2r_setsm3
                push    bp              ; save local ptr. to message
                mov     dx, 0000h       ; line 0, column 0
                mov     bl, 0Bh         ; cyan
                mov     bp, offset em_mm_title[@mm] + 1
                call    write_string_rm
                mov     dx, 0100h       ; line 1, column 0
                mov     bl, 0Ch         ; red
                mov     bp, offset em_severe[@mm] + 1
                call    write_string_rm
                mov     dx, 0300h       ; line 3, column 0
                mov     bl, 07h         ; gray
                pop     bp              ; restore local ptr. to msg.
                inc     bp
                call    write_string_rm
                call    show_situation
                mov     dx, 0503h       ; line 5, column 3
                mov     bl, 0Fh         ; white
                mov     bp, offset em_faultchoice[@mm] + 1
                call    write_string_rm
fault_invldkey: sti
                xor     ax, ax
                int     16h             ; wait for keypress
                cli
                cmp     al, 27          ; escape?
                je      f2r_chosen_b
                cmp     al, 13          ; enter?
                je      f2r_chosen_r
                jmp     fault_invldkey  ; none of them?
f2r_recover:    call    f2r_setsm3      ; set text video mode, if necessary
                push    bp bp           ; save local ptr. to message
                mov     dx, 0000h       ; line 0, column 0
                mov     bl, 0Bh         ; cyan
                mov     bp, offset em_mm_title[@mm] + 1
                call    write_string_rm
                mov     dx, 0100h       ; line 1, column 0
                mov     bl, 0Ch         ; red
                mov     bp, offset em_recover[@mm] + 1
                call    write_string_rm
                mov     dx, 0300h       ; line 3, column 0
                mov     bl, 07h         ; gray
                pop     bp              ; restore local ptr. to msg.
                inc     bp
                call    write_string_rm
                mov     dx, 0503h       ; line 5, column 3
                mov     bl, 0Fh         ; white
                mov     bp, offset em_benignchoice[@mm] + 1
                call    write_string_rm
                call    show_situation
                pop     bp              ; * (to pass it to "f2r_chosen_i")
recover_invkey: sti
                xor     ax, ax
                int     16h             ; wait for keypress
                cli
                cmp     al, 32          ; space?
                je      f2r_chosen_i
                cmp     al, 27          ; escape?
                je      f2r_chosen_b
                cmp     al, 13          ; enter?
                je      f2r_chosen_r
                jmp     recover_invkey  ; none of them?
f2r_chosen_i:   and     byte ptr [bp + 15], 0FDh
               ;jmp     f2r_chosen_r
f2r_chosen_r:   call    f2r_offsm3
                jmp     f2r_ignore
f2r_chosen_b:   mov     word ptr ss:[POSTresetflag], 0000h
                jmp     f2r_reset_jump
f2r_ignore:     mov     ebx, ss:f2r_ebx[@mm]
                mov     ecx, ss:f2r_ecx[@mm]
                mov     edx, ss:f2r_edx[@mm]
                mov     esi, ss:f2r_esi[@mm]
                mov     edi, ss:f2r_edi[@mm]
                mov     ebp, ss:f2r_ebp[@mm]
                jmp     f2r_retfromexc
show_situation: mov     dx, 003Fh       ; line 0, column 63
                mov     bl, 0Ah         ; light green
                mov     bp, offset em_codeindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0048h       ; line 0, column 72
                mov     eax, ss:f2r_errorcode[@mm]
                call    write_hexval_rm
                mov     dx, 013Fh       ; line 1, column 63
                mov     bl, 0Ah         ; light green
                mov     bp, offset em_eipindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0148h       ; line 1, column 72
                mov     eax, ss:f2r_old_eip[@mm]
                call    write_hexval_rm
                mov     dx, 023Fh       ; line 2, column 63
                mov     bl, 0Ah         ; light green
                mov     bp, offset em_eflagsindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0248h       ; line 2, column 72
                mov     eax, ss:f2r_eflags[@mm]
                call    write_hexval_rm
                mov     bl, 0Fh         ; intensified white
                mov     dx, 043Fh       ; line 4, column 63
                mov     bp, offset em_eaxindex[@mm] + 1
                call    write_string_rm
                mov     dx, 053Fh       ; line 5, column 63
                mov     bp, offset em_ebxindex[@mm] + 1
                call    write_string_rm
                mov     dx, 063Fh       ; line 6, column 63
                mov     bp, offset em_ecxindex[@mm] + 1
                call    write_string_rm
                mov     dx, 073Fh       ; line 7, column 63
                mov     bp, offset em_edxindex[@mm] + 1
                call    write_string_rm
                mov     dx, 083Fh       ; line 8, column 63
                mov     bp, offset em_esiindex[@mm] + 1
                call    write_string_rm
                mov     dx, 093Fh       ; line 9, column 63
                mov     bp, offset em_ediindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0A3Fh       ; line 10, column 63
                mov     bp, offset em_espindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0B3Fh       ; line 11, column 63
                mov     bp, offset em_ebpindex[@mm] + 1
                call    write_string_rm
                mov     dx, 0443h       ; line 4, column 67
                mov     eax, ss:f2r_eax[@mm]
                call    write_hexval_rm
                mov     dx, 0543h       ; line 5, column 67
                mov     eax, ss:f2r_ebx[@mm]
                call    write_hexval_rm
                mov     dx, 0643h       ; line 6, column 67
                mov     eax, ss:f2r_ecx[@mm]
                call    write_hexval_rm
                mov     dx, 0743h       ; line 7, column 67
                mov     eax, ss:f2r_edx[@mm]
                call    write_hexval_rm
                mov     dx, 0843h       ; line 8, column 67
                mov     eax, ss:f2r_esi[@mm]
                call    write_hexval_rm
                mov     dx, 0943h       ; line 9, column 67
                mov     eax, ss:f2r_edi[@mm]
                call    write_hexval_rm
                mov     dx, 0A43h       ; line 10, column 67
                mov     eax, ss:f2r_esp[@mm]
                call    write_hexval_rm
                mov     dx, 0B43h       ; line 11, column 67
                mov     eax, ss:f2r_ebp[@mm]
                call    write_hexval_rm
                mov     dx, 044Ch       ; line 4, column 76
                mov     eax, ss:f2r_eax[@mm]
                call    write_idstr_rm
                mov     dx, 054Ch       ; line 5, column 76
                mov     eax, ss:f2r_ebx[@mm]
                call    write_idstr_rm
                mov     dx, 064Ch       ; line 6, column 76
                mov     eax, ss:f2r_ecx[@mm]
                call    write_idstr_rm
                mov     dx, 074Ch       ; line 7, column 76
                mov     eax, ss:f2r_edx[@mm]
                call    write_idstr_rm
                mov     dx, 084Ch       ; line 8, column 76
                mov     eax, ss:f2r_esi[@mm]
                call    write_idstr_rm
                mov     dx, 094Ch       ; line 9, column 76
                mov     eax, ss:f2r_edi[@mm]
                call    write_idstr_rm
                mov     dx, 0A4Ch       ; line 10, column 76
                mov     eax, ss:f2r_esp[@mm]
                call    write_idstr_rm
                mov     dx, 0B4Ch       ; line 11, column 76
                mov     eax, ss:f2r_ebp[@mm]
                call    write_idstr_rm
                ret
f2r_setsm3:     mov     ax, 0B800h;@tx/16
                mov     ds, ax
                mov     ax, 00150h;@rec/16
                mov     es, ax
                mov     cx, 80*25
                xor     si, si
                xor     di, di
                cld
                rep     movsw
                push    ds
                pop     es
                mov     cx, 80*25
                xor     di, di
                mov     ax, 0720h
                cld
                rep     stosw
                xor     ax, ax
                mov     ds, ax
                mov     es, ax
                ret
f2r_offsm3:     mov     ax, 00150h;@rec/16
                mov     ds, ax
                mov     ax, 0B800h;@tx/16
                mov     es, ax
                mov     cx, 80*25
                xor     si, si
                xor     di, di
                cld
                rep     movsw
                xor     ax, ax
                mov     ds, ax
                mov     es, ax
                ret
f2r_reset_trap: lidt    fword ptr ss:ivtr[@mm]
                mov     ax, IDATA
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
                mov     ss, ax
                mov     eax, cr0
                and     eax, 0FFFFFFFEh
                mov     cr0, eax
                db      0EAh
                dw      $ + @mm + 4, 0
                xor     ax, ax
                mov     ss, ax
                mov     esp, @rmstop
                mov     ds, ax
                mov     es, ax
                mov     fs, ax
                mov     gs, ax
f2r_reset_jump: db      0EAh            ; far jump to absolute ptr.
                dw      0FFF0h          ; offset  FFF0
                dw      0F000h          ; segment F000
;
write_string_rm:movzx   cx, byte ptr ss:[bp-1]
                mov     ax, 1301h
                xor     bh, bh
                int     10h
                ret
write_hexval_rm:mov     cx, 8
                xor     bp, bp
wrhvrm_loop:    dec     bp
                mov     bx, ax
                and     bx, 0Fh
                mov     bh, ss:bin2asc[@mm + bx]
                mov     ss:b2a_bfr[@mm + bp + 8], bh
                ror     eax, 4
                loop    wrhvrm_loop
                mov     bp, offset b2a_bfr[@mm]
                mov     ax, 1301h
                mov     bx, 07h
                mov     cx, 8
                int     10h
                ret
write_idstr_rm: cmp     al, 20h
                jb      notastring
                cmp     ah, 20h
                jb      notastring
                bswap   eax
                cmp     al, 20h
                jb      notastring
                cmp     ah, 20h
                jb      notastring
                mov     dword ptr ss:b2a_bfr[@mm], eax
                mov     bp, offset b2a_bfr[@mm]
                mov     ax, 1301h
                mov     bx, 19h
                mov     cx, 4
                int     10h
notastring:     ret
;
align           16
                dw      2 * 8 dup (0DB87h)
MemoryManager   ends

; 
;  L.C. System32                                                            
; 
;  Mass memory frame: 0B00h-7DFFh (28.75 Kb)
;  Memory frame: 8900h-FBFFh (28.75 Kb)
;  Additional memory frames: FC00h-9FBFFh (576 Kb)
; 
; 

System32        segment para    public  use32   'System32'
                assume  CS:System32, DS:System32, SS:nothing
;
htl             equ     21              ; hypertext lines on screen
KBhead          equ     41Ah            ; head of keyboard's circular buffer
KBtail          equ     41Ch            ; tail of keyboard's circular buffer
;
ADDRESS8900: mov word ptr ds:[@rmds], 0 ; Initialize real-mode segments
             mov word ptr ds:[@rmes], 0
             mov word ptr ds:[@rmfs], 0
             mov word ptr ds:[@rmgs], 0
; ------------- Perform hard data backup checksum ----------------------------
                xor     eax, eax
                mov     esi, @hdb
                mov     ecx, 41Ah
hdbchecksum_mk: add     eax, [esi]
                inc     esi
                dec     ecx
                jnz     hdbchecksum_mk
                mov     hdbchecksum[@cb], eax
; ------------- Initialize tiriticks counter ---------------------------------
                mov     dword ptr ds:[@tts+0], 0
                mov     dword ptr ds:[@tts+4], 0
                call    ReSyncTTCounter
                mov     dword ptr ds:[@fmuth + 1 * 4], offset cp_uth[@cb]
; ------------- Mount central memory -----------------------------------------
                call    MountMemory
; ------------- Initialize GAT as void ---------------------------------------
                mov     edi, @gat
                xor     eax, eax
                mov     ecx, @gatsz
                cld
                rep     stosd
; ------------- Compile resident GAT -----------------------------------------
                mov     edi, @gat
                mov     dword ptr [edi], @gatsz - 1
                add     edi, 4
                lea     esi, ResidentGAT[@cb]
                mov     ecx, ResidentGATsz
                cld
                rep     movsd
; ------------- Build new-style 8x16 text font -------------------------------
                lea     edi, @FM[@cb]
                lea     esi, [edi + 64]
                mov     edx, 128
nextchunk:      mov     ebx, 04
nextshape:      mov     ecx, 16
charshape:      mov     al, [esi]
                mov     [edi], al
                add     esi, 4
                inc     edi
                dec     ecx
                jnz     charshape
                sub     esi, 63
                dec     ebx
                jnz     nextshape
                add     esi, 64 - 4
                sub     edx, 4
                jnz     nextchunk
; ------------- Send font to characters generator ----------------------------
                mov     ax, 1100h
                mov     bx, 1000h
                xor     dx, dx
                mov     cx, 128
                lea     ebp, @FM[@cb]
                int     10h
; ------------- Disable blinking for intensity bit ---------------------------
                mov     ax, 1003h
                xor     bx, bx
                int     10h
; ------------- Disable cursor (make it invisible) ---------------------------
                mov     ax, 0103h
                mov     cx, 2808h
                int     10h
; ------------- Run system shell ---------------------------------------------
         mov dword ptr ds:[@cp], '' ; Initialize cp area as empty
         mov dword ptr ds:[@ht], '' ; Initialize ht area as empty
                mov     edi, @cp        ; Setup control panel's hypertext
                mov     eax, offset ht_commandlist[@cb]
                call    wrh_next        ; (special call, ONLY for this case)
                xor     ebx, ebx        ; Start (new) with the system screen
                mov     eax, offset SystemScreen[@cb]
                call    WrHyper         ; (loading this other page as normal)
freshscreen:    call    writestatusline ; Control panel's interface
                call    writehtcontest
                call    writecpdiv
waitcpselect:   cmp     moreflag[@cb], 1
                jbe     skipmoreblink
                mov     al, byte ptr ds:[46Ch]
                shr     al, 2
                and     al, 1
                mov     ah, ds:[@tx + htl*160 + 80]
                and     ah, 8
                shr     ah, 3
                cmp     al, ah
                je      skipmorearrow
                sub     moreflag[@cb], al
                shl     al, 3
                mov     ds:[@tx + htl*160 + 80], al
                jmp     skipmorearrow
skipmoreblink:  mov     al, moreflag[@cb]
                shl     al, 3
                xor     al, 8
                cmp     al, ds:[@tx + htl*160 + 80]
                je      skipmorearrow
                mov     ds:[@tx + htl*160 + 80], al
skipmorearrow:  mov     ax, word ptr ds:[KBhead]
                cmp     ax, word ptr ds:[KBtail]
                je      waitcpselect
                xor     ax, ax
                int     16h
                cmp     ax, 4D00h
                je      cursorright
                cmp     ax, 4B00h
                je      cursorleft
                cmp     ax, 5000h
                je      cursordown
                cmp     ax, 4800h
                je      cursorup
                cmp     ax, 1C0Dh
                je      select
                cmp     ax, 3920h
                je      select
                cmp     ax, 5100h
                je      nexthtrow
                cmp     ax, 4900h
                je      prevhtrow
                cmp     ax, 7600h
                je      nexthtpag
                cmp     ax, 8400h
                je      prevhtpag
                cmp     ax, 4700h
                je      hthome
                jmp     waitcpselect
m2m             MACRO   dest, source    ; MOV Memory-To-Memory, thru stack
                push    source
                pop     dest
                ENDM
strup           MACRO   idstring        ; bring 4-byte ID string to uppercase
                and     dword ptr idstring, 0DFDFDFDFh
                ENDM
idup            MACRO   idstring        ; bring drive ID string to uppercase
                and     dword ptr idstring, 0DFDFFFFFh
                ENDM
_idup           MACRO   idstring        ; reverse drive ID string to uppercase
                and     dword ptr idstring, 0FFFFDFDFh
                ENDM
dmon            MACRO   dmonD, dmonB    ; update "disk monitor" in status line
                ror dmonD, 24
                mov byte ptr ds:[@tx+3840+2*(offset cpdmon-offset cpsl)],dmonB
                ror dmonD, 8
                mov byte ptr ds:[@tx+3842+2*(offset cpdmon-offset cpsl)],'D'
                mov byte ptr ds:[@tx+3844+2*(offset cpdmon-offset cpsl)],'0'
                mov byte ptr ds:[@tx+3846+2*(offset cpdmon-offset cpsl)],dmonB
                ENDM
dmon____clear   MACRO                   ; clear "disk monitor" in status line
                mov byte ptr ds:[@tx+3840+2*(offset cpdmon-offset cpsl)],32
                mov byte ptr ds:[@tx+3842+2*(offset cpdmon-offset cpsl)],32
                mov byte ptr ds:[@tx+3844+2*(offset cpdmon-offset cpsl)],32
                mov byte ptr ds:[@tx+3846+2*(offset cpdmon-offset cpsl)],32
                ENDM
status          MACRO   sb              ; assert function status
                mov     byte ptr cpstatsp[@cb], sb
                call    assert_status
                ENDM
assert_status:  pushfd                  ; coupled with the above macro
                pushad
                test    quietmode[@cb], 1
                jnz     ast_quiet
                call    write_status
                inc     logoffset[@cb]
                mov     cpstatsp[@cb], 2
                call    write_status
ast_quiet:      popad
                popfd
                ret
write_status:   movzx edi,logoffset[@cb]; coupled with the above macro
                cmp   edi, 23
                jb    wrs_inlogb
                xor   edi, edi
                mov   logoffset[@cb], 0
wrs_inlogb:     mov   al, cpstatsp[@cb]
                mov   ds:[@tx + edi*2 +3840+2*(offset cpstat-offset cpsl)], al
                mov   ds:[@cb + edi + offset cpstat], al
                mov   ds:[@fnstat + edi], al
                ret
end_____fn      MACRO
                jmp     common_fn_ret
                ENDM
common_fn_ret:  status  24              ; coupled with the above macro
                mov     quietmode[@cb], 0
                clc
                ret
fail____fn      MACRO
                jmp     common_fn_fail
                ENDM
common_fn_fail: push    edx             ; coupled with the above macro
                mov     dl, endstatus[@cb]
                status  dl
                status  25
                mov     quietmode[@cb], 0
                pop     edx
                stc
                ret
fullscreenfocal:mov     focal_left[@cb], -1
                mov     focal_top[@cb], -1
                mov     focal_right[@cb], 80
                mov     focal_bottom[@cb], 25
                ret
cursorright:    m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], lowysensitivity[@cb]
                call    fullscreenfocal
                m2m     focal_left[@cb], cursorpos_x[@cb]
                call    move2nearestaa
                jmp     waitcpselect
cursorleft:     m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], lowysensitivity[@cb]
                call    fullscreenfocal
                m2m     focal_right[@cb], cursorpos_x[@cb]
                call    move2nearestaa
                jmp     waitcpselect
cursordown:     m2m     aaxsensitivity[@cb], lowxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                call    fullscreenfocal
                m2m     focal_top[@cb], cursorpos_y[@cb]
                call    move2nearestaa
                jmp     waitcpselect
cursorup:       m2m     aaxsensitivity[@cb], lowxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                call    fullscreenfocal
                m2m     focal_bottom[@cb], cursorpos_y[@cb]
                call    move2nearestaa
                jmp     waitcpselect
select:         cmp     selectedarea[@cb], 0
                je      waitcpselect
                xor     ebx, ebx
select_nrm:     mov     eax, sa_start[@cb]
                mov     ecx, sa_width[@cb]
select_hlg:     xor     byte ptr [eax+1], 0Fh
                add     eax, 2
                dec     ecx
                jnz     select_hlg
                mov     eax, dword ptr ds:[@tts]
select_delay_1: cmp     eax, dword ptr ds:[@tts]
                je      select_delay_1
                mov     eax, dword ptr ds:[@tts]
                add     eax, sa_hlgtiming[@cb]
select_delay_2: cmp     dword ptr ds:[@tts], eax
                jb      select_delay_2
                test    ebx, ebx
                mov     bl, 1
                jz      select_nrm
                mov     esi, selectedarea[@cb]  ; (pointer within @aa)
                mov     esi, [esi + 2]          ; (pointer within @cp or @ht)
                cmp     dword ptr [esi + 4], offset tag_ted[@cb]
                je      runted
                cmp     dword ptr [esi + 4], offset tag_lnk[@cb]
                je      runcontentlink
runexecutable:  mov     eax, [esi + 8]          ; eax = ptr to GOID string
                push    esi                     ; (preserve)
                call    GATLookup               ; eax now = ptr to code
                pop     esi                     ; (restore)
                jc      waitcpselect
                mov     execodeptr[@cb], eax    ; preserve code pointer
                mov     eax, [esi + 12]         ; (parameter/spec #2)
                mov     ebx, [esi + 16]         ; (parameter/spec #3)
                mov     ecx, [esi + 20]         ; (parameter/spec #4)
                mov     edx, [esi + 24]         ; (parameter/spec #5)
                mov     edi, [esi + 32]         ; (parameter/spec #7)
                mov     ebp, [esi + 36]         ; (parameter/spec #8)
                mov     esi, [esi + 28]         ; (parameter/spec #6 - last)
                push    cursorpos_x[@cb]
                push    cursorpos_y[@cb]
                call  dword ptr execodeptr[@cb] ; enter the executable module
                pop     cursorpos_y[@cb]
                pop     cursorpos_x[@cb]
                m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                call    fullscreenfocal
                call    move2nearestaa
                jmp     waitcpselect
runcontentlink: mov     eax, [esi + 8]          ; eax = ptr to GOID string
                push    esi                     ; (preserve)
                call    GATLookup               ; eax now = ptr to data
                pop     esi                     ; (restore)
                jc      waitcpselect
                mov     ebx, [esi + 12]         ; behaviour indication
                push    cursorpos_x[@cb]
                push    cursorpos_y[@cb]
                call    WrHyper                 ; loading the new page...
                pop     cursorpos_y[@cb]
                pop     cursorpos_x[@cb]
                m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                call    fullscreenfocal
                call    move2nearestaa
                jmp     waitcpselect
runted:         mov     eax, [esi + 8]          ; find out text buffer pointer
                pushad                          ;  (will be held in variable
                call    GATLookup               ;  "tedbuffer" for later
                mov     tedbuffer[@cb], eax     ;  retraces and/or edits)
                popad
                jc      waitcpselect
                mov     edi, selectedarea[@cb]  ; compute pointer to the spot,
                movzx   eax, byte ptr [edi + 1] ;  in video memory, being the
                movzx   edi, byte ptr [edi]     ;  upper-left corner of the
                mov     tedsheet_y[@cb], eax    ;  TED. -Pointer goes in EDI.
                mov     tedsheet_x[@cb], edi
                mov     edx, 160
                mul     edx
                add     edi, edi
                add     edi, eax
                add     edi, @tx
                mov     tedcursor_x[@cb], 0     ; relative to upperleft corner
                mov     tedcursor_y[@cb], 0     ;     -    -      -        -
                mov     txstartcol[@cb],  0     ;     -    -      -        -
                mov     txstartrow[@cb],  0     ;     -    -      -        -
                mov     ebx, 1                  ; inflation counters
                mov     ted_vcols[@cb], 1
                mov     ecx, 1
                mov     ted_vrows[@cb], 1
ted_inflate_h:  mov     ted_vcols[@cb], ebx     ; horizontal inflation
                call    ted_retr_text
                call    wait1tick
                inc     ebx
                cmp     ebx, [esi + 16]
                ja      ted_inflate_v
                inc     tedsheet_x[@cb]
                cmp     tedsheet_x[@cb], 80
                jb      ted_inflate_h
                sub     edi, 2
                jmp     ted_inflate_h
ted_inflate_v:  inc     ecx                     ; vertical inflation
                cmp     ecx, [esi + 20]
                ja      ted_control
                inc     tedsheet_y[@cb]
                cmp     tedsheet_y[@cb], 24
                jb      ted_inflate_v1
                sub     edi, 160
ted_inflate_v1: mov     ted_vrows[@cb], ecx
                call    ted_retr_text
                call    wait1tick
                jmp     ted_inflate_v
ted_control:    call    ted_mark_cursor
                pushad
                mov     eax, txstartcol[@cb]
                add     eax, tedcursor_x[@cb]
                inc     eax
                lea     ebx, cptedc[@cb]
                mov     ecx, 3
                call    In2Dec
                mov     eax, txstartrow[@cb]
                add     eax, tedcursor_y[@cb]
                inc     eax
                lea     ebx, cptedl[@cb]
                mov     ecx, 5
                call    In2Dec
                mov     cptedl[@cb + 5], ':'
                call    writestatusline
                popad
ted_waitkey:    xor     ax, ax
                int     16h
                cmp     ax, 4D00h
                je      ted_cursorright
                cmp     ax, 4B00h
                je      ted_cursorleft
                cmp     ax, 5000h
                je      ted_cursordown
                cmp     ax, 4800h
                je      ted_cursorup
                cmp     ax, 4700h
                je      ted_home_key
                cmp     ax, 4F00h
                je      ted_end_key
                cmp     ax, 011Bh
                je      ted_close
                cmp     ax, 0E08h
                je      ted_bksp
                cmp     ax, 5300h
                je      ted_delete
                cmp     ax, 1C0Dh
                je      ted_insertcrlf
                cmp     al, 32
                jnb     ted_insertchar
                jmp     ted_waitkey
ted_close:      mov     dword ptr cptedl[@cb], 20202020h
                mov     byte ptr cptedl[@cb + 4], 20h
                mov     dword ptr cptedc[@cb - 1], 20202020h
                jmp     freshscreen
ted_cursorright:mov     eax, [esi + 16]
                dec     eax
                cmp     tedcursor_x[@cb], eax
                jnl     ted_nextcol
                call    ted_mark_cursor
                inc     tedcursor_x[@cb]
                jmp     ted_control
ted_nextcol:    inc     txstartcol[@cb]
                call    ted_retr_text
                jmp     ted_control
ted_cursorleft: cmp     tedcursor_x[@cb], 1
                jl      ted_prevcol
                call    ted_mark_cursor
                dec     tedcursor_x[@cb]
                jmp     ted_control
ted_prevcol:    cmp     txstartcol[@cb], 1
                jl      ted_waitkey
                dec     txstartcol[@cb]
                call    ted_retr_text
                jmp     ted_control
ted_cursordown: mov     eax, [esi + 20]
                dec     eax
                cmp     tedcursor_y[@cb], eax
                jnl     ted_nextrow
                call    ted_mark_cursor
                inc     tedcursor_y[@cb]
                jmp     ted_control
ted_nextrow:    inc     txstartrow[@cb]
                call    ted_retr_text
                jmp     ted_control
ted_cursorup:   cmp     tedcursor_y[@cb], 1
                jl      ted_prevrow
                call    ted_mark_cursor
                dec     tedcursor_y[@cb]
                jmp     ted_control
ted_prevrow:    cmp     txstartrow[@cb], 1
                jl      ted_waitkey
                dec     txstartrow[@cb]
                call    ted_retr_text
                jmp     ted_control
ted_home_key:   mov     tedcursor_x[@cb], 0     ;  effect
                mov     txstartcol[@cb], 0      ;  of a CR
                call    ted_retr_text
                jmp     ted_control
ted_end_key:    call    ted_retr_text           ; ted_retr_text finds out the
                call    ted_mark_cursor         ;  actual ptr of the byte
                cmp     tedcursorig[@cb], -1    ;  pointed by the cursor, and
                je      ted_waitkey             ;  -1 means no char under cr.
                mov     eax, tedcursorig[@cb]
                mov     edx, [esi + 16]
                dec     edx
ted_find_eol:   cmp     byte ptr [eax], 13
                je      ted_end_retr
                cmp     eax, endmarkorig[@cb]
                jnb     ted_end_retr
                inc     eax
                cmp     tedcursor_x[@cb], edx
                jl      ted_cr2right
                inc     txstartcol[@cb]
                jmp     ted_find_eol
ted_cr2right:   inc     tedcursor_x[@cb]
                jmp     ted_find_eol
ted_end_retr:   call    ted_retr_text
                jmp     ted_control
ted_insertcrlf: mov     ticode[@cb], 10         ; 10 = LF
                call    ted_insertcode
                jc      ted_waitkey
                mov     ticode[@cb], 13         ; 13 = CR (will go before LF)
                call    ted_inswocoupd          ; (ins w/o cursor orgn update)
                mov     tedcursor_x[@cb], 0     ;  effect
                mov     txstartcol[@cb], 0      ;  of a CR
                call    ted_retr_text
                call    ted_mark_cursor
                jmp     ted_cursordown          ; (forces a LF for the cursor)
ted_insertchar: mov     ticode[@cb], al
                call    ted_insertcode
                jc      ted_waitkey
                call    ted_retr_text
                call    ted_mark_cursor
                jmp     ted_cursorright
ted_bksp:       call    ted_retr_text           ; ted_retr_text finds out the
                call    ted_mark_cursor         ;  actual ptr of the byte
                cmp     tedcursorig[@cb], -1    ;  pointed by the cursor, and
                je      ted_waitkey             ;  -1 means no char under cr.
                mov     eax, tedcursorig[@cb]
                cmp     eax, tedbuffer[@cb]
                jbe     ted_waitkey
                cmp     byte ptr [eax - 1], 10
                je      tbk_cursorup
                cmp     tedcursor_x[@cb], 1
                jl      tbk_prevcol
                dec     tedcursor_x[@cb]
                jmp     tbk_again
tbk_prevcol:    cmp     txstartcol[@cb], 1
                jl      tbk_again
                dec     txstartcol[@cb]
                jmp     tbk_again
tbk_cursorup:   cmp     tedcursor_y[@cb], 1
                jl      tbk_prevrow
                dec     tedcursor_y[@cb]
                jmp     tbk_end_key
tbk_prevrow:    cmp     txstartrow[@cb], 1
                jl      tbk_end_key
                dec     txstartrow[@cb]
tbk_end_key:    push    tedcursorig[@cb]
                push    tedcursor_x[@cb]
                push    txstartcol[@cb]
                mov     tedcursor_x[@cb], 0
                mov     txstartcol[@cb], 0
                call    ted_retr_text
                mov     eax, tedcursorig[@cb]
                pop     txstartcol[@cb]
                pop     tedcursor_x[@cb]
                pop     tedcursorig[@cb]
                cmp     eax, -1
                je      tbk_again
                mov     edx, [esi + 16]
                dec     edx
tbk_find_eol:   cmp     byte ptr [eax], 13
                je      tbk_again
                cmp     eax, endmarkorig[@cb]
                jnb     tbk_again
                inc     eax
                cmp     tedcursor_x[@cb], edx
                jl      tbk_cr2right
                inc     txstartcol[@cb]
                jmp     tbk_find_eol
tbk_cr2right:   inc     tedcursor_x[@cb]
                jmp     tbk_find_eol
tbk_again:      mov     eax, tedcursorig[@cb]
                cmp     eax, tedbuffer[@cb]
                jbe     ted_waitkey
                dec     endmarkorig[@cb]
                dec     tedcursorig[@cb]
                mov     eax, tedcursorig[@cb]
                mov     ch, [eax]
                mov     ticode[@cb], ch
tbk_backspacer: mov     ch, [eax + 1]
                mov     [eax], ch
                cmp     eax, endmarkorig[@cb]
                jae     tbk_endloop
                inc     eax
                jmp     tbk_backspacer
tbk_endloop:    cmp     ticode[@cb], 10
                jne     tbk_endbksp
                mov     eax, tedcursorig[@cb]
                cmp     byte ptr [eax - 1], 13
                je      tbk_again
tbk_endbksp:    call    ted_retr_text
                jmp     ted_control
ted_delete:     call    ted_retr_text           ; ted_retr_text finds out the
                call    ted_mark_cursor         ;  actual ptr of the byte
                cmp     tedcursorig[@cb], -1    ;  pointed by the cursor, and
                je      ted_waitkey             ;  -1 means no char under cr.
tdl_again:      mov     eax, tedcursorig[@cb]
                cmp     eax, endmarkorig[@cb]
                jnb     ted_waitkey
                dec     endmarkorig[@cb]
                mov     ch, [eax]
                mov     ticode[@cb], ch
tdl_backspacer: mov     ch, [eax + 1]
                mov     [eax], ch
                cmp     eax, endmarkorig[@cb]
                jae     tdl_endloop
                inc     eax
                jmp     tdl_backspacer
tdl_endloop:    cmp     ticode[@cb], 13
                jne     tdl_enddel
                mov     eax, tedcursorig[@cb]
                cmp     byte ptr [eax], 10
                je      tdl_again
tdl_enddel:     call    ted_retr_text
                jmp     ted_control
ted_insertcode: call    ted_retr_text           ; ted_retr_text finds out the
                call    ted_mark_cursor         ;  actual ptr of the byte
                cmp     tedcursorig[@cb], -1    ;  pointed by the cursor, and
                je      tic_fail                ;  -1 means no char under cr.
                mov     ecx, [esi + 12]         ; ecx = buffer size
                sub     ecx, endmarkorig[@cb]   ; ecx - ptr to null terminator
                add     ecx, tedbuffer[@cb]     ; ecx + origin of text buffer
                cmp     ecx, 3                  ; (this means at least 3 bytes
                jle     tic_fail                ;  must be available for ins.)
ted_inswocoupd: mov     eax, endmarkorig[@cb]   ; offs of null terminator code
tic_makeroom:   mov     ch, [eax]               ; now rescans the buffer
                mov     [eax + 1], ch           ;  backwards, moving all bytes
                dec     eax                     ;  1 place forward to make
                cmp     eax, tedcursorig[@cb]   ;  room for the inserted code:
                jnb     tic_makeroom            ;  loop ends at "tedcursorig".
                mov     ch, ticode[@cb]         ; finally inserts the code:
                mov     [eax + 1], ch           ; (eax=tedcursorig-1, see jnb)
                inc     endmarkorig[@cb]        ; increments end mark position
                clc                             ;  and the function returns
                ret                             ;  successfully.
tic_fail:       stc                             ; failure = not enough room
                ret                             ;  in text buffer for insert.
ted_retr_text:  pushad
                mov     eax, tedcursor_y[@cb]   ;  (compute tedcursptr as the
                mov     edx, 80                 ;   offset of the cursor cell
                mul     edx                     ;   relatively to the TED's
                add     eax, tedcursor_x[@cb]   ;   upper left corner and
                shl     eax, 1                  ;   aligned to screen's width
                mov     tedcursptr[@cb], eax    ;   and granularity: 80x2)
                mov     eax, tedbuffer[@cb]     ;  (EAX = buffer origin)
                mov     ecx, [esi + 12]         ;  (ECX = buffer size)
                mov     txcol[@cb], 0           ;  (starting text retrace X)
                mov     txrow[@cb], 0           ;  (starting text retrace Y)
                mov     tedcursorig[@cb], -1    ;  (invalidate insert point)
                mov     dh, tedfgattrs[@cb]
                call    ted_shownewline
ted_next_char:  mov     dl, [eax]
                test    dl, dl
                jz      ted_end_text
                cmp     dl, 10
                jne     ted_no_lf
                inc     txrow[@cb]
                call    ted_shownewline
                jmp     ted_ctrl_code
ted_no_lf:      cmp     dl, 13
                jne     ted_no_caret
                and     dh, 0F0h
                or      dh, 002h
                mov     dl, 127
ted_no_caret:   cmp     dl, 32
                jnb     ted_known_code
                and     dh, 0F0h
                or      dh, 004h
                mov     dl, '?'
ted_known_code: call    ted_calc_coords
                pushf
                inc     txcol[@cb]
                popf
                jc      ted_ctrl_code
                call    ted_checkcrorig
                mov     [edi + ebx], dx
ted_ctrl_code:  mov     dh, tedfgattrs[@cb]
                inc     eax
                dec     ecx
                jnz     ted_next_char
                mov     tedcursorig[@cb], -1    ; (no null terminators found!
                and     dh, 0F0h                ;  a bad thing: it could mean
                or      dh, 004h                ;  dirty memory. The mark is
                jmp     ted_end_signal          ;  RED and no insert allowed.)
ted_end_text:   and     dh, 0F0h                ; (null terminator regularly
                or      dh, 002h                ;  found: green mark, insert
                mov     endmarkorig[@cb], eax   ;  is allowed and optimized)
ted_end_signal: call    ted_calc_coords
                jc      ted_endmark_nv
                mov     dl, 3
                test    dh, 004h
                jnz     ted_endmark_red
                call    ted_checkcrorig
ted_endmark_red:mov     [edi + ebx], dx
ted_endmark_nv: mov     ebp, ted_vrows[@cb]
ted_clear_eop:  inc     txrow[@cb]
                push    ebp
                call    ted_shownewline
                pop     ebp
                dec     ebp
                jnz     ted_clear_eop
                popad
                ret
ted_checkcrorig:cmp     ebx, tedcursptr[@cb]
                jne     ted_nocrsorig
                mov     tedcursorig[@cb], eax
ted_nocrsorig:  ret
ted_shownewline:mov     dl, 20h
                mov     txcol[@cb], 0
                push    txstartcol[@cb]
                mov     txstartcol[@cb], 0
                mov     ebp, ted_vcols[@cb]
ted_clear_eol:  call    ted_calc_coords
                jc      ted_found_eol
                mov     [edi + ebx], dx
                inc     txcol[@cb]
                dec     ebp
                jnz     ted_clear_eol
ted_found_eol:  pop     txstartcol[@cb]
                mov     txcol[@cb], 0
                ret
ted_calc_coords:mov     ebx, txrow[@cb]
                sub     ebx, txstartrow[@cb]
                cmp     ebx, ted_vrows[@cb]
                jnb     ted_cc_fail
                mov     ebx, txcol[@cb]
                sub     ebx, txstartcol[@cb]
                cmp     ebx, ted_vcols[@cb]
                jnb     ted_cc_fail
                push    eax
                push    edx
                mov     eax, txrow[@cb]
                sub     eax, txstartrow[@cb]
                mov     edx, 160
                mul     edx
                mov     edx, txcol[@cb]
                sub     edx, txstartcol[@cb]
                add     eax, edx
                add     eax, edx
                mov     ebx, eax
                pop     edx
                pop     eax
                clc
                ret
ted_cc_fail:    stc
                ret
ted_mark_cursor:mov     eax, tedcursor_y[@cb]
                mov     edx, 160
                mul     edx
                mov     ecx, [esi + 16]
ted_hmarkloop:  xor     byte ptr [edi + eax + 1], 88h
                add     eax, 2
                loop    ted_hmarkloop
                mov     eax, tedcursor_x[@cb]
                add     eax, tedcursor_x[@cb]
                mov     ecx, [esi + 20]
ted_vmarkloop:  xor     byte ptr [edi + eax + 1], 88h
                add     eax, 160
                loop    ted_vmarkloop
                ret
wait1tick:      push    ebp
                mov     ebp, ds:[46Ch]
waitloop:       cmp     ebp, ds:[46Ch]
                je      waitloop
                pop     ebp
                ret
move2nearestaa: m2m     sa_prevstart[@cb], sa_start[@cb]
                m2m     sa_prevend[@cb], sa_end[@cb]
                mov     sa_changed[@cb], 0
                m2m     areastocheck[@cb], activeareas[@cb]
                cmp     areastocheck[@cb], 0
                je      aanoselection
                mov     esi, @aa
                mov     ebx, cursorpos_x[@cb]
                mov     ecx, cursorpos_y[@cb]
                m2m     aux?4[@cb], maxaadistance[@cb]
checknextaa:    movzx   edx, byte ptr [esi]
                cmp     edx, focal_left[@cb]
                jle     aatoofar
                cmp     edx, focal_right[@cb]
                jge     aatoofar
                sub     edx, ebx
                mov     aux?1[@cb], edx
                movzx   edx, byte ptr [esi + 1]
                cmp     edx, focal_top[@cb]
                jle     aatoofar
                cmp     edx, focal_bottom[@cb]
                jge     aatoofar
                sub     edx, ecx
                mov     aux?2[@cb], edx
                fild    aux?1[@cb]
                fmul    aaxsensitivity[@cb]
                fmul    st, st
                fild    aux?2[@cb]
                fmul    aaysensitivity[@cb]
                fmul    st, st
                faddp
                fsqrt
                fst     aux?3[@cb]
                fcomp   aux?4[@cb]
                fstsw   ax
                sahf
                jnb     aatoofar
                m2m     aux?4[@cb], aux?3[@cb]
                mov     resultingarea[@cb], esi
                mov     sa_changed[@cb], 1
aatoofar:       add     esi, 6
                dec     areastocheck[@cb]
                jnz     checknextaa
                test    sa_changed[@cb], 1
                jz      aanoselection
                m2m     selectedarea[@cb], resultingarea[@cb]
                mov     esi, resultingarea[@cb]
                movzx   edx, byte ptr [esi]
                mov     cursorpos_x[@cb], edx
                movzx   edx, byte ptr [esi + 1]
                mov     cursorpos_y[@cb], edx
                mov     eax, cursorpos_y[@cb]
                mov     edx, 160
                mul     edx
                add     eax, cursorpos_x[@cb]
                add     eax, cursorpos_x[@cb]
                mov     edi, eax
                add     edi, @tx
                mov     esi, dword ptr [esi + 2]
                mov     edx, [esi]
                shl     edx, 2
                add     esi, edx
                add     esi, 8
                mov     sa_start[@cb], edi
                mov     sa_width[@cb], 0
aahighlight:    cmp     byte ptr [esi], MAX_TAG_SPECS
                ja      aakeepgoing
                cmp     dword ptr [esi+4], offset tag_end[@cb]
                je      aaendarea
                mov     eax, [esi]
                shl     eax, 2
                cmp     dword ptr [esi+4], offset tag_cri[@cb]
                jne     aanotacritter
                not     byte ptr [edi+1]
                add     edi, 2
                inc     sa_width[@cb]
aanotacritter:  add     esi, eax
                add     esi, 7
                jmp     aaskippingtag
aakeepgoing:    cmp     dword ptr [esi], ''
                je      aaendarea
aaskippingtag:  not     byte ptr [edi+1]
                add     edi, 2
                inc     esi
                inc     sa_width[@cb]
                jmp     aahighlight
aaendarea:      mov     sa_end[@cb], edi
                cmp     sa_prevstart[@cb], 0
                je      aanoselection
                mov     edi, sa_prevstart[@cb]
aadeselect:     cmp     edi, sa_prevend[@cb]
                jnb     aanoselection
                not     byte ptr [edi+1]
                add     edi, 2
                jmp     aadeselect
aanoselection:  ret
nexthtrow:      inc     starthtline[@cb]
                jmp     freshscreen
nexthtpag:      add     starthtline[@cb], htl
                jmp     freshscreen
prevhtrow:      cmp     starthtline[@cb], 0
                je      waitcpselect
                dec     starthtline[@cb]
                jmp     freshscreen
prevhtpag:      cmp     starthtline[@cb], 0
                je      waitcpselect
                sub     starthtline[@cb], htl
                jnc     freshscreen
                mov     starthtline[@cb], 0
                jmp     freshscreen
hthome:         cmp     starthtline[@cb], 0
                je      waitcpselect
                mov     starthtline[@cb], 0
                jmp     freshscreen
writestatusline:mov     ecx, som[@cb]
                mov     eax, som[@cb]
                shr     eax, 5
                sub     ecx, eax
                inc     ecx
                mov     eax, tom[@cb]
                shl     eax, 4
                xor     edx, edx
                div     ecx
                cmp     eax, 15
                jbe     umm_inlimit
                mov     eax, 15
umm_inlimit:    lea     edi, cpmemt[@cb]
                mov     dword ptr [edi], 04040404h
                push    eax
umm_nextfrag:   sub     eax, 4
                js      umm_fractions
                mov     byte ptr [edi], 7
                inc     edi
                jmp     umm_nextfrag
umm_fractions:  pop     eax
                and     eax, 3
                add     eax, 4
                mov     [edi], al
                mov     edi, @tx + 24 * 160
                mov     ecx, 80
                mov     ax, 0F420h
                cld
                rep     stosw
                mov     edi, @tx + 24 * 160
                lea     esi, cpsl[@cb]
wrhelplineloop: mov     al, [esi]
                test    al, al
                jz      wrhelplineend
                mov     [edi], al
                add     edi, 2
                inc     esi
                jmp     wrhelplineloop
wrhelplineend:  ret
writecpdiv:     mov     edi, @tx + htl * 160
                mov     ah, currenthtattrs[@cb]
                and     ah, 0F0h
                or      ah, 7
                mov     al, 8
                mov     ecx, 40
                cld
                rep     stosw
                mov     byte ptr [edi + 1], ah
                add     edi, 2
                mov     ecx, 39
                rep     stosw
                ret
writehtcontest: mov     activeareas[@cb], 0
                mov     selectedarea[@cb], 0
                mov     sa_start[@cb], 0
                mov     currenthtattrs[@cb], 70h
                mov     inlinkhtattrs[@cb], 70h
                m2m     currenthtline[@cb], starthtline[@cb]
                mov     currenthtcol[@cb], 0
                mov     edi, @tx + htl * 160 + 160
                mov     esi, @cp
                call    writenextcode
                mov     currenthtattrs[@cb], 1Fh
                mov     inlinkhtattrs[@cb], 0Bh
                mov     currenthtline[@cb], 0
                mov     currenthtcol[@cb], 0
                mov     edi, @tx
                mov     esi, @ht
                call    writenextcode
                test    appflag[@cb], 1
                jnz     sameselection
                m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                call    fullscreenfocal
                call    move2nearestaa
sameselection:  ret
writenextcode:  cmp     dword ptr [esi], ''
                mov     moreflag[@cb], 0
                je      endhtcontest
                mov     moreflag[@cb], 7
                cmp     dword ptr [esi], MAX_TAG_SPECS
                ja      writenormalchar
                mov     tag_visibility[@cb], 0
                mov     eax, starthtline[@cb]
                cmp     currenthtline[@cb], eax
                jb      tagnotvisible
                add     eax, htl
                cmp     currenthtline[@cb], eax
                jnb     endhtcontest
                or      tag_visibility[@cb], 1
tagnotvisible:  add     esi, 8
                call    dword ptr [esi - 4]
                sub     esi, 8
                mov     eax, [esi]
                shl     eax, 2
                add     esi, eax
                add     esi, 7 ; +1 will be performed later so it's adding 8
                jmp     writeaftertag
writenormalchar:cmp     byte ptr [esi], 28
                jb      writeaftertag
                cmp     byte ptr [esi], 127
                jne     dontforcecaret
                mov     currenthtcol[@cb], 0
                mov     eax, currenthtline[@cb]
                inc     currenthtline[@cb]
                cmp     eax, starthtline[@cb]
                jb      writeaftertag
                mov     eax, starthtline[@cb]
                add     eax, htl
                cmp     currenthtline[@cb], eax
                jnb     endhtcontest
                mov     eax, edi
                sub     eax, @tx
                mov     ebx, 160
                xor     edx, edx
                div     ebx
                inc     eax
                xor     edx, edx
                mul     ebx
                add     eax, @tx
                mov     ch, currenthtattrs[@cb]
                mov     cl, 20h
clearendofline: cmp     edi, eax
                jnb     writeaftertag
                mov     [edi], cx
                add     edi, 2
                jmp     clearendofline
dontforcecaret: mov     eax, starthtline[@cb]
                cmp     currenthtline[@cb], eax
                jb      forwardcursor
                add     eax, htl
                cmp     currenthtline[@cb], eax
                jnb     endhtcontest
                mov     al, [esi]
                mov     ah, currenthtattrs[@cb]
                mov     [edi], ax
                add     edi, 2
forwardcursor:  inc     currenthtcol[@cb]
                cmp     currenthtcol[@cb], 80
                jb      writeaftertag
                mov     currenthtcol[@cb], 0
                inc     currenthtline[@cb]
writeaftertag:  inc     esi
                jmp     writenextcode
endhtcontest:   cmp     edi, @tx + htl * 160
                jnb     realendhtcont
                mov     ah, currenthtattrs[@cb]
                mov     al, 20h
                mov     [edi], ax
                add     edi, 2
                jmp     endhtcontest
realendhtcont:  ret
tag_void:       ret
tag_end:        mov     al, prelinkhtattrs[@cb]
                mov     currenthtattrs[@cb], al
                test    tedtagopen[@cb], 1
                mov     tedtagopen[@cb], 0
                jz      tag_end_no_ted
                test    tag_visibility[@cb], 1
                jz      tag_end_no_ted
                mov     ah, currenthtattrs[@cb]
                mov     al, 3
                mov     [edi], ax
                add     edi, 2
                inc     currenthtcol[@cb]
                cmp     currenthtcol[@cb], 80
                jb      tag_end_no_ted
                mov     currenthtcol[@cb], 0
                inc     currenthtline[@cb]
tag_end_no_ted: ret
tag_wht:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 007h
                ret
tag_blk:        and     currenthtattrs[@cb], 0F8h
                ret
tag_red:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 004h
                ret
tag_grn:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 002h
                ret
tag_blu:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 001h
                ret
tag_yel:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 006h
                ret
tag_mag:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 005h
                ret
tag_cya:        and     currenthtattrs[@cb], 0F8h
                or      currenthtattrs[@cb], 003h
                ret
tag_lgt:        or      currenthtattrs[@cb], 008h
                ret
tag_drk:        and     currenthtattrs[@cb], not 008h
                ret
tag_pap:        mov     al, currenthtattrs[@cb]
                and     currenthtattrs[@cb], 0Fh
                and     al, 0Fh
                shl     al, 4
                or      currenthtattrs[@cb], al
                ret
tag_inv:        mov     al, currenthtattrs[@cb]
                mov     ah, currenthtattrs[@cb]
                and     al, 00Fh
                and     ah, 0F0h
                shl     al, 4
                shr     ah, 4
                or      al, ah
                mov     currenthtattrs[@cb], al
                ret
tag_col:        test    tag_visibility[@cb], 1
                jz      tag_col_end
                mov     eax, edi
                sub     eax, @tx
                mov     ebx, 160
                xor     edx, edx
                div     ebx
                xor     edx, edx
                mul     ebx
                add     eax, @tx
                add     eax, [esi]
                add     eax, [esi]
                cmp     edi, eax
                jb      hastoclear
                mov     edi, eax
                ret
hastoclear:     mov     ch, currenthtattrs[@cb]
                mov     cl, 20h
cleartabstop:   mov     [edi], cx
                add     edi, 2
                cmp     edi, eax
                jb      cleartabstop
tag_col_end:    m2m     currenthtcol[@cb], <dword ptr [esi]>
                ret
tag_cri:        test    tag_visibility[@cb], 1
                jz      tag_cri_end
                mov     ah, currenthtattrs[@cb]
                mov     al, [esi]
                shl     al, 1
                add     al, 4
                mov     [edi], ax
                add     edi, 2
                inc     currenthtcol[@cb]
                cmp     currenthtcol[@cb], 80
                jb      tag_cri_noline
                mov     currenthtcol[@cb], 0
                inc     currenthtline[@cb]
tag_cri_noline: inc     al
                mov     [edi], ax
                add     edi, 2
                inc     currenthtcol[@cb]
                cmp     currenthtcol[@cb], 80
                jb      tag_cri_end
                mov     currenthtcol[@cb], 0
                inc     currenthtline[@cb]
tag_cri_end:    ret
tag_ted:        mov     al, currenthtattrs[@cb]
                mov     prelinkhtattrs[@cb], al
                test    tag_visibility[@cb], 1
                jz      top_of_aa
                mov     tedtagopen[@cb], 1
                mov     ah, currenthtattrs[@cb]
                mov     al, 2
                mov     [edi], ax
                add     edi, 2
                inc     currenthtcol[@cb]
                cmp     currenthtcol[@cb], 80
                jb      tag_lnk
                mov     currenthtcol[@cb], 0
                inc     currenthtline[@cb]
tag_lnk:        nop     ; serves to distinguish <lnk> from <exe> upon running
tag_exe:        mov     al, currenthtattrs[@cb]
                mov     prelinkhtattrs[@cb], al
                test    tag_visibility[@cb], 1
                jz      top_of_aa
                cmp     activeareas[@cb], @aarang
                jnb     top_of_aa
                mov     al, inlinkhtattrs[@cb]
                mov     currenthtattrs[@cb], al
                mov     eax, edi
                sub     eax, @tx
                shr     eax, 1
                mov     ebx, 80
                xor     edx, edx
                div     ebx
                mov     ebx, activeareas[@cb]
                shl     ebx, 2
                add     ebx, activeareas[@cb]
                add     ebx, activeareas[@cb]
                mov     byte ptr ds:[@aa + ebx], dl
                mov     byte ptr ds:[@aa + ebx + 1], al
                sub     esi, 8
                m2m     <dword ptr ds:[@aa + ebx + 2]>, esi
                add     esi, 8
                inc     activeareas[@cb]
top_of_aa:      ret
tag_foc:        test    tag_visibility[@cb], 1
                jz      tag_foc_notv
                m2m     aaxsensitivity[@cb], defxsensitivity[@cb]
                m2m     aaysensitivity[@cb], defysensitivity[@cb]
                m2m     cursorpos_x[@cb], <dword ptr [esi]>
                m2m     cursorpos_y[@cb], <dword ptr [esi + 4]>
                pushad
                call    fullscreenfocal
                call    move2nearestaa
                popad
tag_foc_notv:   ret
; control panel's inline data 
hdbchecksum     dd      ?
logoffset       db      0
cpstatsp        db      ?
endstatus       db      ?
quietmode       db      0
historyptr      dd      0
currenthtattrs  db      ?
currenthtline   dd      ?
currenthtcol    dd      ?
starthtline     dd      ?
tag_visibility  db      ?
prelinkhtattrs  db      ?
inlinkhtattrs   db      ?
tedbuffer       dd      ?
ted_vcols       dd      ?
ted_vrows       dd      ?
txstartcol      dd      ?
txstartrow      dd      ?
txcol           dd      ?
txrow           dd      ?
tedbgattrs      db      0F0h
tedfgattrs      db      0F8h
tedsheet_x      dd      ?
tedsheet_y      dd      ?
tedcursor_x     dd      ?
tedcursor_y     dd      ?
tedcursptr      dd      -1
tedcursorig     dd      -1
endmarkorig     dd      ?
ticode          db      ?
moreflag        db      0
activeareas     dd      0
tedtagopen      db      ?
selectedarea    dd      0
resultingarea   dd      0
maxaadistance   dd      2.5e6
aaxsensitivity  dd      1.0
aaysensitivity  dd      2.0
defxsensitivity dd      1.0
defysensitivity dd      2.0
lowxsensitivity dd      4.0
lowysensitivity dd      8.0
cursorpos_x     dd      -1
cursorpos_y     dd      -1
focal_left      dd      ?
focal_top       dd      ?
focal_right     dd      ?
focal_bottom    dd      ?
areastocheck    dd      ?
sa_hlgtiming    dd      1 * 2730
sa_width        dd      ?
sa_start        dd      0
sa_end          dd      ?
sa_prevstart    dd      ?
sa_prevend      dd      ?
sa_changed      db      ?
execodeptr      dd      ?
cpsl            db      'Hactar 2001/A', 32, 09         ; (status line)
cpmemt          db      32, 32, 32, 32, 09              ; (memory meter)
cpdmon          db      32, 32, 32, 32, 09              ; (disk activity mon.)
cpstat          db      02, 22 dup (32), 09             ; (fn. status log)
cptedl          db      32, 32, 32, 32, 32, 32          ; (TED line, 5+colon)
cptedc          db      32, 32, 32, 09                  ; (TED column)
cpyear          db      32, 32, 32, 32, '/'             ; (current year)
cpmonth         db      32, 32, '/'                     ; (current month)
cpday           db      32, 32, 09                      ; (current day)
cphour          db      32, 32, ':'                     ; (current hour)
cpminutes       db      32, 32, ':'                     ; (current minutes)
cpseconds       db      32, 32, '.'                     ; (current seconds)
cptenths        db      32, 0                           ; (tenths of second)
ht_commandlist  db      '<lnkSystemScreen,#0><cri#6>System Info&end'
                db      '<col#16><exeTimerHook,#0><cri#5>System Timer&end'
                db      '<col#32><exeMemoryProbe><cri#7>Memory&end'
                db      '<col#48><exeDiskProbe><cri#8>Storage&end'
                db      '<col#64><exeSettings><cri#4>Settings&end\'
                db      '<exePrompt><cri#9>Prompt&end'
                db      '<col#16><exeResetHook,#0><cri#11>System Reset&end'
                db      '<col#32><exeResetHook,#1><cri#11>S/ware Reset&end'
                db      '<col#48><exeResetHook,#2><cri#11>H/ware Reset&end'
                db      '<col#64><exeResetHook,#3><cri#11>Power OFF&end'
                db      '<col#79>', 32, 0
ResidentGAT     dd      offset SystemScreen[@cb]
                dd      offset ResetHook[@cb]
                dd      offset MemoryProbe[@cb]
                dd      offset MProbeIO[@cb]
                dd      offset MProbeED[@cb]
                dd      offset MProbeOP[@cb]
                dd      offset MProbePO[@cb]
                dd      offset MProbeDO[@cb]
                dd      offset ManLock[@cb]
                dd      offset ManLoSZ[@cb]
                dd      offset ManUnlock[@cb]
                dd      offset DiskProbe[@cb]
                dd      offset MountID[@cb]
                dd      offset MountDev[@cb]
                dd      offset ProbeIO[@cb]
                dd      offset ProbeED[@cb]
                dd      offset ProbeID[@cb]
                dd      offset ProbePO[@cb]
                dd      offset ProbeOP[@cb]
                dd      offset ProbeDO[@cb]
                dd      offset FormtID[@cb]
                dd      offset FormtLF[@cb]
                dd      offset FormtVL[@cb]
                dd      offset FormtD1[@cb]
                dd      offset FormtD2[@cb]
                dd      offset GATLoad[@cb]
                dd      offset GATUnload[@cb]
                dd      offset GATLookup[@cb]
                dd      offset ReSyncTTCounter[@cb]
                dd      offset DateTime2TT[@cb]
                dd      offset TT2DateTime[@cb]
                dd      offset WrHyper[@cb]
                dd      offset WrBack[@cb]
                dd      offset SystemReset[@cb]
                dd      offset MountMemory[@cb]
                dd      offset LockMemory[@cb]
                dd      offset UnlockMemory[@cb]
                dd      offset MountVolume[@cb]
                dd      offset VolumeIO[@cb]
ResidentGATsz   equ     ($ - ResidentGAT) / 4
; System32 flat-mode interrupt handlers 
fmtthandler:    pushad                                  ; FM timertick handler
                add     dword ptr ds:[@tts + 0], 2730   ; (increment tts/LSD)
                adc     dword ptr ds:[@tts + 4], 0      ; (increment tts/MSD)
                call    dword ptr ds:[@fmuth + 1 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 2 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 3 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 4 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 5 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 6 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 7 * 4]   ; (USER TICK HANDLER)
                call    dword ptr ds:[@fmuth + 8 * 4]   ; (USER TICK HANDLER)
                mov     ebx, ds:[@rmivt + IRQ0*4]       ; * save current ptr.
                mov     eax, ds:[@hdb + IRQ0*4]         ; (get original ptr)
                mov     ds:[@rmivt + IRQ0*4], eax       ; (the OLD vector 8)
                pushfd                                  ; simulate interrupt
                db      09Ah                            ; call far (16:32)
                dd      @f2r_it + IRQ0*4                ; (to INTCON table)
                dw      ICODE                           ; (within MM code)
                mov     ds:[@rmivt + IRQ0*4], ebx       ; *restore int 8 ptr.
                popad                                   ; (EOI being sent by
                iretd                                   ;  original handler)
dummy_fmuth:    ret                                     ; dummyusertickhandler
; control panel's personal "user tick handler" 
cp_uth:         mov     eax, ds:[@tts + 4]
                mov     ebx, ds:[@tts + 0]
                call    TT2DateTime
                mov     eax, rtc_year[@cb]
                lea     ebx, cpyear[@cb]
                mov     ecx, 4
                call    In2Dec
                mov     eax, rtc_month[@cb]
                lea     ebx, cpmonth[@cb]
                mov     ecx, 2
                call    In2Dec
                mov     eax, rtc_day[@cb]
                lea     ebx, cpday[@cb]
                mov     ecx, 2
                call    In2Dec
                mov     eax, rtc_hour[@cb]
                lea     ebx, cphour[@cb]
                mov     ecx, 2
                call    In2Dec
                mov     eax, rtc_minutes[@cb]
                lea     ebx, cpminutes[@cb]
                mov     ecx, 2
                call    In2Dec
                mov     eax, rtc_seconds[@cb]
                lea     ebx, cpseconds[@cb]
                mov     ecx, 2
                call    In2Dec
                mov     eax, rtc_tenths[@cb]
                lea     ebx, cptenths[@cb]
                mov     ecx, 1
                call    In2Dec
                call    writestatusline
                ret
; control panel's resident procedures 
                db      'SystemScreen', 12
SystemScreen    db      '\  '
                db      'HSP Hactar (Lasting Child)'
                db      ' - '
                db      'Version A'
                db      ' - '
                db      'Release date 2001.10.07', 0
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
ResetScreen0    db      '\  System environment reset:'
                db      '\  &invDO NOT REMOVE BOOT DISK!&inv\\', 0
ResetScreen1    db      '\  Warm/Software reset:\\', 0
ResetScreen2    db      '\  Cold/Hardware reset:\\', 0
ResetScreen3    db      '\  &invPower OFF:&inv'
                db      '\  REMOVE MEDIA FROM DRIVE(S) BEFORE YOU CONFIRM!\\', 0
ResetMessage    db      '    You will lose any unsaved data.\'
                db      '    Are you sure you want to proceed?\\'
                db      '<col#50><exeWrBack>NO&end', 32, 32, 0
ResetLink0      db      '<exeSystemReset,#0>YES&end', 0
ResetLink1      db      '<exeSystemReset,#1>YES&end', 0
ResetLink2      db      '<exeSystemReset,#2>YES&end', 0
ResetLink3      db      '<exeSystemReset,#3>YES&end', 0
                db      'ResetHook', 9
ResetHook:      cmp     eax, 3
                lea     ebx, ResetScreen3[@cb]
                lea     ecx, ResetLink3[@cb]
                je      rh_showmsg
                cmp     eax, 2
                lea     ebx, ResetScreen2[@cb]
                lea     ecx, ResetLink2[@cb]
                je      rh_showmsg
                cmp     eax, 1
                lea     ebx, ResetScreen1[@cb]
                lea     ecx, ResetLink1[@cb]
                je      rh_showmsg
                cmp     eax, 0
                lea     ebx, ResetScreen0[@cb]
                lea     ecx, ResetLink0[@cb]
                je      rh_showmsg
                jmp     rh_outofrange
rh_showmsg:     push    ecx
                mov     eax, ebx
                mov     ebx, 1
                call    WrHyper
                lea     eax, ResetMessage[@cb]
                mov     ebx, 2
                call    WrHyper
                pop     eax
                mov     ebx, 2
                call    WrHyper
rh_outofrange:  ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
mp_manlocks     dd      0
mp_blockno      dd      ?
mp_blocktop     dd      ?
MemoryProbe0    db      '\ Central memory mappings:\\'
                db      '  contiguous RAM: '
MemoryProbe1    db      'xxxxxxxxxx\'
                db      '  reserved:       '
MemoryProbe2    db      'xxxxxxxxxx\'
                db      '  locked:         '
MemoryProbe3    db      'xxxxxxxxxx\'
                db      '  available:      '
MemoryProbe4    db      'xxxxxxxxxx\'
                db      '  locked blocks:  '
MemoryProbe5    db      'xxxx\', 0
MemoryProbe6    db      '\  START    END      OWNER', 0
MemoryProbe7    db      '\  '
MemoryProbe8    db      'xxxxxxxx '
MemoryProbe9    db      'xxxxxxxx '
MemoryProbe10   db      'xxxxxxxxxxxxxxxxxxxx', 0
MemoryProbe11   db      '\ Probe memory:'
                db      ' <tedMProbePO,#11,#9,#1,#9,#1,,>FIRSTBYTE&end'
                db      '  <tedMProbeOP,#7,#6,#1,#5,#1,,>OPCODE&end'
                db      '  <tedMProbeED,#147,#16,#8,#16,#8,,>EDIT&end'
                db      '  <exeMProbeDO>EXEC&end'
                db      '\\ Manual memory lock:'
                db      '\\ <tedManLoSZ,#13,#11,#1,#11,#1,,>BLOCK SIZE&end (decimal bytes)'
                db      '\  <exeManLock>LOCK BLOCK&end  <exeManUnlock>UNLOCK LAST&end\', 0
                db      '(manually locked)', 17 ; 
ManLockOwner    label   byte                    ; 
                db      'ManLoSZ', 7            ; 
ManLoSZ         db      13 dup (0)              ; 
                db      'ManLock', 7            ; 
ManLock:        lea     eax, ManLoSZ[@cb]       ; 
                call    Dec2In
                lea     ebx, ManLockOwner[@cb]
                call    LockMemory
                jc      MemoryProbe
                inc     mp_manlocks[@cb]
                jmp     MemoryProbe
                db      'ManUnlock', 9          ; 
ManUnlock:      cmp     mp_manlocks[@cb], 0     ; 
                je      MemoryProbe
                call    UnlockMemory
                jc      MemoryProbe
                dec     mp_manlocks[@cb]
                jmp     MemoryProbe
                db      'MProbeIO', 8           ; 
MProbeIO        db      64 dup (0)              ; 
                db      'MProbeED', 8           ; Ŀedit buffer:
MProbeED        db      8 dup ('0000000000000000', 13, 10) ; len 147 bytes,
                db      0, 0, 0                 ; ٳ16 col/16 rows
                db      'MProbePO', 8           ; 
MProbePO        db      11 dup (0)              ; 
                db      'MProbeOP', 8           ; 
MProbeOP        db      7 dup (0)               ; 
                db      'MProbeDO', 8           ; 
MProbeDO:       lea     esi, MProbeED[@cb]      ; 
                lea     edi, MProbeIO[@cb]
                call    ProbeTRforward
                jc      mem_Ewrong
                lea     eax, MProbePO[@cb]
                call    Hex2In
                mov     esi, eax
                lea     edi, MProbeIO[@cb]
                mov     ecx, 64
                strup   <MProbeOP[@cb]>
                cmp     dword ptr MProbeOP[@cb], 'DAER'
                je      mem_copy_in
                cmp     dword ptr MProbeOP[@cb], 'TIRW'
                je      mem_copy_out
                jmp     MemoryProbe
mem_copy_out:   xchg    esi, edi
mem_copy_in:    cld
                rep     movsb
mem_Ewrong:     lea     esi, MProbeIO[@cb]
                lea     edi, MProbeED[@cb]
                call    ProbeTRback
                jmp     MemoryProbe
ProbeTRforward: mov     ecx, 64
ProbeTRforwLP:  mov     ax, [esi]
                cmp     al, 'A'
                jb      ProbeTRFalNUM
                sub     al, 'A'
                cmp     al, 5
                ja      ProbeTRforwERR
                add     al, 10
                jmp     ProbeTRFalACC
ProbeTRFalNUM:  sub     al, '0'
                cmp     al, 9
                ja      ProbeTRforwERR
ProbeTRFalACC:  cmp     ah, 'A'
                jb      ProbeTRFahNUM
                sub     ah, 'A'
                cmp     ah, 5
                ja      ProbeTRforwERR
                add     ah, 10
                jmp     ProbeTRFahACC
ProbeTRFahNUM:  sub     ah, '0'
                cmp     ah, 9
                ja      ProbeTRforwERR
ProbeTRFahACC:  shl     al, 4
                or      al, ah
                mov     [edi], al
                inc     edi
                add     esi, 2
                dec     ecx
                jz      ProbeTRforwEND
                test    ecx, 7
                jnz     ProbeTRforwLP
                add     esi, 2
                jmp     ProbeTRforwLP
ProbeTRforwEND: clc
                ret
ProbeTRforwERR: stc
                ret
ProbeTRback:    mov     ecx, 64
ProbeTRbackLP:  pushad
                movzx   eax, byte ptr [esi]
                mov     ebx, edi
                mov     ecx, 2
                call    In2Hex
                popad
                inc     esi
                add     edi, 2
                dec     ecx
                jz      ProbeTRbackEND
                test    ecx, 7
                jnz     ProbeTRbackLP
                mov     word ptr [edi], 0A0Dh
                add     edi, 2
                jmp     ProbeTRbackLP
ProbeTRbackEND: ret
                db      'MemoryProbe', 11       ; 
MemoryProbe:    xor     eax, eax                ; 
                xor     ebx, ebx
                call    LockMemory
                pushad
                call    UnlockMemory
                popad
                push    ecx
                mov     eax, ebx
                lea     ebx, MemoryProbe4[@cb]
                mov     ecx, 10
                call    In2Dec
                pop     eax
                lea     ebx, MemoryProbe3[@cb]
                mov     ecx, 10
                call    In2Dec
                mov     eax, eom[@cb]
                sub     eax, som[@cb]
                lea     ebx, MemoryProbe2[@cb]
                mov     ecx, 10
                call    In2Dec
                mov     eax, eom[@cb]
                lea     ebx, MemoryProbe1[@cb]
                mov     ecx, 10
                call    In2Dec
                mov     dword ptr MemoryProbe5[@cb], 'enon'
                mov     eax, mt_locks[@cb]
                test    eax, eax
                jz      mp_nolockings1
                lea     ebx, MemoryProbe5[@cb]
                mov     ecx, 4
                call    In2Dec
mp_nolockings1: lea     eax, MemoryProbe0[@cb]
                mov     ebx, 1
                call    WrHyper
                mov     eax, mp_manlocks[@cb]
                cmp     eax, mt_locks[@cb]
                jb      mp_mlforbid
                lea     eax, MemoryProbe11[@cb]
                mov     ebx, 2
                call    WrHyper
mp_mlforbid:    cmp     mt_locks[@cb], 0
                je      mp_nolockings2
                lea     eax, MemoryProbe6[@cb]
                mov     ebx, 2
                call    WrHyper
                mov     eax, mt_locks[@cb]
                mov     mp_blockno[@cb], eax
                m2m     mp_blocktop[@cb], tom[@cb]
mp_blockdisp:   dec     mp_blockno[@cb]
                js      mp_nolockings2
                mov     eax, mp_blockno[@cb]
                mov     eax, dword ptr [mt_iptr + eax*4]
                add     eax, bom[@cb]
                lea     ebx, MemoryProbe8[@cb]
                mov     ecx, 8
                call    In2Hex
                mov     eax, mp_blocktop[@cb]
                add     eax, bom[@cb]
                dec     eax
                lea     ebx, MemoryProbe9[@cb]
                mov     ecx, 8
                call    In2Hex
                mov     eax, mp_blockno[@cb]
                mov     eax, dword ptr [mt_owner + eax*4]
                movzx   ebx, byte ptr [eax - 1]
                sub     eax, ebx
                dec     eax
                xor     ecx, ecx
mp_copyowner:   mov     dl, [eax]
                mov     MemoryProbe10[@cb + ecx], dl
                inc     eax
                inc     ecx
                dec     ebx
                jnz     mp_copyowner
mp_clearowner:  cmp     ecx, 20
                jnb     mp_clearend
                mov     byte ptr MemoryProbe10[@cb + ecx], 32
                inc     ecx
                jmp     mp_clearowner
mp_clearend:    mov     eax, mp_blockno[@cb]
                shr     eax, 5
                mov     ecx, mp_blockno[@cb]
                and     ecx, 31
                mov     ebx, 1
                shl     ebx, cl
                mov     eax, mp_blockno[@cb]
                mov     eax, dword ptr [mt_iptr + eax*4]
                mov     mp_blocktop[@cb], eax
                lea     eax, MemoryProbe7[@cb]
                mov     ebx, 2
                call    WrHyper
                jmp     mp_blockdisp
mp_nolockings2: ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
dp_currentid    dd      ?
formtconfirm    db      0
DiskProbe0      db      '\ System drives:\'
                db      '\  ID    MOUNTED  BUS   SYST  LOCK  CAPACITY(Kb)  NAME\', 0
DiskProbe1      db      '\  '
DiskProbe2      db      'xxxx  '
DiskProbe3      db      'xxx      '
DiskProbe4      db      'xxxx  '
DiskProbe5      db      'xxx   '
DiskProbe6      db      'xxx   '
DiskProbe7      db      'xxxxxxxxxx    '
DiskProbe8      db      'xxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 0
DiskProbe9      db      '\\ Mount device:'
                db      ' <tedMountID,#7,#5,#1,#5,#1,,>ID&end'
                db      '   <exeMountDev>MOUNT NOW&end'
                db      '\\ Probe device:'
                db      ' <tedProbeID,#7,#5,#1,#5,#1,,>ID&end'
                db      '  <tedProbePO,#11,#9,#1,#9,#1,,>FIRSTBYTE&end'
                db      '  <tedProbeOP,#7,#6,#1,#5,#1,,>OPCODE&end'
                db      '  <tedProbeED,#147,#16,#8,#16,#8,,>EDIT&end'
                db      '  <exeProbeDO>EXEC&end'
                db      '\\ Format/configure volume:'
                db      ' <tedFormtID,#7,#5,#1,#5,#1,,>ID&end'
                db      '  <tedFormtLF,#11,#11,#1,#9,#1,,>LARGESTFILE&end'
                db      '  <tedFormtVL,#67,#15,#1,#64,#1,,>VOLUMELABEL&end'
                db      '  <exeFormtD1>EXEC&end'
                db      '  <exeFormtD2>CONFIRM&end', 0
                db      'MountID', 7            ; 
MountID         db      7 dup (0)               ; 
                db      'MountDev', 8           ; 
MountDev:       _idup   <MountID[@cb]>          ; 
                mov     eax, dword ptr MountID[@cb]
                bswap   eax
                call    MountVolume
                jmp     DiskProbe
                db      'ProbeIO', 7            ; 
ProbeIO         db      64 dup (0)              ; 
                db      'ProbeED', 7            ; Ŀedit buffer:
ProbeED         db      8 dup ('0000000000000000', 13, 10) ; len 147 bytes,
                db      0, 0, 0                 ; ٳ16 col/16 rows
                db      'ProbeID', 7            ; 
ProbeID         db      7 dup (0)               ; 
                db      'ProbePO', 7            ; 
ProbePO         db      11 dup (0)              ; 
                db      'ProbeOP', 7            ; 
ProbeOP         db      7 dup (0)               ; 
                db      'ProbeDO', 7            ; 
ProbeDO:        lea     esi, ProbeED[@cb]       ; 
                lea     edi, ProbeIO[@cb]
                call    ProbeTRforward
                jc      disk_Ewrong
                lea     eax, ProbePO[@cb]
                call    Hex2In
                mov     ebx, eax
                xor     eax, eax
                mov     ecx, 64
                _idup   <ProbeID[@cb]>
                mov     edx, dword ptr ProbeID[@cb] 
                bswap   edx
                strup   <ProbeOP[@cb]>
                mov     esi, dword ptr ProbeOP[@cb]
                bswap   esi
                lea     edi, ProbeIO[@cb]
                call    VolumeIO
disk_Ewrong:    lea     esi, ProbeIO[@cb]
                lea     edi, ProbeED[@cb]
                call    ProbeTRback
                jmp     DiskProbe
                db      'FormtID', 7            ; 
FormtID         db      7 dup (0)               ; 
                db      'FormtLF', 7            ; 
FormtLF         db      11 dup (0)              ; 
                db      'FormtVL', 7            ; 
FormtVL         db      67 dup (0)              ; 
                db      'FormtD1', 7            ; 
FormtD1:        mov     formtconfirm[@cb], 1    ; 
                ret
                db      'FormtD2', 7            ; 
FormtD2:        _idup   <FormtID[@cb]>          ; 
                test    formtconfirm[@cb], 1
                jnz     dp_goformat
                ret
dp_goformat:    lea     eax, FormtLF[@cb]
                call    Dec2In
                mov     ecx, eax
                xor     ebx, ebx
                lea     edx, FormtVL[@cb]
                mov     eax, dword ptr FormtID[@cb]
                bswap   eax
                call    FormatVolume
                jmp     DiskProbe
                db      'DiskProbe', 9          ; 
DiskProbe:      mov     formtconfirm[@cb], 0    ; 
                lea     eax, DiskProbe0[@cb]
                mov     ebx, 1
                call    WrHyper
                mov     dp_currentid[@cb], 'RD01'
dp_nextdrive:   mov     eax, dp_currentid[@cb]
                bswap   eax
                mov     dword ptr DiskProbe2[@cb], eax
                mov     eax, dp_currentid[@cb]
                call    internal_TranslateDriveID
                mov     dword ptr DiskProbe3[@cb], ' SEY'
                cmp     byte ptr DIBr[@cb + edi + 00], -1
                jne     dp_ismounted
                mov     dword ptr DiskProbe3[@cb], '  ON'
                mov     byte ptr DiskProbe3[@cb + 4], 0
                jmp     dp_isnotmounted
dp_ismounted:   call    sub_extend_DIB
                mov     byte ptr DiskProbe3[@cb + 4], 32
                mov     dword ptr DiskProbe4[@cb], 'ASIL'
                test    byte ptr DIBr[@cb + edi + 15], df_enhanced
                jz      dp_islisa
                mov     dword ptr DiskProbe4[@cb], 'CHNE'
dp_islisa:      mov     dword ptr DiskProbe5[@cb], '  ON'
                test    byte ptr DIBr[@cb + edi + 15], df_systemdisk
                jz      dp_isnotfmtd
                mov     dword ptr DiskProbe5[@cb], ' SEY'
dp_isnotfmtd:   mov     dword ptr DiskProbe6[@cb], ' SEY'
                test    byte ptr DIBr[@cb + edi + 15], df_write_enable
                jz      dp_islocked
                mov     dword ptr DiskProbe6[@cb], '  ON'
dp_islocked:    mov     eax, mvl_dev_size_m[@cb]
                mov     ebx, mvl_dev_size_l[@cb]
                shrd    ebx, eax, 10
                mov     eax, ebx
                lea     ebx, DiskProbe7[@cb]
                mov     ecx, 10
                call    In2Dec
                mov     byte ptr DiskProbe8[@cb], '?'
                mov     byte ptr DiskProbe8[@cb+1], 0
                test    byte ptr DIBr[@cb + edi + 15], df_systemdisk
                jz      dp_isnotmounted ; (at this point means "has no label")
                xor     eax, eax
                mov     ebx, 28h
                mov     ecx, 29
                mov     edx, dp_currentid[@cb]
                mov     esi, 'READ'
                lea     edi, DiskProbe8[@cb]
                mov     quietmode[@cb], 1
                call    VolumeIO
                mov     DiskProbe8[@cb+29], 0
dp_isnotmounted:lea     eax, DiskProbe1[@cb]
                mov     ebx, 2
                call    WrHyper
                inc     dp_currentid[@cb]
                cmp     dp_currentid[@cb], 'RD05'
                je      dp_endlist
                cmp     dp_currentid[@cb], 'MD05'
                je      dp_endlist
                jmp     dp_nextdrive
dp_endlist:     cmp     dp_currentid[@cb], 'MD05'
                je      dp_endatall
                mov     dp_currentid[@cb], 'MD01'
                jmp     dp_nextdrive
dp_endatall:    lea     eax, DiskProbe9[@cb]
                mov     ebx, 2
                call    WrHyper
                ret
; resident font and UDG table 
@FM             db      64 dup (0)
@00             db      00000010b,00000001b,00000001b,01000000b
blank_more      db      00000000b,00000000b,00000000b,00000000b
textarea_start  db      00001000b,00000001b,00000001b,01000000b
textarea_marker db      00010000b,00000000b,00000000b,00000000b
textarea_end    db      00000000b,00000001b,00000001b,01000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      11111111b,00000001b,00000001b,01000000b
                db      01111110b,00000000b,00000000b,00000000b
                db      00111100b,00000001b,00000001b,01000000b
                db      00011000b,00000000b,00100000b,00000000b
                db      00000000b,00000001b,00010001b,01000000b
                db      00000000b,00000000b,00001000b,00000000b
                db      00010000b,00000001b,00000101b,01000000b
                db      00110001b,00000011b,00001000b,01100000b
                db      01110011b,00000111b,00010001b,01110000b
                db      11110111b,00001111b,00100000b,01111000b
@04             db      00000000b,00000000b,00000000b,00000000b
meter_udgs      db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      11011011b,11011011b,11011011b,11011011b
                db      11011011b,11011011b,11011011b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,11000000b,11011000b,11011011b
                db      00000000b,00000000b,00000000b,00000000b
@08             db      00000010b,00001000b,00000000b,00000000b
horiz_gradient  db      00000000b,00000000b,00000000b,00000000b
divider         db      00001000b,00001000b,00000000b,00000000b
ctrlkey_udg     db      00010000b,00000000b,00000000b,00000000b
                db      00000001b,00001000b,01100010b,00000100b
                db      01000000b,00000000b,10010010b,00000100b
                db      10000100b,00001000b,10000111b,11110100b
                db      00001000b,00001000b,10000010b,10000100b
                db      00000001b,00001000b,10000010b,10000100b
                db      00100011b,00001000b,10010010b,10000100b
                db      01100110b,00000000b,01100011b,10000101b
                db      11000100b,00001000b,00000000b,00000000b
                db      10010000b,00000000b,00000000b,00000000b
                db      00110001b,00001000b,00000000b,00000000b
                db      01110011b,00000000b,00000000b,00000000b
                db      11110111b,00001000b,00000000b,00000000b
@12             db      00000000b,00000000b,00000000b,00000000b
settings_udg    db      00010010b,00100000b,00000000b,10000000b
clock_udg       db      00110011b,00100000b,00001000b,00001000b
                db      00110011b,00100000b,00000100b,00000000b
                db      00011110b,00100000b,00100010b,00000010b
                db      00001100b,00101110b,00000001b,00000000b
                db      00001100b,00100010b,00000000b,00000000b
                db      00001100b,00100010b,01001111b,11000001b
                db      00001100b,00100010b,00000000b,00110000b
                db      00001100b,01110010b,00000000b,00001100b
                db      00001100b,01110010b,00100000b,00000010b
                db      00001100b,01110010b,00000000b,00000000b
                db      00011110b,01110010b,00001000b,00001000b
                db      00110011b,01110010b,00000000b,10000000b
                db      00110011b,01110010b,00000000b,00000000b
                db      00010010b,01110010b,00000000b,00000000b
@16             db      00000000b,00000000b,00000000b,00000000b
system_udg      db      01111110b,00000000b,00010101b,10101000b
memory_udg      db      01000011b,00000000b,00010101b,10101000b
                db      01000011b,11111110b,00000000b,00000110b
                db      01111111b,10000011b,01101100b,00000000b
                db      01000011b,10000011b,00001100b,00000110b
                db      01111111b,10000011b,01100000b,00000000b
                db      01000011b,11111111b,00000000b,00000110b
                db      01000011b,11111111b,01100000b,00000000b
                db      01000000b,00000000b,00001110b,10100110b
                db      00011011b,11011011b,01100000b,00000000b
                db      00000000b,00000000b,00001011b,00000110b
                db      00110111b,10101010b,01100000b,00000000b
                db      00000000b,00000000b,00010101b,10101000b
                db      01101111b,10101100b,00010101b,10101000b
                db      00000000b,00000000b,00000000b,00000000b
@20             db      00000000b,00000000b,00000000b,00000000b
storage_udg     db      01010101b,10101010b,00000000b,00000000b
logo_udg        db      00000000b,00000001b,00011100b,00000000b
                db      10001111b,01110011b,00001110b,10000000b
                db      00001111b,01110001b,00000111b,10000100b
                db      10001111b,01110011b,00000011b,10001000b
                db      00000000b,00000001b,00000001b,11000000b
                db      10010101b,10101011b,00000000b,11100000b
                db      00000000b,00000001b,11111111b,01110111b
                db      10010000b,00000011b,00000000b,11100000b
                db      00000000b,00000001b,00000001b,11000000b
                db      10010000b,00000011b,00000011b,10001000b
                db      00110000b,00000001b,00000111b,10000100b
                db      10110000b,00000011b,00001110b,10000000b
                db      01010101b,10101010b,00011100b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@24             db      00000000b,00000000b,00000000b,00000000b
fn_success_udg  db      00000000b,00000000b,00111111b,11111100b
fn_failure_udg  db      00000000b,00000000b,01111111b,11111110b
reset_udg       db      00000000b,00000000b,01100000b,00000110b
                db      00000011b,11000000b,01100001b,10000110b
                db      00000011b,01100011b,01100101b,10100110b
                db      00000110b,00110110b,01101001b,10010110b
                db      00000110b,00011100b,01101001b,10010110b
                db      11001100b,00110110b,01101001b,10010110b
                db      01101100b,01100011b,01101000b,00010110b
                db      00111000b,11000000b,01100111b,11100110b
                db      00011000b,00000000b,01100000b,00000110b
                db      00000000b,00000000b,01111111b,11111110b
                db      00000000b,00000000b,00111111b,11111100b
                db      00011000b,00011000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@28             db      00000000b,00000000b,00000000b,00000000b
crsr_arrows     db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00001000b,00001000b,00001000b,00011100b
                db      00011000b,00001100b,00011100b,00011100b
                db      00111111b,01111110b,00111110b,00011100b
                db      01111111b,01111111b,01111111b,01111111b
                db      00111111b,01111110b,00011100b,00111110b
                db      00011000b,00001100b,00011100b,00011100b
                db      00001000b,00001000b,00011100b,00001000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@32             db      00000000b,00000000b,00000000b,00000000b
ASCII_start     db      00000000b,00001100b,00011000b,00000000b
                db      00000000b,00001100b,00110011b,00100010b
                db      00000000b,00001100b,01100110b,00100010b
                db      00000000b,00011000b,11001100b,01111111b
                db      00000000b,00011000b,00011000b,00100010b
                db      00000000b,00011000b,00000000b,00101010b
                db      00000000b,00110000b,00000000b,00101010b
                db      00000000b,00110000b,00000000b,00100010b
                db      00000000b,00110000b,00000000b,01111111b
                db      00000000b,00000000b,00000000b,00100010b
                db      00000000b,01100000b,00000000b,00100010b
                db      00000000b,01100000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@36             db      00001100b,00000000b,00111100b,00000000b
                db      00001100b,00000011b,01111110b,00011000b
                db      00111110b,01100011b,01100110b,00011000b
                db      01111111b,01100110b,01100100b,00001000b
                db      01100000b,00000110b,00110000b,00001000b
                db      01100000b,00001100b,00011000b,00000000b
                db      01111110b,00001100b,00110000b,00000000b
                db      00111111b,00011000b,01100000b,00000000b
                db      00000011b,00011000b,11000110b,00000000b
                db      00000011b,00110000b,11011111b,00000000b
                db      01111111b,00110011b,11000110b,00000000b
                db      00111110b,01100011b,11000110b,00000000b
                db      00001100b,01100000b,01111110b,00000000b
                db      00001100b,00000000b,00111100b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@40             db      00001100b,00110000b,00000000b,00000000b
                db      00001100b,00110000b,00001000b,00000000b
                db      00011000b,00011000b,00001000b,00000000b
                db      00011000b,00011000b,01111111b,00011000b
                db      00011000b,00011000b,00011100b,00011000b
                db      00110000b,00001100b,00010100b,00011000b
                db      00110000b,00001100b,00100010b,01111110b
                db      00110000b,00001100b,00000000b,01111110b
                db      00110000b,00001100b,00000000b,00011000b
                db      00110000b,00001100b,00000000b,00011000b
                db      00110000b,00001100b,00000000b,00011000b
                db      00011000b,00011000b,00000000b,00000000b
                db      00011000b,00011000b,00000000b,00000000b
                db      00011000b,00011000b,00000000b,00000000b
                db      00001100b,00110000b,00000000b,00000000b
                db      00001100b,00110000b,00000000b,00000000b
@44             db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000011b
                db      00000000b,00000000b,00000000b,00000011b
                db      00000000b,00000000b,00000000b,00000110b
                db      00000000b,00000000b,00000000b,00000110b
                db      00000000b,00000000b,00000000b,00001100b
                db      00000000b,01111110b,00000000b,00001100b
                db      00000000b,01111110b,00000000b,00011000b
                db      00000000b,00000000b,00000000b,00011000b
                db      00000000b,00000000b,00000000b,00110000b
                db      00000000b,00000000b,00000000b,00110000b
                db      00011000b,00000000b,00011000b,01100000b
                db      00011000b,00000000b,00011000b,01100000b
                db      00001000b,00000000b,00000000b,00000000b
                db      00001000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@48             db      00000000b,00000000b,00000000b,00000000b
                db      00111110b,00000100b,00111110b,00111110b
                db      01111111b,00001100b,01111111b,01111111b
                db      01100011b,00011100b,01100011b,01100011b
                db      01100011b,00001100b,00000011b,00000011b
                db      01100011b,00001100b,00000110b,00000011b
                db      01101011b,00001100b,00001100b,00011110b
                db      01101011b,00001100b,00011000b,00011110b
                db      01100011b,00001100b,00110000b,00000011b
                db      01100011b,00001100b,01100000b,00000011b
                db      01100011b,00001100b,01100000b,01100011b
                db      01111111b,00001100b,01111111b,01111111b
                db      00111110b,00001100b,00111111b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@52             db      00000000b,00000000b,00000000b,00000000b
                db      01000000b,01111111b,00111110b,01111111b
                db      01100000b,01111111b,01111111b,01111111b
                db      01100000b,01100000b,01100000b,00000011b
                db      01100000b,01100000b,01100000b,00000011b
                db      01100000b,01100000b,01100000b,00000011b
                db      01100100b,01111110b,01111110b,00001111b
                db      01100110b,01111111b,01111111b,00001111b
                db      01100110b,00000011b,01100011b,00000011b
                db      01111111b,00000011b,01100011b,00000011b
                db      00111111b,01100011b,01100011b,00000011b
                db      00000110b,01111111b,01111111b,00000011b
                db      00000110b,00111110b,00111110b,00000011b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@56             db      00000000b,00000000b,00000000b,00000000b
                db      00111110b,00111110b,00000000b,00000000b
                db      01111111b,01111111b,00000000b,00000000b
                db      01100011b,01100011b,00000000b,00000000b
                db      01100011b,01100011b,00011000b,00011000b
                db      01100011b,01100011b,00011000b,00011000b
                db      00111110b,01111111b,00000000b,00000000b
                db      01111111b,00111111b,00000000b,00000000b
                db      01100011b,00000011b,00011000b,00011000b
                db      01100011b,00000011b,00011000b,00011000b
                db      01100011b,00000011b,00000000b,00001000b
                db      01111111b,00000011b,00000000b,00001000b
                db      00111110b,00000010b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@60             db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00111110b
                db      00000000b,00000000b,00000000b,01111111b
                db      00001100b,00000000b,00110000b,01100011b
                db      00011000b,00111100b,00011000b,00000011b
                db      00110000b,00111100b,00001100b,00000111b
                db      01100000b,00000000b,00000110b,00001110b
                db      00110000b,00000000b,00001100b,00001100b
                db      00011000b,00111100b,00011000b,00001100b
                db      00001100b,00111100b,00110000b,00000000b
                db      00000100b,00000000b,00100000b,00000000b
                db      00000000b,00000000b,00000000b,00001100b
                db      00000000b,00000000b,00000000b,00001100b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@64             db      00000000b,00000000b,00000000b,00000000b
                db      00111100b,00111110b,01111110b,00111110b
                db      01100110b,01111111b,01111111b,01111111b
                db      01000010b,01100011b,01100011b,01100011b
                db      11000011b,01100011b,01100011b,01100000b
                db      10011001b,01100011b,01100011b,01100000b
                db      10100101b,01111111b,01111110b,01100000b
                db      10100111b,01111111b,01111110b,01100000b
                db      10011011b,01100011b,01100011b,01100000b
                db      11000000b,01100011b,01100011b,01100000b
                db      01000000b,01100011b,01100011b,01100011b
                db      01100000b,01100011b,01111111b,01111111b
                db      00111100b,01100011b,01111110b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@68             db      00000000b,00000000b,00000000b,00000000b
                db      01111110b,01111111b,01111111b,00111110b
                db      01111111b,01111111b,01111111b,01111111b
                db      01100011b,01100000b,01100000b,01100011b
                db      01100011b,01100000b,01100000b,01100000b
                db      01100011b,01100000b,01100000b,01100000b
                db      01100011b,01111110b,01111110b,01100110b
                db      01100011b,01111110b,01111110b,01100111b
                db      01100011b,01100000b,01100000b,01100011b
                db      01100011b,01100000b,01100000b,01100011b
                db      01100011b,01100000b,01100000b,01100011b
                db      01111111b,01111111b,01100000b,01111111b
                db      01111110b,01111111b,01100000b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@72             db      00000000b,00000000b,00000000b,00000000b
                db      01100011b,00011000b,00000011b,01100000b
                db      01100011b,00011000b,00000011b,01100011b
                db      01100011b,00011000b,00000011b,01100110b
                db      01100011b,00011000b,00000011b,01101100b
                db      01100011b,00011000b,00000011b,01111000b
                db      01111111b,00011000b,00000011b,01111000b
                db      01111111b,00011000b,00000011b,01101100b
                db      01100011b,00011000b,00000011b,01101100b
                db      01100011b,00011000b,01100011b,01100110b
                db      01100011b,00011000b,01100011b,01100110b
                db      01100011b,00011000b,01111111b,01100011b
                db      01100011b,00011000b,00111110b,01100011b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@76             db      00000000b,00000000b,00000000b,00000000b
                db      01100000b,01000001b,01000011b,00111110b
                db      01100000b,01100011b,01100011b,01111111b
                db      01100000b,01110111b,01110011b,01100011b
                db      01100000b,01111111b,01111011b,01100011b
                db      01100000b,01111111b,01111111b,01100011b
                db      01100000b,01101011b,01101111b,01100011b
                db      01100000b,01100011b,01100111b,01100011b
                db      01100000b,01100011b,01100011b,01100011b
                db      01100000b,01100011b,01100011b,01100011b
                db      01100000b,01100011b,01100011b,01100011b
                db      01111111b,01100011b,01100011b,01111111b
                db      01111111b,01100011b,01100011b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@80             db      00000000b,00000000b,00000000b,00000000b
                db      01111110b,00111110b,01111110b,00111110b
                db      01111111b,01111111b,01111111b,01111111b
                db      01100011b,01100011b,01100011b,01100011b
                db      01100011b,01100011b,01100011b,01100000b
                db      01100011b,01100011b,01100011b,01100000b
                db      01111111b,01100011b,01111111b,01111110b
                db      01111110b,01100011b,01111110b,00111111b
                db      01100000b,01100011b,01100111b,00000011b
                db      01100000b,01100011b,01100011b,00000011b
                db      01100000b,01100000b,01100011b,01100011b
                db      01100000b,01111011b,01100011b,01111111b
                db      01100000b,00111011b,01100011b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@84             db      00000000b,00000000b,00000000b,00000000b
                db      01111110b,01100011b,01000001b,01000001b
                db      01111110b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01101011b
                db      00011000b,01100011b,01100011b,01111111b
                db      00011000b,01100011b,01110111b,01111111b
                db      00011000b,01100011b,00111110b,01110111b
                db      00011000b,01111111b,00011100b,01100011b
                db      00011000b,00111110b,00001000b,01000001b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@88             db      00000000b,00000000b,00000000b,00000000b
                db      01100011b,01100011b,01111111b,00111100b
                db      01100011b,01100011b,01111111b,00111100b
                db      00110110b,01100011b,00000110b,00110000b
                db      00110110b,01100011b,00000110b,00110000b
                db      00001100b,01100011b,00001100b,00110000b
                db      00001100b,01111111b,00001100b,00110000b
                db      00011000b,00111111b,00011000b,00110000b
                db      00011000b,00000011b,00011000b,00110000b
                db      00110110b,00000011b,00110000b,00110000b
                db      00110110b,00000011b,00110000b,00110000b
                db      01100011b,01111111b,01111111b,00111100b
                db      01100011b,01111110b,01111111b,00111100b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
@92             db      00000000b,00000000b,00000000b,00000000b
                db      01100000b,00111100b,00001000b,00000000b
                db      01100000b,00111100b,00011100b,00000000b
                db      00110000b,00001100b,00111110b,00000000b
                db      00110000b,00001100b,00000000b,00000000b
                db      00011000b,00001100b,00001000b,00000000b
                db      00011000b,00001100b,00001000b,00000000b
                db      00001100b,00001100b,00001000b,00000000b
                db      00001100b,00001100b,00001000b,00000000b
                db      00000110b,00001100b,00001000b,00000000b
                db      00000110b,00001100b,00001000b,00000000b
                db      00000011b,00111100b,00001000b,00000000b
                db      00000011b,00111100b,00001000b,00000000b
                db      00000000b,00000000b,00000000b,01111111b
                db      00000000b,00000000b,00000000b,11111110b
                db      00000000b,00000000b,00000000b,00000000b
@96             db      00000000b,00000000b,00000000b,00000000b
                db      00100000b,00000000b,01100000b,00000000b
                db      00110000b,00000000b,01100000b,00000000b
                db      00011000b,00000000b,01100000b,00000000b
                db      00001100b,00000000b,01100000b,00000000b
                db      00000000b,00000000b,01100000b,00000000b
                db      00000000b,00111110b,01111110b,00111110b
                db      00000000b,01111111b,01111111b,01111111b
                db      00000000b,00000011b,01100011b,01100000b
                db      00000000b,01100011b,01100011b,01100000b
                db      00000000b,01100011b,01100011b,01100000b
                db      00000000b,01111111b,01111111b,01111111b
                db      00000000b,00111111b,00111110b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
I00             db      00000000b,00000000b,00000000b,00000000b
                db      00000011b,00000000b,00000111b,00000000b
                db      00000011b,00000000b,00001111b,00000000b
                db      00000011b,00000000b,00001100b,00000000b
                db      00000011b,00000000b,00001100b,00000000b
                db      00000011b,00000000b,00111111b,00000000b
                db      00111111b,00111110b,00111111b,00111110b
                db      01111111b,01111111b,00001100b,01111111b
                db      01100011b,01100011b,00001100b,01100011b
                db      01100011b,01100011b,00001100b,01100011b
                db      01100011b,01100000b,00001100b,01100011b
                db      01111111b,01111111b,00001100b,01111011b
                db      00111110b,00111110b,00001100b,00111011b
                db      00000000b,00000000b,00000000b,00000011b
                db      00000000b,00000000b,00000000b,00111111b
                db      00000000b,00000000b,00000000b,00111110b
I04             db      00000000b,00000000b,00000000b,00000000b
                db      01100000b,00000000b,00000000b,01100000b
                db      01100000b,00011000b,00000110b,01100000b
                db      01100000b,00011000b,00000110b,01100001b
                db      01100000b,00000000b,00000000b,01100011b
                db      01100000b,00000000b,00000000b,01100110b
                db      01111110b,00011000b,00000110b,01101100b
                db      01111111b,00011000b,00000110b,01111000b
                db      01100011b,00011000b,00000110b,01111000b
                db      01100011b,00011000b,00000110b,01101100b
                db      01100011b,00011000b,00000110b,01100110b
                db      01100011b,00011000b,00000110b,01100011b
                db      01100011b,00011000b,00000110b,01100001b
                db      00000000b,00000000b,00000110b,00000000b
                db      00000000b,00000000b,01111110b,00000000b
                db      00000000b,00000000b,01111100b,00000000b
I08             db      00000000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00110110b,00111110b,00111110b
                db      00011000b,01111111b,01111111b,01111111b
                db      00011000b,01111111b,01100011b,01100011b
                db      00011000b,01101011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011100b,01100011b,01100011b,01111111b
                db      00001100b,01100011b,01100011b,00111110b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
I12             db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00111110b,00111110b,00111111b,00111111b
                db      01111111b,01111111b,01111111b,01111111b
                db      01100011b,01100011b,01100000b,01100000b
                db      01100011b,01100011b,01100000b,01100011b
                db      01100011b,01100011b,01100000b,00000011b
                db      01111111b,01111111b,01100000b,01111111b
                db      01111110b,00111111b,01100000b,01111110b
                db      01100000b,00000011b,00000000b,00000000b
                db      01100000b,00000011b,00000000b,00000000b
                db      01100000b,00000011b,00000000b,00000000b
I16             db      00000000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      01111110b,00000000b,00000000b,00000000b
                db      01111110b,00000000b,00000000b,00000000b
                db      00011000b,00000000b,00000000b,00000000b
                db      00011000b,01100011b,01100011b,01100011b
                db      00011000b,01100011b,01100011b,01101011b
                db      00011000b,01100011b,01100011b,01111111b
                db      00011000b,01100011b,01100011b,01111111b
                db      00011000b,01100011b,00110110b,01110111b
                db      00011110b,01111111b,00011100b,01100011b
                db      00001110b,00111111b,00001000b,01000001b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
                db      00000000b,00000000b,00000000b,00000000b
I20             db      00000000b,00000000b,00000000b,00001100b
                db      00000000b,00000000b,00000000b,00011100b
                db      00000000b,00000000b,00000000b,00011000b
                db      00000000b,00000000b,00000000b,00011000b
                db      00000000b,00000000b,00000000b,00011000b
                db      00000000b,00000000b,00000000b,00011100b
                db      01100011b,01100011b,01111111b,00001100b
                db      00110110b,01100011b,01111110b,00111000b
                db      00011100b,01100011b,00001100b,00111000b
                db      00001000b,01100011b,00011000b,00001100b
                db      00011100b,01100011b,00110000b,00011100b
                db      00110110b,01111111b,01111111b,00011000b
                db      01100011b,00111111b,01111111b,00011000b
                db      00000000b,00000011b,00000000b,00011000b
                db      00000000b,00111111b,00000000b,00011100b
                db      00000000b,00111110b,00000000b,00001100b
I24             db      00000000b,00011000b,00000000b,00000000b
ASCII_end       db      00011000b,00011100b,00000000b,00000000b
enter_tail      db      00011000b,00001100b,00000000b,00001110b
                db      00011000b,00001100b,00000000b,00001110b
                db      00011000b,00001100b,00000000b,00001110b
                db      00011000b,00011100b,00110011b,00001110b
                db      00011000b,00011000b,01111011b,00001110b
                db      00011000b,00001110b,01101111b,11111110b
                db      00011000b,00001110b,01100110b,11111110b
                db      00011000b,00011000b,00000000b,11111110b
                db      00011000b,00011100b,00000000b,00000000b
                db      00011000b,00001100b,00000000b,00000000b
                db      00011000b,00001100b,00000000b,00000000b
                db      00011000b,00001100b,00000000b,00000000b
                db      00011000b,00011100b,00000000b,00000000b
                db      00000000b,00011000b,00000000b,00000000b
; generic data for System32 resident functions 
eax2            label   dword                   ; fake register
 ax2            label   word                    ; 16 bits LSW
 al2            label   byte                    ;  8 bits LSB
                db      0
 ah2            label   byte                    ;  8 bits MSB
                db      0, 0, 0
ebx2            label   dword                   ; fake register
 bx2            label   word                    ; 16 bits LSW
 bl2            label   byte                    ;  8 bits LSB
                db      0
 bh2            label   byte                    ;  8 bits MSB
                db      0, 0, 0
ecx2            label   dword                   ; fake register
 cx2            label   word                    ; 16 bits LSW
 cl2            label   byte                    ;  8 bits LSB
                db      0
 ch2            label   byte                    ;  8 bits MSB
                db      0, 0, 0
edx2            label   dword                   ; fake register
 dx2            label   word                    ; 16 bits LSW
 dl2            label   byte                    ;  8 bits LSB
                db      0
 dh2            label   byte                    ;  8 bits MSB
                db      0, 0, 0
efx             label   dword                   ; special fake register
 fx             label   word                    ; 16 bits LSW
 fl             label   byte                    ;  8 bits LSB
                db      0
 fh             label   byte                    ;  8 bits MSB
                db      0
fxh             label   word                    ; 16 bits MSW, 16 to 32
efl             label   byte                    ;  8 bits from 16 to 24
                db      0
efh             label   byte                    ;  8 bits from 24 to 32
                db      0
aux?0           dd      0                       ; single temporary variable
aux?1           dd      0                       ; temporary variable nr. 1
aux?2           dd      0                       ; temporary variable nr. 2
aux?3           dd      0                       ; temporary variable nr. 3
aux?4           dd      0                       ; temporary variable nr. 4
aux?5           dd      0                       ; temporary variable nr. 5
aux?6           dd      0                       ; temporary variable nr. 6
aux?7           dd      0                       ; temporary variable nr. 7
aux?8           dd      0                       ; temporary variable nr. 8
aux?9           dd      0                       ; temporary variable nr. 9
aux?a           dd      0                       ; temporary variable nr. 10
aux?b           dd      0                       ; temporary variable nr. 11
aux?c           dd      0                       ; temporary variable nr. 12
aux?d           dd      0                       ; temporary variable nr. 13
aux?e           dd      0                       ; temporary variable nr. 14
aux?f           dd      0                       ; temporary variable nr. 15
; GATLoad 
; > Input:      EAX - pointer to object code (preceeded by its GOID)
; > Output:     (nothing)
                db      'GATLoad', 7            ; Global Object IDentifier
GATLoad:        push    eax                     ; Before doing anything,
                call    GATLookup               ;  perform a lookup to see
                pop     eax                     ;  if this object wasn't
                jnc     GATLoaderror            ;  already loaded to the GAT
                mov     ebp, dword ptr ds:[@gat]; number of slots in GAT
                mov     esi, @gat + 4           ; first slot
loadlookup:     cmp     dword ptr [esi], 0      ; check if this slot is empty
                je      loademptyslot           ;  (get there if yes)
                add     esi, 4                  ;  (else keep running
                dec     ebp                     ;   but when the GAT ends,
                jnz     loadlookup              ;   stop and return error)
GATLoaderror:   stc
                ret
loademptyslot:  mov     [esi], eax              ; load new GAT slot
                clc
                ret
; GATUnload 
; > Input:      EAX - pointer to GOID of object to unload (GOID is a string)
; > Output:     (nothing)
                db      'GATUnload', 9          ; Global Object IDentifier
GATUnload:      call    GATLookup               ; Find the object's GAT slot
                jc      GATUnloaderror          ;  (fail if not found)
                mov     dword ptr [esi], 0      ; unload object's GAT slot
                clc                             ;  (remembering ESI is loaded
                ret                             ;   by "GATLookup", see it...)
GATUnloaderror: stc
                ret
; GATLookup 
; > Input:      EAX - pointer to GOID of object (the GOID is a string)
; > Output:     EAX - pointer to corresponding object from the GAT
                db      'GATLookup', 9          ; Global Object IDentifier
GOIDptr         dd      ?                       ; temporary for preserving it
GATLookup:      mov     GOIDptr[@cb], eax       ; preserve original pointer
                mov     ebp, dword ptr ds:[@gat]; number of slots in GAT
                mov     esi, @gat + 4           ; first slot
lookupnextslot: cmp     dword ptr [esi], 0      ; check if this slot is empty
                je      nolongerequals          ;  (get there if yes)
                mov     ebx, [esi]              ; get pointer to slot's GOID
                dec     ebx                     ;  (1 byte before code/data)
                movzx   edx, byte ptr [ebx]     ; GOID size goes in edx
                sub     ebx, edx                ; now [ebx]=first GOID's byte
                mov     eax, GOIDptr[@cb]       ; and [eax]=ref. GOID's bytes
lookupnextbyte: mov     ch, [eax]               ; 
                mov     cl, [ebx]               ; 
                cmp     ch, cl                  ; 
                jne     nolongerequals          ; 
                cmp     ch, 0                   ; 
                je      nolongerequals          ; 
                cmp     cl, 0                   ; 
                je      nolongerequals          ; 
                cmp     ch, '>'                 ; 
                je      nolongerequals          ; 
                cmp     cl, '>'                 ;  just comparing the two
                je      nolongerequals          ;  strings: four possible
                cmp     ch, ','                 ;  terminators: NULL > , &
                je      nolongerequals          ; 
                cmp     cl, ','                 ; 
                je      nolongerequals          ; 
                cmp     ch, '&'                 ; 
                je      nolongerequals          ; 
                cmp     cl, '&'                 ; 
                je      nolongerequals          ; 
                inc     eax                     ; 
                inc     ebx                     ; 
                dec     edx                     ; 
                jnz     lookupnextbyte          ; 
                jmp     GLfound                 ; 
nolongerequals: add     esi, 4                  ; forwarding to next slot
                dec     ebp                     ;  until end of GAT
                jnz     lookupnextslot
                stc                             ; if coming here,
                ret                             ;  this object wasn't found
GLfound:        mov     eax, [esi]              ; if coming here,
                clc                             ;  object WAS found,
                ret                             ;  so return its pointer
; ReSyncTTCounter 
; > Input:      (nothing)
; > Output:     (nothing)
rtc_year        dd      1975
rtc_month       dd      12
rtc_day         dd      31
rtc_hour        dd      0
rtc_minutes     dd      0
rtc_seconds     dd      0
rtc_tenths      dd      0
rtc_dpmtable    dw      000, 031, 059
                dw      090, 120, 151
                dw      181, 212, 243
                dw      273, 304, 334, 365
                db      'ReSyncTTCounter', 15   ; Global Object IDentifier
ReSyncTTCounter:cli
                mov     ah, 04h                 ; get date from realtime clock
                int     1Ah
                jc      rtc_error
                movzx   ebx, ch                 ; century
                call    _2digitsBCDconv
                imul    ebx, 100
                mov     rtc_year[@cb], ebx
                movzx   ebx, cl                 ; year
                call    _2digitsBCDconv
                add     rtc_year[@cb], ebx
                movzx   ebx, dh                 ; month
                call    _2digitsBCDconv
                mov     rtc_month[@cb], ebx
                movzx   ebx, dl                 ; day
                call    _2digitsBCDconv
                mov     rtc_day[@cb], ebx
                mov     ah, 02h                 ; get time from realtime clock
                int     1Ah
                jc      rtc_error
                movzx   ebx, ch                 ; hour
                call    _2digitsBCDconv
                mov     rtc_hour[@cb], ebx
                movzx   ebx, cl                 ; minutes
                call    _2digitsBCDconv
                mov     rtc_minutes[@cb], ebx
                movzx   ebx, dh                 ; seconds
                call    _2digitsBCDconv
                mov     rtc_seconds[@cb], ebx
                call    rtc_convdate
                mov     ds:[@tts + 0], eax
                mov     ds:[@tts + 4], edx
rtc_success:    sti
                clc
                ret
rtc_error:      sti
                stc
                ret
_2digitsBCDconv:push    ebx                     ; BCD input: ebx (00h to 99h)
                shr     ebx, 4                  ; binary output: ebx (0 to 99)
                mov     ebp, ebx
                lea     ebx, [ebx + ebx * 8]
                add     ebx, ebp
                pop     ebp
                and     ebp, 0Fh
                add     ebx, ebp
                ret
rtc_convdate:   mov     eax, rtc_year[@cb]      ; computing TTS since 12/31/75
                sub     eax, 1976               ; input: rtc_x variables
                jc      rtc_error               ; output: EAX:EDX
                mov     ebx, eax                ; EBX = (year - 1976)
                imul    eax, 365                ; EAX = (year - 1976) * 365
                add     ebx, 2                  ; EBX = (year - 1974)
                shr     ebx, 2                  ; EBX = (year - 1974) / 4
                lea     eax, [eax + ebx + 1]    ; EAX = EAX + EBX + 1
                push    eax                     ;  (+1 is last day of 1975)
                mov     eax, rtc_year[@cb]      ;  (while EBX gives how many
                xor     edx, edx                ;   29ths of februaries have
                mov     ecx, 400                ;   passed...)
                div     ecx                     ; (year / 400)
                mov     ebx, eax                ;  because every 400 years
                pop     eax                     ;  there's an additional day
                add     eax, ebx                ;  like happened in year 2000.
                mov     ebx, rtc_month[@cb]
                movzx   ebx, rtc_dpmtable[@cb + ebx*2 - 2]
                add     eax, ebx                ; add #days before this month
                add     eax, rtc_day[@cb]       ; add #days within this month
                dec     eax                     ; -1 (day is based at 1)
                mov     ebx, 4294402905 ; (tts per day)
                mul     ebx
                push    eax edx
                mov     eax, rtc_hour[@cb]
                mov     edx, 178933454 ; (tts per hour)
                mul     edx
                mov     ebx, eax
                pop     edx eax
                add     eax, ebx
                adc     edx, 0
                mov     ebx, rtc_minutes[@cb]
                imul    ebx, 2982224 ; (tts per minute)
                add     eax, ebx
                adc     edx, 0
                mov     ebx, rtc_seconds[@cb]
                imul    ebx, 49704 ; (tts per second)
                add     eax, ebx
                adc     edx, 0
                ret
; DateTime2TT 
; > Input:      EAX - year:             above or equal to 1976.
;               EBX - month:            1 to 12
;               ECX - day of month:     1 to 31
;               EDX - hour:             0 to 23, zero is midnight
;               ESI - minutes:          0 to 59
;               EDI - seconds:          0 to 59
; > Output:     EAX:EBX - tiriticks value corresponding to entry datetime.
;                         (EAX is the most significant doubleword.)
                db      'DateTime2TT', 11       ; Global Object IDentifier
DateTime2TT:    mov     rtc_year[@cb], eax
                mov     rtc_month[@cb], ebx
                mov     rtc_day[@cb], ecx
                mov     rtc_hour[@cb], edx
                mov     rtc_minutes[@cb], esi
                mov     rtc_seconds[@cb], edi
                call    rtc_convdate
                mov     ebx, edx
                ret
; TT2DateTime 
; > Input:      EAX:EBX - timer ticks value to convert.
;                         (EAX is the most significant doubleword.)
; > Output:     EAX - year:             above or equal to 1976.
;               EBX - month:            1 to 12
;               ECX - day of month:     1 to 31
;               EDX - hour:             0 to 23, zero is midnight
;               ESI - minutes:          0 to 59
;               EDI - seconds:          0 to 59
;               EBP - tenths of second: 0 to 9
tt2_days        dd      ?
tt2_rest        dd      ?
                db      'TT2DateTime', 11       ; Global Object IDentifier
TT2DateTime:    sub     ebx, 4294402905 ; (missing day from 1975)
                sbb     eax, 0
                mov     edx, eax                ; counting days
                mov     eax, ebx
                mov     ecx, 4294402905 ; (tts per day)
                div     ecx
                mov     tt2_days[@cb], eax
                mov     tt2_rest[@cb], edx
                xor     edx, edx                ; counting years
                mov     ecx, 365
                div     ecx
                push    eax                     ; counting abnormal years
                add     eax, 1976               ;  every 400 years
                xor     edx, edx                ;  (case of year 2000)
                mov     ecx, 400
                div     ecx
                sub     tt2_days[@cb], eax
                pop     eax
                add     eax, 2                  ; counting 29ths of februaries
                cmp     eax, 2
                jl      tt2_error
                shr     eax, 2
                sub     tt2_days[@cb], eax
                mov     eax, tt2_days[@cb]      ; computing year
                xor     edx, edx
                mov     ecx, 365
                div     ecx
                add     eax, 1976
                mov     rtc_year[@cb], eax
                mov     tt2_days[@cb], edx
                mov     eax, 1                  ; computing month
                mov     ebx, tt2_days[@cb]
tt2_findmonth:  cmp     bx, rtc_dpmtable[@cb + eax * 2]
                jbe     tt2_foundmonth
                inc     eax
                jmp     tt2_findmonth
tt2_foundmonth: mov     rtc_month[@cb], eax
                m2m     rtc_day[@cb], tt2_days[@cb]
                movzx   ebx, rtc_dpmtable[@cb + eax * 2 - 2]
                sub     rtc_day[@cb], ebx       ; computing day of month
                inc     rtc_day[@cb]
                mov     eax, tt2_rest[@cb]      ; computing hour
                xor     edx, edx
                mov     ecx, 178933454 ; (tts per hour)
                div     ecx
                mov     rtc_hour[@cb], eax
                mov     eax, edx                ; computing minutes
                xor     edx, edx
                mov     ecx, 2982224 ; (tts per minute)
                div     ecx
                mov     rtc_minutes[@cb], eax
                mov     eax, edx                ; computing seconds
                xor     edx, edx
                mov     ecx, 49704 ; (tts per second)
                div     ecx
                mov     rtc_seconds[@cb], eax
                mov     eax, edx                ; computing tenths of second
                xor     edx, edx
                mov     ecx, 4971 ; (tts per tenth of second)
                div     ecx
                mov     rtc_tenths[@cb], eax
                mov     eax, rtc_year[@cb]      ; returns resulting datetime
                mov     ebx, rtc_month[@cb]
                mov     ecx, rtc_day[@cb]
                mov     edx, rtc_hour[@cb]
                mov     esi, rtc_minutes[@cb]
                mov     edi, rtc_seconds[@cb]
                mov     ebp, rtc_tenths[@cb]
tt2_success:    clc
                ret
tt2_error:      stc
                ret
; WrHyper 
; > Input:      EAX - address of hypertext page to write
;               EBX - 0 (new) clear contest, clear history, then write
;                     1 (top) clear contest, then write
;                     2 (append) don't clear, append to existing contest
; > Output:     (nothing)
MAX_TAG_SPECS   equ     8                       ; max n.of parameters per tag
htsize          dd      0                       ; initial hypertext size
appflag         db      0                       ; append flag
tag_resproutine dd      ?                       ; responsible routine for tag
tag_specs       dd      ?                       ; #parameters for current tag
tag_spec_1      dd      ?                       ; current tag parameter #1
tag_spec_1_size dd      ?
tag_spec_2      dd      ?                       ; current tag parameter #2
tag_spec_2_size dd      ?
tag_spec_3      dd      ?                       ; current tag parameter #3
tag_spec_3_size dd      ?
tag_spec_4      dd      ?                       ; current tag parameter #4
tag_spec_4_size dd      ?
tag_spec_5      dd      ?                       ; current tag parameter #5
tag_spec_5_size dd      ?
tag_spec_6      dd      ?                       ; current tag parameter #6
tag_spec_6_size dd      ?
tag_spec_7      dd      ?                       ; current tag parameter #7
tag_spec_7_size dd      ?
tag_spec_8      dd      ?                       ; current tag parameter #8
tag_spec_8_size dd      ?
                db      'WrHyper', 7            ; Global Object IDentifier
WrHyper:        cmp     ebx, 2
                jbe     wrh_idst1_pass
                stc
                ret
wrh_idst1_pass: mov     appflag[@cb], 1
                cmp     ebx, 2
                jne     wrh_new
                mov     edi, historyptr[@cb]
                and     edi, @hrange
                mov     dword ptr ds:[@hist + edi*4], 1
                inc     edi
                and     edi, @hrange
                mov     dword ptr ds:[@hist + edi*4], eax
                add     historyptr[@cb], 2
                jmp     wrh_append
wrh_new:        mov     appflag[@cb], 0
                cmp     ebx, 1
                je      wrh_top
                push    eax
                mov     edi, @hist
                mov     ecx, @hrange + 1
                xor     eax, eax
                cld
                rep     stosd
                pop     eax
wrh_top:        mov     edi, historyptr[@cb]
                and     edi, @hrange
                mov     dword ptr ds:[@hist + edi*4], eax
                inc     historyptr[@cb]
                mov     htsize[@cb], 0
                mov     starthtline[@cb], 0
                mov     selectedarea[@cb], 0
                mov     sa_start[@cb], 0
wrh_append:     mov     edi, @ht
                add     edi, htsize[@cb]
wrh_next:       mov     edx, [eax]              ; hypertext translator
                test    dl, dl
                jz      wrh_end
                cmp     dx, '''<'
                je      wrh_parse_dbllt         ; '<    writes "less than"
                cmp     dx, '''>'
                je      wrh_parse_dblgt         ; '>    writes "greater than"
                cmp     dx, '''\'
                je      wrh_parse_dblbs         ; '\    writes a backslash
                cmp     dx, '''&'
                je      wrh_parse_dblam         ; '&    writes an ampersand
                cmp     dl, '\'
                je      wrh_parse_caret         ; \     carriage return
                cmp     dl, '<'
                jne     wrh_nolt                ; <     alternate tag opener
                mov     dl, '&'                 ; <     becomes ampersand
                jmp     wrh_lt
wrh_nolt:       cmp     dl, '&'
                jne     wrh_notatag             ; "&" and "<" are tag openers
wrh_lt:         mov     tag_resproutine[@cb], offset tag_void[@cb]
                mov     tag_specs[@cb], 0
                mov     tag_spec_1[@cb], 0
                mov     tag_spec_2[@cb], 0
                mov     tag_spec_3[@cb], 0
                mov     tag_spec_4[@cb], 0
                mov     tag_spec_5[@cb], 0
                mov     tag_spec_6[@cb], 0
                mov     tag_spec_7[@cb], 0
                mov     tag_spec_8[@cb], 0
                cmp     edx, 'dne&'
                je      wrh_parse_end           ; &end  block terminator
                cmp     edx, 'thw&'
                je      wrh_parse_wht           ; &wht  writes in white
                cmp     edx, 'klb&'
                je      wrh_parse_blk           ; &blk  writes in black
                cmp     edx, 'der&'
                je      wrh_parse_red           ; &red  writes in red
                cmp     edx, 'nrg&'
                je      wrh_parse_grn           ; &grn  writes in green
                cmp     edx, 'ulb&'
                je      wrh_parse_blu           ; &blu  writes in blue
                cmp     edx, 'ley&'
                je      wrh_parse_yel           ; &yel  writes in yellow
                cmp     edx, 'gam&'
                je      wrh_parse_mag           ; &mag  writes in magenta
                cmp     edx, 'ayc&'
                je      wrh_parse_cya           ; &cya  writes in cyan
                cmp     edx, 'tgl&'
                je      wrh_parse_lgt           ; &lgt  light text color
                cmp     edx, 'krd&'
                je      wrh_parse_drk           ; &drk  dark text color
                cmp     edx, 'pap&'
                je      wrh_parse_pap           ; &pap  set paper color=text
                cmp     edx, 'vni&'
                je      wrh_parse_inv           ; &inv  xchange text<-->paper
                cmp     edx, 'loc&'
                je      wrh_parse_col           ; <col  move at column...
                cmp     edx, 'irc&'
                je      wrh_parse_cri           ; <cri  insert text critter
                cmp     edx, 'knl&'
                je      wrh_parse_lnk           ; <lnk  content link
                cmp     edx, 'exe&'
                je      wrh_parse_exe           ; <exe  executable link
                cmp     edx, 'det&'
                je      wrh_parse_ted           ; <ted  text editing area
                jmp     wrh_end                 ; else  invalid tag (break)
wrh_parsed_tag: m2m     <dword ptr [edi+00]>, tag_specs[@cb]
                m2m     <dword ptr [edi+04]>, tag_resproutine[@cb] ; in sys32
                m2m     <dword ptr [edi+08]>, tag_spec_1[@cb]
                m2m     <dword ptr [edi+12]>, tag_spec_2[@cb]
                m2m     <dword ptr [edi+16]>, tag_spec_3[@cb]
                m2m     <dword ptr [edi+20]>, tag_spec_4[@cb]
                m2m     <dword ptr [edi+24]>, tag_spec_5[@cb]
                m2m     <dword ptr [edi+28]>, tag_spec_6[@cb]
                m2m     <dword ptr [edi+32]>, tag_spec_7[@cb]
                m2m     <dword ptr [edi+36]>, tag_spec_8[@cb]
                add     edi, 8
                mov     edx, tag_specs[@cb]
                shl     edx, 2
                add     edi, edx
                jmp     wrh_next
wrh_notatag:    mov     [edi], dl
                inc     edi
                inc     eax
                jmp     wrh_next
wrh_end:        mov     dword ptr [edi], ''
                sub     edi, @ht
                mov     htsize[@cb], edi
                call    writestatusline
                call    writehtcontest
                call    writecpdiv
                mov     appflag[@cb], 0
                clc
                ret
wrh_parse_caret:mov     dl, 127
                jmp     wrh_notatag
wrh_parse_dblbs:inc     eax
                mov     dl, '\'
                jmp     wrh_notatag
wrh_parse_dblam:inc     eax
                mov     dl, '&'
                jmp     wrh_notatag
wrh_parse_dbllt:inc     eax
                mov     dl, '<'
                jmp     wrh_notatag
wrh_parse_dblgt:inc     eax
                mov     dl, '>'
                jmp     wrh_notatag
wrh_parse_end:  mov     tag_resproutine[@cb], offset tag_end[@cb]
                jmp     wrh_simple_tag
wrh_parse_wht:  mov     tag_resproutine[@cb], offset tag_wht[@cb]
                jmp     wrh_simple_tag
wrh_parse_blk:  mov     tag_resproutine[@cb], offset tag_blk[@cb]
                jmp     wrh_simple_tag
wrh_parse_red:  mov     tag_resproutine[@cb], offset tag_red[@cb]
                jmp     wrh_simple_tag
wrh_parse_grn:  mov     tag_resproutine[@cb], offset tag_grn[@cb]
                jmp     wrh_simple_tag
wrh_parse_blu:  mov     tag_resproutine[@cb], offset tag_blu[@cb]
                jmp     wrh_simple_tag
wrh_parse_yel:  mov     tag_resproutine[@cb], offset tag_yel[@cb]
                jmp     wrh_simple_tag
wrh_parse_mag:  mov     tag_resproutine[@cb], offset tag_mag[@cb]
                jmp     wrh_simple_tag
wrh_parse_cya:  mov     tag_resproutine[@cb], offset tag_cya[@cb]
                jmp     wrh_simple_tag
wrh_parse_lgt:  mov     tag_resproutine[@cb], offset tag_lgt[@cb]
                jmp     wrh_simple_tag
wrh_parse_drk:  mov     tag_resproutine[@cb], offset tag_drk[@cb]
                jmp     wrh_simple_tag
wrh_parse_pap:  mov     tag_resproutine[@cb], offset tag_pap[@cb]
                jmp     wrh_simple_tag
wrh_parse_inv:  mov     tag_resproutine[@cb], offset tag_inv[@cb]
wrh_simple_tag: add     eax, 4
                cmp     byte ptr [eax], '>'
                jne     wrh_st_nogt
                inc     eax
wrh_st_nogt:    mov     tag_specs[@cb], 0
                jmp     wrh_parsed_tag
wrh_parse_col:  mov     tag_specs[@cb], 1       ; <col#spec1>
                call    wrh_get_specs           ; #spec1
                jc      wrh_end                 ;  column to move to (0..79)
                cmp     tag_spec_1[@cb], 79
                ja      wrh_end
                mov     tag_resproutine[@cb], offset tag_col[@cb]
                jmp     wrh_parsed_tag
wrh_parse_cri:  mov     tag_specs[@cb], 1       ; <cri#spec1>
                call    wrh_get_specs           ; #spec1
                jc      wrh_end                 ;  critter identificator 0..11
                cmp     tag_spec_1[@cb], 11
                ja      wrh_end
                mov     tag_resproutine[@cb], offset tag_cri[@cb]
                jmp     wrh_parsed_tag
wrh_parse_lnk:  mov     tag_specs[@cb], 2       ; <lnkspec1,#spec2>
                call    wrh_get_specs           ; spec1
                jc      wrh_end                 ;  GOID of page to link
                cmp     tag_spec_2[@cb], 2      ; #spec2 - destination frame
                ja      wrh_end                 ;  0=top, 1=new, 2=append
                mov     tag_resproutine[@cb], offset tag_lnk[@cb]
                jmp     wrh_parsed_tag
wrh_parse_exe:  mov     tag_specs[@cb], -1      ; <exespec1,[undef'd params]>
                call    wrh_get_specs           ; spec1 GOID of module
                jc      wrh_end                 ; [#]spec2-spec8 parameters
                mov     tag_resproutine[@cb], offset tag_exe[@cb]
                jmp     wrh_parsed_tag
wrh_parse_ted:  mov     tag_specs[@cb], 8       ; <tedspec1,#spec2...spec8>
                call    wrh_get_specs           ;  <spec1> GOID of text buffer
                jc      wrh_end                 ; #<spec2> text buffer size
                cmp     tag_spec_3[@cb], 78     ; #<spec3> visib. cols 0/1..78
                ja      wrh_end                 ; #<spec4> visible rows 1..24
                cmp     tag_spec_4[@cb], 24     ; #<spec5> effective columns
                ja      wrh_end                 ; #<spec6> effective rows
                cmp     tag_spec_4[@cb], 0      ; #<spec7> behaviour flags
                je      wrh_end                 ;  <spec8> GOID of manager
                mov     tag_resproutine[@cb], offset tag_ted[@cb]
                jmp     wrh_parsed_tag
wrh_get_specs:  add     eax, 4                  ; gets all parameters for tag
                mov ebx, offset tag_spec_1[@cb] ;  parameters are returned as
                mov     [ebx], eax              ;  pointers and lay within the
                add     ebx, 8                  ;  area provided as the origin
                xor     ecx, ecx                ;  for the whole hypertext.
wrh_gs_loop:    cmp     byte ptr [eax], 0       ;  if tag_specs on entry does
                je      wrh_gs_fail             ;  not match the ecx counter
                cmp     byte ptr [eax], '&'     ;  on exit, the call fails.
                je      wrh_gs_fail             ;  if extraneous & signs or
                cmp     byte ptr [eax], '>'     ;  null terminators are found
                je      wrh_gs_specend          ;  in the parameter string/s,
                cmp     byte ptr [eax], ','     ;  the call fails, either.
                je      wrh_gs_specend          ;  parameter strings, if more
                inc     eax                     ;  than one, are separated by
                jmp     wrh_gs_loop             ;  commas, and the parameters'
wrh_gs_specend: mov     edx, eax                ;  list must end with a ">".
                sub     edx, [ebx - 8]          ;  also, this subroutine will
                mov     [ebx - 4], edx          ;  take care of updating the
                mov     edx, [ebx - 8]          ;  value of eax so that it
                cmp     byte ptr [edx], '#'     ;  will point to the character
                jne     wrh_gs_isstring         ;  immediately following the
                pushad                          ;  "greater". Of course, this
                sub     ebx, 8                  ;  subroutine must be called
                inc     dword ptr [ebx]         ;  only if one or more specs
                dec     dword ptr [ebx + 4]     ;  are SUPPOSED to be found,
                call    wrh_gs_isnumber         ;  and exactly in the quantity
                popad                           ;  specified by "tag_specs".
                jc      wrh_gs_fail             ;  While for the type of spec,
wrh_gs_isstring:inc     eax                     ;  an integer value is given
                mov     [ebx], eax              ;  by a string made as "#val"
                add     ebx, 8                  ;  where "val" is the number,
                inc     ecx                     ;  because if the hash sign
                cmp     ecx, MAX_TAG_SPECS      ;  is not found as first char.
                ja      wrh_gs_fail             ;  of the said spec, it is
                cmp     byte ptr [eax-1], '>'   ;  considered a text string,
                jne     wrh_gs_loop             ;  and the pointer to that
wrh_gs_end:     cmp     tag_specs[@cb], -1      ;  string is returned instead.
                je      wrh_gs_tsundef
                cmp     ecx, tag_specs[@cb]
                je      wrh_gs_success
wrh_gs_fail:    stc
                ret
wrh_gs_tsundef: mov     tag_specs[@cb], ecx
wrh_gs_success: clc
                ret
wrh_gs_isnumber:push    eax                     ; converts parameter string
                mov     ecx, 1                  ;  indicated by [ebx] on entry
                xor     esi, esi                ;  to a number. the value of
                mov     edx, [ebx]              ;  location pointed by [ebx]
                add     edx, [ebx + 4]          ;  is holding the pointer to
wrh_gs_in_loop: dec     edx                     ;  the parameter string, on
                cmp     byte ptr [edx], '0'     ;  entry, and if all goes well
                jb      wrh_gs_in_end           ;  the pointer held by [ebx]
                cmp     byte ptr [edx], '9'     ;  is replaced by the value of
                ja      wrh_gs_in_end           ;  the parameter string.
                push    edx                     ;  if there are any symbols
                xor     eax, eax                ;  outside digits 0 to 9 in
                mov     al, [edx]               ;  the parameter string, the
                sub     al, '0'                 ;  call fails.
                mul     ecx                     ;  if resulting decimal value
                add     esi, eax                ;  cannot fit the 32-bit range
                pushf                           ;  the call fails, either.
                mov     eax, ecx                ;  typically, to convert spec.
                mov     ecx, 10                 ;  number X, ebx should be
                mul     ecx                     ;  loaded as...
                mov     ecx, eax                ;  ebx=offset tag_spec_x[@cb]
                popf                            ;  also, but this is performed
                pop     edx                     ;  by "wrh_get_specs", the
                jc      wrh_gs_in_break         ;  location following [ebx]
                jmp     wrh_gs_in_loop          ;  in 32-bit (so it's [ebx+4])
wrh_gs_in_end:  pop     eax                     ;  must contain the length of
                cmp     edx, [ebx]              ;  the parameter string, in
                jnb     wrh_gs_in_fail          ;  bytes, because the number
                mov     [ebx], esi              ;  needs to be decoded by
                clc                             ;  starting from last byte
                ret                             ;  in that string, which is
wrh_gs_in_break:pop     eax                     ;  therefore calculated as...
wrh_gs_in_fail: stc                             ;  ptr to last byte=
                ret                             ;   (ebx + [ebx + 4] - 1)
; WrBack 
; > Input:      (nothing)
; > Output:     (nothing)
                db      'WrBack', 6             ; Global Object IDentifier
WrBack:         mov     edx, 2
                mov     esi, historyptr[@cb]
wb_findparent:  sub     esi, 2
                and     esi, @hrange
                cmp     dword ptr ds:[@hist + esi*4], 0
                je      wb_top
                cmp     dword ptr ds:[@hist + esi*4], 1
                jne     wb_foundparent
                add     edx, 2
                jmp     wb_findparent
wb_foundparent: mov     ecx, 1
wb_findtop:     cmp     dword ptr ds:[@hist + esi*4 - 4], 1
                jne     wb_foundtop
                sub     esi, 2
                and     esi, @hrange
                inc     ecx
                add     edx, 2
                jmp     wb_findtop
wb_foundtop:    mov     eax, dword ptr ds:[@hist + esi*4]
                sub     historyptr[@cb], edx
                mov     ebx, 1
wb_extendloop:  push    ecx esi
                call    WrHyper
                pop     esi ecx
                mov     ebx, 2
                add     esi, 2
                and     esi, @hrange
                mov     eax, dword ptr ds:[@hist + esi*4]
                dec     ecx
                jnz     wb_extendloop
                ret
wb_top:         stc
                ret
; SystemReset 
; > Input:      EAX - reset mode, can assume one of these values:
;                     0 reset Lasting Child system environment only
;                     1 BIOS warm-boot, equals to Ctrl+Alt+Canc
;                     2 BIOS cold-boot, equals hardware reset
;                     3 power off, turns off the whole machine
;                     - any other values turn to hardware reset
; > Output:     (nothing, and never returns)
resetmsg        db      'System reset.'         ; Reset message...
sdstatus1       db      '\\  Power off:'
                db      '\   all devices power supply going down in 3 seconds.', 0
sdstatus2       db      '\\  (power management not supported)', 0
sdstatus3       db      '\\  (can''t enable power management)', 0
sdstatus4       db      '\\  (unable to enter POWER OFF state)', 0
                db      'SystemReset', 11       ; Global Object IDentifier
SystemReset:    cmp     eax, 3
                jne     sr_noAPMservice
                lea     eax, sdstatus1[@cb]
                mov     ebx, 2
                call    WrHyper
                mov     eax, ds:[46Ch]          ; 
                add     eax, 55                 ;  waiting 3 seconds for
sr_wait_3secs:  cmp     ds:[46Ch], eax          ;  (eventual) drives to stop
                jb      sr_wait_3secs           ; 
                mov     ax, 5300h
                xor     bx, bx
                int     15h                     ; APM installcheck
                jnc     sr_step_1
                lea     eax, sdstatus2[@cb]
                mov     ebx, 2
                call    WrHyper
                ret
sr_step_1:      mov     ax, 5301h
                xor     bx, bx
                int     15h                     ; Connect R/M interface
sr_step_2:      mov     ax, 530Eh
                xor     bx, bx
                mov     cx, 0101h
                int     15h                     ; Declare version 1.1+ driver
                jnc     sr_step_3
sr_err_unsupp:  lea     eax, sdstatus3[@cb]
                mov     ebx, 2
                call    WrHyper
                ret
sr_step_3:      mov     ax, 530Dh
                mov     bx, 0001h
                mov     cx, 0001h
                int     15h                     ; Engage power management
                jc      sr_err_unsupp
sr_step_4:      mov     ax, 5307h
                mov     bx, 0001h
                mov     cx, 0003h
                int     15h                     ; Power off
                lea     eax, sdstatus4[@cb]
                mov     ebx, 2
                call    WrHyper
                ret
sr_noAPMservice:cmp     eax, 1
                jne     sr_nonsoft
                mov     word ptr ds:[POSTresetflag], 1234h
                int     32h
sr_nonsoft:     cmp     eax, 2
                jne     sr_nonhard
sr_default:     mov     word ptr ds:[POSTresetflag], 0000h
                int     32h
sr_nonhard:     cmp     eax, 0                  ; BEGIN ENV. RESET CODE
                jne     sr_default
                xor     eax, eax                ; 
                mov     esi, @hdb               ; 
                mov     ecx, 41Ah               ; 
hdbchecksum_ver:add     eax, [esi]              ;  Verify hard data backup
                inc     esi                     ;  checksum: if wrong, go
                dec     ecx                     ;  for an hardware reset.
                jnz     hdbchecksum_ver         ; 
                cmp     hdbchecksum[@cb], eax   ; 
                jne     sr_default              ; 
                mov     word ptr ds:[@rmds], 0  ;  Reset real-mode segments
                mov     word ptr ds:[@rmes], 0  ;  for RM interrupt calls...
                cmp     byte ptr ds:[449h], 3   ; 
                je      sr_mode3                ;  set color text video mode
                mov     ax, 3                   ;  if not already set...
                int     10h                     ; 
                mov     eax, ds:[46Ch]          ; 
                add     eax, 37                 ;  waiting a couple of secs
sr_wait_2secs:  cmp     ds:[46Ch], eax          ;  for monitors to adjust...
                jb      sr_wait_2secs           ; 
                jmp     sr_goforreset           ; * proceed
sr_mode3:       mov     edi, @tx                ; 
                mov     ecx, 80 * 25            ;  clearing text from screen
                mov     ax, 0720h               ;  when mode 3 already set
                cld                             ; 
                rep     stosw                   ; 
sr_goforreset:  cli                             ; * environment reset! IRQ off
                mov     esi, @hdb               ; 
                xor     edi, edi                ;  restoring hardware data
                mov     ecx, 41Ah               ;  (R-M IVT and BDA)
                cld                             ; 
                rep     movsb                   ; 
; ------------- Disconnect flat-mode timer ticks handler ---------------------
                mov ds:[@fmidt + IRQ0*8 + 0], word ptr @f2r_it + IRQ0*4
                mov ds:[@fmidt + IRQ0*8 + 2], word ptr ICODE
                mov ds:[@fmidt + IRQ0*8 + 4], word ptr int_code
                mov ds:[@fmidt + IRQ0*8 + 6], word ptr 0000h
; ------------- End disconnect flat-mode timer ticks handler -----------------
                sti                             ; * IRQ on again
                mov     ecx, 65536              ;  waiting a few cycles for
                loop    $                       ;  interrupts to adjust...
sr_retry:       mov     ax, 0201h               ;  reading bootstrap loader
                mov     bx, 7C00h               ;  offset of bootstrap loader
                mov     cx, 0001h               ;  sector 1 of track 0
                mov     dl, ds:[7C02h]          ;  drive ID of latest boot
                xor     dh, dh                  ;  head 0
                int     13h                     ;  LEGACY ISA DRIVE SERVICES
                jnc     sr_loaded               ;  check error status
                cmp     ah, 11h                 ;  error code: corrected ECC
                je      sr_loaded               ;  (that wasn't a real error)
                cmp     ah, 06h                 ;  error code: disk changed
                je      sr_retry                ;  (ok, try again)
                jmp     sr_default              ;  critical error: hard reset
sr_loaded:      lea     bp, resetmsg[@cb]       ;  write "System reset."
                mov     dx, 1700h               ;  at line 23, column zero,
                mov     bx, 0Ch                 ;  intensified red, page 0,
                mov     cx, 13                  ;  message is 13 characters,
                mov     ax, 1301h               ;  BIOS function (textout),
                int     10h                     ;  BIOS VIDEO SERVICES
                mov     eax, ds:[46Ch]          ; 
                add     eax, 18                 ;  waiting 1 more second
sr_wait_1sec:   cmp     ds:[46Ch], eax          ;  for drive to rest a while
                jb      sr_wait_1sec            ; 
                mov word ptr ds:[64h*4], 7C00h  ;  set real-mode int. 64h
                mov word ptr ds:[64h*4+2], 0    ;  to fetch bootstrap loader
                int     64h                     ; * fetch it!
; In2Dec 
; > Input:      EAX - integer value to convert (32-bit unsigned integer)
;               EBX - ptr to destination buffer for decimal ASCII result
;               ECX - size of destination buffer in bytes (also no. of digits)
; > Output:     (nothing)
                db      'In2Dec', 6             ; Global Object IDentifier
In2Dec:         cmp     ecx, 0
                jle     i2d_return
                mov     esi, 10
i2d_loop:       xor     edx, edx
                div     esi
                add     dl, '0'
                mov     [ebx + ecx - 1], dl
                loop    i2d_loop
i2d_return:     ret
; In2Hex 
; > Input:      EAX - integer value to convert (32-bit unsigned integer)
;               EBX - ptr to destination buffer for hexadecimal ASCII result
;               ECX - size of destination buffer in bytes (also no. of digits)
; > Output:     (nothing)
                db      'In2Hex', 6             ; Global Object IDentifier
In2Hex:         cmp     ecx, 0
                jle     i2h_return
i2h_loop:       mov     dl, al
                and     dl, 0Fh
                cmp     dl, 10
                jnb     i2h_letter
                add     dl, '0'
                jmp     i2h_digit
i2h_letter:     add     dl, 'A' - 10
i2h_digit:      mov     [ebx + ecx - 1], dl
                shr     eax, 4
                loop    i2h_loop
i2h_return:     ret
; Dec2In 
; > Input:      EAX - ptr to source string (null-terminated)
; > Output:     EAX - unsigned integer value corresponding to the string
                db      'Dec2In', 6             ; Global Object IDentifier
Dec2In:         mov     edi, eax
d2i_fwd:        cmp     byte ptr [eax], 0
                je      d2i_start
                inc     eax
                jmp     d2i_fwd
d2i_start:      xor     ebx, ebx
                mov     esi, 1
d2i_loop:       dec     eax
                cmp     eax, edi
                jl      d2i_return
                mov     ch, [eax]
                sub     ch, '0'
                jc      d2i_fail
                cmp     ch, 10
                jnb     d2i_fail
                push    eax
                movzx   eax, ch
                mul     esi
                add     ebx, eax
                mov     eax, esi
                mov     edx, 10
                mul     edx
                mov     esi, eax
                pop     eax
                jmp     d2i_loop
d2i_return:     mov     eax, ebx
                clc
                ret
d2i_fail:       xor     eax, eax
                stc
                ret
; Hex2In 
; > Input:      EAX - ptr to source string (null-terminated)
; > Output:     EAX - unsigned integer value corresponding to the string
                db      'Hex2In', 6             ; Global Object IDentifier
Hex2In:         mov     edi, eax
h2i_fwd:        cmp     byte ptr [eax], 0
                je      h2i_start
                inc     eax
                jmp     h2i_fwd
h2i_start:      xor     ebx, ebx
                mov     cl, 0
h2i_loop:       dec     eax
                cmp     eax, edi
                jl      h2i_return
                mov     ch, [eax]
h2i_ver_lower:  cmp     ch, 'a'
                jb      h2i_ver_upper
                cmp     ch, 'f'
                ja      h2i_ver_upper
                sub     ch, 'a' - 10
                jmp     h2i_eval
h2i_ver_upper:  cmp     ch, 'A'
                jb      h2i_ver_digit
                cmp     ch, 'F'
                ja      h2i_ver_digit
                sub     ch, 'A' - 10
                jmp     h2i_eval
h2i_ver_digit:  sub     ch, '0'
                jc      h2i_fail
                cmp     ch, 10
                jnb     h2i_fail
h2i_eval:       movzx   edx, ch
                shl     edx, cl
                add     ebx, edx
                add     cl, 4
                jmp     h2i_loop
h2i_return:     mov     eax, ebx
                clc
                ret
h2i_fail:       xor     eax, eax
                stc
                ret
; MountMemory 
; > Input:      (none)
; > Output:     EAX - total free memory in bytes (may be zero)
; > Notes:      This function must be called at least once before locking
;               any memory blocks of any size. It scans the system memory
;               map, searching for the largest block of free and contiguous
;               memory, then stores the base pointer and size of that memory
;               block to "bom" (Beginning Of Memory) and "som" (Size Of Mem.)
KEYMMBASE       EQU               100000h       ; first analyzed address
smap_blockbase  dd                    0,0       ;  SMAP block info (64-bit)
smap_blocksize  dd                    0,0       ;  SMAP block info (64-bit)
smap_blocktype  dd                      0       ;  SMAP block info (32-bit)
smap_maxblsize  dd                      ?       ; largest block for now
bom             dd              KEYMMBASE       ; beginning of memory
eom             dd              KEYMMBASE       ; end of memory (absolute)
som             dd                      0       ; size of memory
tom             dd                      0       ; top of memory (rel. to bom)
                db      'MountMemory', 11       ; Global Object IDentifier
MountMemory:    mov     smap_maxblsize[@cb], 0  ; zero for maximum blk. size
                mov     tom[@cb], 0             ; zero is actual top of memory
                xor     ebx, ebx                ; specify for first block.
mme_nextblock:  mov     eax, 0E820h             ; get memory block info.
                mov     ecx, 00014h             ; result buffer size.
                mov     edx, 'SMAP'             ; 'SMAP' BIOS signature.
                lea     edi, smap_blockbase[@cb]; offset of results buffer.
                int     15h                     ; BIOS call.
                jc      mme_compute             ; (alternate end of mem. map)
                cmp     eax, 'SMAP'             ; function check.
                jne     mme_nosysmap            ; (if not supported)
                cmp     smap_blocktype[@cb], 1  ; is block available for use?
                jne     mme_unavailable         ; (if not available)
                test    smap_blockbase[@cb+4], 0FFFFFFFFh
                jnz     mme_unavailable         ; (if out of address range)
                cmp     smap_blockbase[@cb+0], KEYMMBASE
                jnb     mme_notreserved         ; (if not reserved)
                mov     eax, KEYMMBASE
                sub     eax, smap_blockbase[@cb]
                sub     smap_blocksize[@cb], eax
                jc      mme_unavailable         ; (if entirely blw. KEYMMBASE)
                jz      mme_unavailable         ; (if entirely blw. KEYMMBASE)
                mov     smap_blockbase[@cb], KEYMMBASE
mme_notreserved:test    smap_blocksize[@cb+4], 0FFFFFFFFh
                jz      mme_insizerange         ; (if no larger than 4 Gb)
                mov     smap_blocksize[@cb+0], 0FFFFFFFFh
                mov     eax, smap_blockbase[@cb]; (if larger than 4 Gb)
                sub     smap_blocksize[@cb], eax
mme_insizerange:mov     eax, smap_blockbase[@cb]; (proceed as normal)
                add     eax, smap_blocksize[@cb]
                cmp     eax, eom[@cb]
                jbe     mme_nottheeom           ; (if it's not the new eom)
                mov     eom[@cb], eax
mme_nottheeom:  mov     eax, smap_maxblsize[@cb]; (check if it's the largest)
                cmp     smap_blocksize[@cb], eax
                jb      mme_unavailable         ; (if smaller than actual)
                m2m     smap_maxblsize[@cb], smap_blocksize[@cb]
                m2m     bom[@cb], smap_blockbase[@cb]
                m2m     som[@cb], smap_blocksize[@cb]
mme_unavailable:test    ebx, ebx                ; end of BIOS memory map?
                jnz     mme_nextblock           ; (no)
                jmp     mme_compute             ; (yes)
mme_nosysmap:   jmp     mme_compute             ; <--- incomplete here
mme_compute:    mov     eax, som[@cb]           ; return size of memory
                clc
                ret
; LockMemory 
; > Input:      EAX - needed memory, in bytes
;               EBX - owner's GOID (optional, must be zero if undeclared)
; > Output:     EAX - address to 1st locked byte
;               EBX - remaining free memory in bytes (after locking)
;               ECX - total used memory so far, bytes (after locking)
mt_locks        dd                      0       ; currently locked blocks
mt_iptr         EQU                   @mt       ; top of memory for each block
mt_owner        EQU       @mlocks*4 + @mt       ; ptr to owner's GOID
                db      'Unknown', 7            ; Global Object IDentifier
UnknownGOID:    ret                             ; (for undeclared block owner)
                db      'LockMemory', 10        ; Global Object IDentifier
LockMemory:     test    ebx, ebx                ; check if caller has a GOID
                jnz     lme_knownowner          ; .
                lea     ebx, UnknownGOID[@cb]   ; .
lme_knownowner: mov     ecx, tom[@cb]           ; check for enough memory
                add     ecx, eax                ; .
                jc      lme_fail                ; (if out of range)
                cmp     ecx, som[@cb]           ; .
                ja      lme_fail                ; (if not enough memory)
                cmp     mt_locks[@cb], @mlocks  ; .
                jnb     lme_fail                ; (if end of memory tables)
lme_avail:      mov     edi, mt_locks[@cb]      ; store informations in tables
                m2m     <dword ptr [mt_iptr + edi*4]>, tom[@cb]
                mov     dword ptr [mt_owner + edi*4], ebx
                inc     mt_locks[@cb]           ; (new block has been locked)
                push    tom[@cb]                ; set EAX = pointer to block
                add     tom[@cb], eax           ; and increase top of memory
                pop     eax                     ; .
                add     eax, bom[@cb]           ; .
                mov     ebx, som[@cb]           ; set EBX = remaining free mem
                sub     ebx, tom[@cb]           ; .
                mov     ecx, tom[@cb]           ; set ECX = used memory so far
                clc                             ; (no errors)
                ret                             ; (quit function)
lme_fail:       mov     eax, 7FFFFFFFh          ; (no, fail, but return info)
                mov     ebx, som[@cb]           ; set EBX = remaining free mem
                sub     ebx, tom[@cb]           ; .
                mov     ecx, tom[@cb]           ; set ECX = used memory so far
                stc                             ; (error status)
                ret                             ; (quit function)
; UnlockMemory 
; > Input:      (nothing)
; > Output:     EBX - remaining free memory in bytes (after unlocking)
;               ECX - total used memory so far, bytes (after unlocking)
                db      'UnlockMemory', 12      ; Global Object IDentifier
UnlockMemory:   cmp     mt_locks[@cb], 0        ; is there a block to unlock?
                je      ume_donothing           ; (no, return with error)
                dec     mt_locks[@cb]           ; (yes, go on)
                mov     esi, mt_locks[@cb]      ; get previous top of memory
                m2m     tom[@cb], <dword ptr [mt_iptr + esi*4]>
                mov     ebx, som[@cb]           ; set EBX = remaining free mem
                sub     ebx, tom[@cb]           ; .
                mov     ecx, tom[@cb]           ; set ECX = used memory so far
                clc                             ; (no errors)
                ret                             ; (quit function)
ume_donothing:  mov     ebx, som[@cb]           ; set EBX = remaining free mem
                sub     ebx, tom[@cb]           ; .
                mov     ecx, tom[@cb]           ; set ECX = used memory so far
                stc                             ; (error status)
                ret                             ; (quit function)
; MountVolume 
; > Input:      EAX - system drive ID string (RD01..RD04, MD01..MD08)
; > Output:     EAX:EBX - addressable volume size in bytes
;                         (upto 16 mega-tera-bytes theoretically,
;                          effectively restricted to a maximum of
;                          "only" one Terabytes due to internal maths)
;               ECX - host bus type, can actually be:
;                     '----' - undefined (unaccessible device)
;                     'LISA' - Legacy ISA, 16-bit device
;                     'NSTD' - Non-STanDard, Legacy ISA assumed (be careful!)
;                     'ENHC' - Phoenix EDD, EISA 32-bit device, ENHanCed drive
;               EDX - host interface type, can actually be:
;                     '----' - undefined (no interface informations)
;                     'BEDD' - Phoenix EDD standard interface
;                     'ATA'  - ATA/ATAPI device using ATA commands
;                     'ATAP' - ATA/ATAPI device using ATAPI commands
;                     'SCSI' - SCSI device
;                     'USB'  - USB device
;                     '1394' - 1394 mass device
;                     'FIBR' - FIBRE channel
;                     (future specifications not listed here
;                      will be handled correctly and returned)
;               ESI - host interface path, can be:
;                     'MAST' or 'SLAV' for ATA/ATAPI devices;
;                     'LUNx' Logical Unit Number x, x=0-7, for SCSI;
;                     'TBDx' TBD code x, x=0-9, for USB devices;
;                     '----' unreported (for other device types);
;                     '----' undefined (no interface informations).
;               * on error, device is not mounted, carry flag is set,
;                 and registers layout is as follows...
;                       EAX = 00000000
;                       EBX = 00000000
;                       ECX = 'NSTD' or '----'
;                       EDX = '----'
;                       ESI = '----'
; > Faults:     A - invalid drive ID
;               B - device acknowledged but unaccessible
;               C - volume larger than 2 Tb and thus not supported
;               D - invalid bytes per sector (not a multiple of 256)
;               E - invalid bytes per sector (zero)
;               F - no media in drive
;               G - device reported invalid geometry
;               H - drive read test failed
;               I - no physical sectors were accessible while scanning volume
;               J - no physical heads were accessible while scanning volume
;               K - no physical tracks were accessible while scanning volume
;               L - drive recalibration failure
; --- Structures and constants relating system drives ----------
driveinfo       struc           ; --------------- L.C. Drive Info Block (DIB).
di_drive_id     db      ?       ; offset +0     ; thats an 8-bit BIOS drive ID
di_nroftracks   dd      ?       ; offset +1     ; number of supported cyls.
di_nrofheads    dd      ?       ; offset +5     ; number of heads per cylinder
di_secspertrack dd      ?       ; offset +9     ; sectors per each track
di_secspercall  db      ?       ; offset +13    ; sectors per single BIOS call
di_bytespersec  db      ?       ; offset +14    ; bytes per sector / 256
di_driveflags   db      ?       ; offset +15    ; drive flags, see below...
                ends            ; --------------- internal size: 16 bytes.
di_size         equ     16      ; size of "driveinfo"
di_sizeshift    equ     4       ; log2 (di_size)
df_write_enable equ     1       ; driveflags: drive is write-enable
df_crosstracker equ     2       ; driveflags: drive can R/W cross-track
df_systemdisk   equ     4       ; driveflags: media is bootable
df_enhanced     equ     8       ; driveflags: enhanced drive specifications
vtocinfo        struc           ; --------------- L.C. VTOC Info Block (VIB).
vi_base         dd      ?       ; offset +0     ; base address of the VTOC
                ends            ; --------------- internal size: 16 bytes.
vi_size         equ     4       ; size of "vtocinfo"
vi_sizeshift    equ     2       ; log2 (vi_size)
; --- MDT, Mounted Devices Table. --------------------
DIBr    label   byte
      ;Removable1..4BIOSIDCylsHeadsSectsPerCallBPS/256Flags
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
DIBm    label   byte
      ;Fixed1..8BIOSIDCylsHeadsSectsPerCallBPS/256Flags
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
        driveinfo      < -1,     0,     0,       0,      0,     0,    000000b>
; --- VTOC access optimization table. ------------------
VIBr    label   byte
      ;Removable1..4VTOCbase
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
VIBm    label   byte
      ;Fixed1..8VTOCbase
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
        vtocinfo       < 0 >
; --- Other buffers relating MountVolume and generic drive access ----
mvl_EDDinfo     db      68              ; buffer size (68 max. / 26 min.)
                db      67 dup (0)      ; remainder of EDDinfo structure
mvl_DNSH        db      ?               ; flag: Device Not Supported by H/w?
mvl_current_id  dd      'RD01'          ; device ID string
mvl_dev_size_m  dd      000000          ; device size (MSD)
mvl_dev_size_l  dd      000000          ; device size (LSD)
mvl_dev_host    dd      '----'          ; host interface bus
mvl_dev_type    dd      '----'          ; host interface type
mvl_dev_path    dd      '----'          ; type-specific path
mvl_geoscan_delta dd    000000          ; temporary (drive geometry scanning)
; --- Extended inline DIB. -----------------------
mvl_bytespersec dd      000000          ; log2 (bytes per sector)
mvl_vtocbase    dd      000000          ; base address of the vtoc
mvl_vtocsize    dd      000000          ; size of the vtoc in bytes
mvl_vtocslots   dd      000000          ; number of slots in the vtoc
; --- Internal function translating system drive ID to BIOS ID -----
internal_TranslateDriveID:
                mov     edx, eax                                ;  parse ID
                idup    edx
tid_cv_rd:      cmp     edx, 'RD01'                             ;  check ID
                jb      tid_cv_md                               ;   validity/1
                cmp     edx, 'RD04'                             ;  check ID
                ja      tid_cv_md                               ;   validity/2
                jmp     tid_cv_ok
tid_cv_md:      cmp     edx, 'MD01'                             ;  check ID
                jb      tid_fail                                ;   validity/3
                cmp     edx, 'MD08'                             ;  check ID
                ja      tid_fail                                ;   validity/4
               ;jmp     tid_cv_ok
tid_cv_ok:      sub     dx, '01'                                ;  convert
                movzx   eax, dh                                 ;   ID-string
                add     eax, eax                                ;   (EDX) to
                lea     eax, [eax * 4 + eax]                    ;   BIOS-ID
                add     al, dl                                  ;   and store
                movzx   edi, al                                 ;   resulting
                shl     edi, di_sizeshift                       ;   code in AL
                shr     edx, 16                                 ;   while the
                cmp     dx, 'MD'                                ;   DIB offset
                jne     tid_success                             ;   from"DIBr"
                or      al, 80h                                 ;   is held
                add     edi, 4 * di_size                        ;   in EDI.
tid_success:    clc
                ret
tid_fail:       stc
                ret
; --- MountVolume ----------------------------
                db      'MountVolume', 11       ; Global Object IDentifier
MountVolume:    status  'm'
                status  'v'
                idup    eax
                dmon    eax, al
                mov     mvl_current_id[@cb], eax                ;  preserve
                mov     mvl_dev_size_m[@cb], 0000               ;  clear
                mov     mvl_dev_size_l[@cb], 0000               ;   dev. size
                mov     mvl_dev_host[@cb], '----'               ;  clear HB
                mov     mvl_dev_type[@cb], '----'               ;  clear DT
                mov     mvl_dev_path[@cb], '----'               ;  clear DP
                call    internal_TranslateDriveID               ;  translate
                mov     endstatus[@cb], 'A'                     ;  FAULT 
                jc      mvl_unsupported                         ;  invalid ID
                mov     DIBr[@cb + edi + 00], al                ;  BIOSIDcode
                mov     byte ptr DIBr[@cb + edi + 15], 00       ;  clr flags
                cmp     edi, offset DIBm - offset DIBr          ;  init flags
                jnb     mvl_fixed                               ; (lock drive)
                or      byte ptr DIBr[@cb + edi + 15], df_write_enable
mvl_fixed:      mov     dl, al                                  ;  check if
                mov     bx, 55AAh                               ;   extension
                mov     ax, 4100h                               ;   is present
                int     13h                                     ;   (BIOS EDD)
                cmp     bx, 0AA55h                              ;  confirm?
                jne     mvl_EDD_chfail                          ;  if not...
                test    cx, 001b                                ;  EDDaccess
                mov     endstatus[@cb], 'B'                     ;  FAULT 
                jz      mvl_unsupported                         ;  if not...
mvl_EDD_chfail: mov     dl, DIBr[@cb + edi + 00]                ;  BIOS ID
                mov     ax, 4800h                               ;   EDD infos
                lea     esi, mvl_EDDinfo[@cb]                   ;   sets SI
                mov     word ptr mvl_EDDinfo[@cb + 0], 68       ;   set size
                int     13h                                     ;   (BIOS EDD)
                jc      mvl_non_EDD                             ;  if failed
                test    word ptr mvl_EDDinfo[@cb + 2], 0000010b ;  CHS valid?
                jnz     mvl_validchs                            ;   (yes)
                test dword ptr mvl_EDDinfo[@cb + 20], 0FFFFFFFFh;  LSD of SPT
                jnz     mvl_unsupported                         ;   (if >2Tb)
                mov     endstatus[@cb], 'C'                     ;  FAULT 
                mov     eax, dword ptr mvl_EDDinfo[@cb + 16]    ;  get/store
                mov     dword ptr DIBr[@cb + edi + 09], eax     ;  sectors
                mov     dword ptr DIBr[@cb + edi + 01], 1       ; (fake)tracks
                mov     dword ptr DIBr[@cb + edi + 05], 1       ; (fake)heads
                mov     byte ptr DIBr[@cb + edi + 13], 127      ; (max)PerCall
                movzx   eax, word ptr mvl_EDDinfo[@cb + 24]     ;  get bps
                shr     eax, 8                                  ;  bps / 256
                mov     byte ptr DIBr[@cb + edi + 14], al       ;  store bps
                or byte ptr DIBr[@cb + edi + 15],df_enhanced    ; (default)
                or byte ptr DIBr[@cb + edi + 15],df_crosstracker; (default)
                jmp     mvl_fakegeomtry                         ; minimal geo.
mvl_validchs:   test    word ptr mvl_EDDinfo[@cb + 2], 1000000b ;  Disk in?
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                jnz     mvl_unsupported                         ;   (no)
                or byte ptr DIBr[@cb + edi + 15],df_enhanced    ;  OK for 1.1
                or byte ptr DIBr[@cb + edi + 15],df_crosstracker;  (default)
                mov     ax, word ptr mvl_EDDinfo[@cb + 24]      ;  get BPS
                test    al, al                                  ;   multiple?
                mov     endstatus[@cb], 'D'                     ;  FAULT 
                jnz     mvl_unsupported                         ;   (no)
                shr     ax, 8                                   ;   (/ 256)
                mov     endstatus[@cb], 'E'                     ;  FAULT 
                jz      mvl_unsupported                         ;   (if zero)
                mov     DIBr[@cb + edi + 14], al
                push    dword ptr mvl_EDDinfo[@cb + 04]         ;  get tracks
                pop     dword ptr DIBr[@cb + edi + 01]
                inc     dword ptr DIBr[@cb + edi + 01]
                push    dword ptr mvl_EDDinfo[@cb + 08]         ;  get heads
                pop     dword ptr DIBr[@cb + edi + 05]
                inc     dword ptr DIBr[@cb + edi + 05]
                push    dword ptr mvl_EDDinfo[@cb + 12]         ;  get SPT
                pop     dword ptr DIBr[@cb + edi + 09]
                mov     eax, dword ptr DIBr[@cb + edi + 09]     ;  PerCall
                cmp     eax, 128                                ;(always re-
                jb      mvl_percall_ok                          ; normalized
                mov     eax, 127                                ; to max 127)
mvl_percall_ok: mov     DIBr[@cb + edi + 13], al                ; (secs/call)
mvl_fakegeomtry:mov     mvl_dev_host[@cb], 'EISA'               ;  defaults
                mov     mvl_dev_type[@cb], 'BEDD'               ;  for 1.1
                cmp     word ptr mvl_EDDinfo[@cb + 30], 0BEDDh  ;  dev info?
                jne     mvl_return                              ;   (no)
                mov     eax, dword ptr mvl_EDDinfo[@cb + 36]    ;  get host
                bswap   eax
                mov     dword ptr mvl_dev_host[@cb], eax
                mov     eax, dword ptr mvl_EDDinfo[@cb + 40]    ;  get interf
                bswap   eax
                mov     dword ptr mvl_dev_type[@cb], eax
                mov     al, mvl_EDDinfo[@cb + 56]               ;  get path
                cmp     mvl_dev_type[@cb], 'SCSI'
                jne     mvl_non_scsi
                mov     mvl_dev_path[@cb], 'LUN0'
                add     byte ptr mvl_dev_path[@cb + 3], al
                jmp     mvl_return                              ; SCSI mounted
mvl_non_scsi:   cmp     word ptr mvl_dev_type[@cb], 'TA'
                jne     mvl_non_ata
                mov     mvl_dev_path[@cb], 'MAST'
                test    al, 1
                jz      mvl_return
                mov     mvl_dev_path[@cb], 'SLAV'
                jmp     mvl_return                              ; ATA mounted
mvl_non_ata:    cmp     word ptr mvl_dev_type[@cb], 'SU'
                jne     mvl_return
                mov     mvl_dev_path[@cb], 'TBD0'
                add     byte ptr mvl_dev_path[@cb + 3], al                
                jmp     mvl_return                              ; USB mounted
mvl_non_EDD:    mov     mvl_dev_host[@cb], 'LISA'               ; (default)
                mov     ax, 0800h                               ;  get params
                mov     dl, DIBr[@cb + edi + 00]                ;   (BIOS DDS)
                push    edi                                     ;  (save)
                int     13h                                     ;  inquire
                pop     edi                                     ;  (restore)
                mov     mvl_DNSH[@cb], 1                        ; * set flag
                jc      mvl_non_std                             ;  on error
                mov     mvl_DNSH[@cb], 0                        ; * clr. flag
                test    ch, ch                                  ;  (see note)
                mov     endstatus[@cb], 'G'                     ;  FAULT 
                jz      mvl_unsupported                         ;  on error
                movzx   eax, ch                                 ;  get tracks
                movzx   ebp, cl                                 ;   (max 1024)
                and     ebp, 11000000b
                shl     ebp, 2
                or      eax, ebp
                mov     dword ptr DIBr[@cb + edi + 01], eax
                inc     dword ptr DIBr[@cb + edi + 01]
                movzx   eax, dh                                 ;  get heads
                test    eax, eax                                ;   (0 = 256)
                jnz     mvl_lt256heads
                mov     eax, 256
mvl_lt256heads: mov     dword ptr DIBr[@cb + edi + 05], eax
                inc     dword ptr DIBr[@cb + edi + 05]
                movzx   eax, cl                                 ;  get SPT..
                and     eax, 00111111b                          ; = percall
                mov     dword ptr DIBr[@cb + edi + 09], eax
                mov     DIBr[@cb + edi + 13], al
                mov     DIBr[@cb + edi + 14], byte ptr 02       ;  BPS = 512
                test    DIBr[@cb + edi + 00], 80h               ;  harddrive:
                jnz     mvl_return                              ;   stop here.
mvl_non_std:    test    mvl_DNSH[@cb], 1                        ; is for this?
                jz      mvl_justafloppy                         ; no
                mov     mvl_dev_host[@cb], 'NSTD'               ; non-std bus
mvl_justafloppy:xor     eax, eax                                ; Cyl0Hd0Sec1
                call    mvl_test_sect                           ; * dummy read
                mov     endstatus[@cb], 'H'                     ;  FAULT 
                jc      mvl_unsupported                         ; cannot read!
                cmp     dword ptr ds:[@io + 20h], 'VPBX'        ; (signature)
                jne     mvl_nofmtdisk                           ; is formated?
;        When non-standard disk has been already formatted, there's no need
;         to extensively scan its geometry, so it gathers informations from
;         the bootstrap loader (all formated volumes are bootable under LC).
                or      DIBr[@cb + edi + 15], df_systemdisk     ; IS FORMATED
                mov     eax, dword ptr ds:[@io + 006h]          ; GET TRACKS
                mov     dword ptr DIBr[@cb + edi + 01], eax
                mov     eax, dword ptr ds:[@io + 00Eh]          ; GET SPT
                mov     dword ptr DIBr[@cb + edi + 09], eax     ;(percall is
                cmp     eax, 127                                ; normalized
                jbe     mvl_ns_lt128                            ; to max 127)
                mov     eax, 127
mvl_ns_lt128:   mov     byte ptr DIBr[@cb + edi + 13], al
                mov     eax, dword ptr ds:[@io + 00Ah]          ; GET HEADS
                mov     dword ptr DIBr[@cb + edi + 05], eax
                mov     cl, byte ptr ds:[@io + 012h]            ; GET the...
                mov     eax, 1                                  ; log2 (bps)
                shl     eax, cl                                 ; and convert
                shr     eax, 8                                  ; to BPS / 256
                mov     byte ptr DIBr[@cb + edi + 14], al       ; ...store
;        If it's a system disk (if disk is formated for access by LC),
;         the VTOC base address is stored in the optimization table (case #1).
                shr     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                mov     eax, dword ptr ds:[@io + 14h]           ; get VTOC b.
                mov     dword ptr VIBr[@cb + edi + 00], eax     ; set VTOC b.
                shl     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                jmp     mvl_sub_mounted                         ; OK, ALL DONE
;        Scan for diskette media type: initialization.
;        Set initial Sectors Per Call to 1, BPS fixed to 512.
mvl_nofmtdisk:  mov     DIBr[@cb + edi + 13], byte ptr 01
                mov     DIBr[@cb + edi + 14], byte ptr 02
;        Scan for diskette media type: part 1.
;        Find number of sectors per head / track.
                mov     eax, 32                                 ;  stepping
                mov     mvl_geoscan_delta[@cb], eax             ;  step value
                mov     dword ptr DIBr[@cb + edi + 09], 0       ;  zero base
mvl_sect_scan:  mov     eax, mvl_geoscan_delta[@cb]             ;  (re)start
                add     eax, dword ptr DIBr[@cb + edi + 09]     ;   scanning
                mov     dword ptr DIBr[@cb + edi + 09], eax     ;   sectors
mvl_sect_ok:    mov     eax, dword ptr DIBr[@cb + edi + 09]     ; Cyl0Hd0SecN
                call    mvl_test_sect                           ; * dummy read
                jc      mvl_no_sect                             ; read error
                mov     eax, dword ptr DIBr[@cb + edi + 09]     ;  expand
                add     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 09], eax     ;   by step.
                jmp     mvl_sect_ok                             ;  run again
mvl_no_sect:    mov     eax, dword ptr DIBr[@cb + edi + 09]     ;  shrink
                sub     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 09], eax     ;   to above.
                shr     mvl_geoscan_delta[@cb], 1               ;  halve step
                cmp     mvl_geoscan_delta[@cb], 0               ;   and retry
                jne     mvl_sect_scan                           ;   until 0.
                inc     dword ptr DIBr[@cb + edi + 09]          ; * END SCAN
                cmp     dword ptr DIBr[@cb + edi + 09], 0       ; * verify
                mov     endstatus[@cb], 'I'                     ;  FAULT 
                je      mvl_unsupported                         ; * invalid
                mov     eax, dword ptr DIBr[@cb + edi + 09]
                mov     DIBr[@cb + edi + 13], al
;        Scan for diskette media type: part 2.
;        Find number of heads.
                mov     eax, 32                                 ;  stepping
                mov     mvl_geoscan_delta[@cb], eax             ;  step value
                mov     dword ptr DIBr[@cb + edi + 05], 0       ;  zero base
mvl_head_scan:  mov     eax, mvl_geoscan_delta[@cb]             ;  (re)start
                add     eax, dword ptr DIBr[@cb + edi + 05]     ;   scanning
                mov     dword ptr DIBr[@cb + edi + 05], eax     ;   heads
mvl_head_ok:    mov     eax, dword ptr DIBr[@cb + edi + 05]     ; Cyl0HdNSec1
                call    mvl_test_head                           ; * dummy read
                jc      mvl_no_head                             ; read error
                mov     eax, dword ptr DIBr[@cb + edi + 05]     ;  expand
                add     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 05], eax     ;   by step.
                jmp     mvl_head_ok                             ;  run again
mvl_no_head:    mov     eax, dword ptr DIBr[@cb + edi + 05]     ;  shrink
                sub     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 05], eax     ;   to above.
                shr     mvl_geoscan_delta[@cb], 1               ;  halve step
                cmp     mvl_geoscan_delta[@cb], 0               ;   and retry
                jne     mvl_head_scan                           ;   until 0.
                inc     dword ptr DIBr[@cb + edi + 05]          ; * END SCAN
                cmp     dword ptr DIBr[@cb + edi + 05], 0       ; * verify
                mov     endstatus[@cb], 'J'                     ;  FAULT 
                je      mvl_unsupported                         ; * invalid
;        Scan for diskette media type: part 3.
;        Find number of tracks.
mvl_nostdnoext: mov     eax, 512                                ;  stepping
                mov     mvl_geoscan_delta[@cb], eax             ;  step value
                mov     dword ptr DIBr[@cb + edi + 01], 0       ;  zero base
mvl_track_scan: mov     eax, mvl_geoscan_delta[@cb]             ;  (re)start
                add     eax, dword ptr DIBr[@cb + edi + 01]     ;   scanning
                mov     dword ptr DIBr[@cb + edi + 01], eax     ;   tracks
mvl_track_ok:   mov     eax, dword ptr DIBr[@cb + edi + 01]     ; CylNHd0Sec1
                call    mvl_test_cyl                            ; * dummy read
                jc      mvl_no_track                            ; read error
                mov     eax, dword ptr DIBr[@cb + edi + 01]     ;  expand
                add     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 01], eax     ;   by step.
                jmp     mvl_track_ok                            ;  run again
mvl_no_track:   mov     eax, dword ptr DIBr[@cb + edi + 01]     ;  shrink
                sub     eax, mvl_geoscan_delta[@cb]             ;   geometry
                mov     dword ptr DIBr[@cb + edi + 01], eax     ;   to above.
                shr     mvl_geoscan_delta[@cb], 1               ;  halve step
                cmp     mvl_geoscan_delta[@cb], 0               ;   and retry
                jne     mvl_track_scan                          ;   until 0.
                inc     dword ptr DIBr[@cb + edi + 01]          ; * END SCAN
                cmp     dword ptr DIBr[@cb + edi + 01], 0       ; * verify
                mov     endstatus[@cb], 'K'                     ;  FAULT 
                je      mvl_unsupported                         ; * invalid
                jmp     mvl_return                              ; * else, ok!
; --- Internal procedures using ISA services to test drives geometry. --
;        1) Performs a dummy read from currently examined drive (DIBr+00):
;               Cylinder EAX
;               Head     zero
;               Sector   one
mvl_test_cyl:   cmp     eax, 1024                               ; (ISA maximum
                jnb     mvl_fail_test                           ; cyl. = 1023)
                mov     cl, 01h                                 ; (sector one)
                mov     dh, 00h                                 ; (head zero)
                mov     ch, al                                  ; (bits 0..7)
                mov     dl, ah                                  ; 
                shl     dl, 6                                   ; (bits 8/9)
                or      cl, dl                                  ; 
                jmp     mvl_test_now                            ;(common code)
;        2) Performs a dummy read from currently examined drive (DIBr+00):
;               Cylinder zero
;               Head     EAX
;               Sector   one
mvl_test_head:  cmp     eax, 256                                ; (ISA maximum
                jnb     mvl_fail_test                           ; head = 255)
                mov     ch, 00h                                 ; (track zero)
                mov     cl, 01h                                 ; (sector one)
                mov     dh, al                                  ; (curr. head)
                jmp     mvl_test_now                            ;(common code)
;        3) Performs a dummy read from currently examined drive (DIBr+00):
;               Cylinder zero
;               Head     zero
;               Sector   EAX + 1
mvl_test_sect:  cmp     eax, 64                                 ; (ISA maximum
                jnb     mvl_fail_test                           ; sector = 63)
                mov     ch, 00h                                 ; (track zero)
                mov     dh, 00h                                 ; (head zero)
                mov     cl, al                                  ; (curr. sect)
                inc     cl                                      ; (plus / one)
;        Common code for tests 1..3, invoking ISA disk services.
;        Reads 1 sector, retries if disk-change line signal asserted.
;        Floppies are recalibrated before reading, hard disks are not.
;        (I have met an HD controller that crashed everything when the
;         disk system - both floppies and hard drives - were recalibrated
;         just before reading from them. I suppose it's a bug, or maybe
;         an interference, anyway hard drives should rarely need to be
;         recalibrated, and many HD controllers just ignore that call.)
mvl_test_now:   test    byte ptr DIBr[@cb + edi + 00], 80h      ;  HD to skip
                jnz     mvl_isharddisk                          ;recalibration
                pushad
                mov     ah, 00h                                 ;(recalibrate)
                mov     dl, byte ptr DIBr[@cb + edi + 00]       ; (curr.drive)
                push    edi                                     ;  (save)
                int     13h                                     ;  (execute)
                pop     edi                                     ;  (restore)
                cmp     ah, 00h                                 ; confirm code
                je      mvl_recal_ok                            ; (success)
                popad                                           ; calibration
                pop     eax                                     ; failure:
                mov     endstatus[@cb], 'L'                     ; fail whole
                jmp     mvl_unsupported                         ; function!
mvl_recal_ok:   popad                                           ; (calibr. ok)
mvl_isharddisk: mov     dl, byte ptr DIBr[@cb + edi + 00]       ; (curr.drive)
                mov     ax, 0201h                               ; (read 1 sec)
                mov     word ptr ds:[@rmes], @seg_io            ; (R-M buffer)
                xor     bx, bx                                  ; (R-M offset)
                push    edi                                     ;  (save)
                int     13h                                     ;  (execute)
                pop     edi                                     ;  (restore)
                mov     word ptr ds:[@rmes], 0                  ; (restore ES)
                cmp     ah, 00h                                 ; confirm code
                je      mvl_test_ok                             ; (success)
                cmp     ah, 11h                                 ;corrected ECC
                je      mvl_test_ok                             ; (success)
                cmp     ah, 06h                                 ; disk change
                je      mvl_test_now                            ; (retry)
mvl_fail_test:  stc                                             ; (failed)
                ret                                             ; (return)
mvl_test_ok:    clc                                             ; (ok)
                ret                                             ; (return)
;        At logical level, drive and media informations have been found.
;         Now physically test the disk, and also see if it's a system disk.
;         Such a test is performed before, but only for a non-standard bus
;         disk, such as floppies and particular hard drives which are neither
;         listed by legacy function 08h of the BIOS int 13h, nor by the EDD.
mvl_return:     pushad                                          ;
                mov     quietmode[@cb], 1                       ; this will
                xor     eax, eax                                ; perform
                xor     ebx, ebx                                ; a read on
                mov     ecx, 512                                ; the first
                mov     edx, mvl_current_id[@cb]                ; 512 bytes
                mov     esi, 'READ'                             ; of the
                mov     edi, @io                                ; media, in
                call    VolumeIO                                ; quiet mode
                popad                                           ;
                mov     endstatus[@cb], 'H'                     ;  FAULT 
                jc      mvl_unsupported                         ; cannot read!
                cmp     dword ptr ds:[@io + 20h], 'VPBX'        ; (signature)
                jne     mvl_sub_mounted                         ; is formated?
                or      DIBr[@cb + edi + 15], df_systemdisk     ; (if yes)
;        If it's a system disk (if disk is formated for access by LC),
;         the VTOC base address is stored in the optimization table (case #2).
                shr     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                mov     eax, dword ptr ds:[@io + 14h]           ; get VTOC b.
                mov     dword ptr VIBr[@cb + edi + 00], eax     ; set VTOC b.
                shl     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                jmp     mvl_sub_mounted
;        This sub-function is entered HERE (NEAR CALL) with EDI = DIB ptr.
;         to find extended informations without effectively re-mounting.
sub_extend_DIB: mov     quietmode[@cb], 1                       ; nonuser call
                cmp     byte ptr DIBr[@cb + edi + 00], -1       ; is mounted?
                je      mvl_unsupported                         ; (if not)
mvl_sub_mounted:mov     eax, dword ptr DIBr[@cb + edi + 01]     ;  Tracks
                mov     ebx, dword ptr DIBr[@cb + edi + 05]     ;  Heads
                mov     ecx, dword ptr DIBr[@cb + edi + 09]     ;  Sectors
                xor     edx, edx                                ;  (clr rem.)
                mul     ecx                                     ;  Tks*Secs
                xor     edx, edx                                ;  (clr rem.)
                mul     ebx                                     ;  Tks*Sec*Hd
                movzx   edx, byte ptr DIBr[@cb + edi + 14]      ; BPS / 256
                shl     edx, 8                                  ; BPS
                mov     ebx, 1
                mov     ecx, 0
mvl_findlog:    cmp     edx, ebx
                je      mvl_foundlog
                shl     ebx, 1
                inc     ecx
                jmp     mvl_findlog
mvl_foundlog:   mov     mvl_bytespersec[@cb], ecx               ;  log2 (BPS)
                mov     ebx, eax                                ; EAX:EBX=size
                xor     eax, eax
                shld    eax, ebx, cl
                shl     ebx, cl
                cmp     eax, 256                                ;  Top bound
                mov     endstatus[@cb], 'C'                     ;  FAULT 
                jnb     mvl_unsupported                         ; My My! >2Tb?
                mov     ecx, eax                                ;  Zero
                or      ecx, ebx                                ;   geometry
                mov     endstatus[@cb], 'G'                     ;  FAULT 
                jz      mvl_unsupported                         ;   check ...
                mov     mvl_dev_size_m[@cb], eax                ; store MSD
                mov     mvl_dev_size_l[@cb], ebx                ; store LSD
                shr     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                mov     eax, dword ptr VIBr[@cb + edi + 00]     ; get VTOC b.
                mov     mvl_vtocbase[@cb], eax                  ; set VTOC b.
                shl     edi, di_sizeshift - vi_sizeshift        ; conv. ptr.
                mov     eax, mvl_dev_size_m[@cb]                ;  calculate
                mov     ebx, mvl_dev_size_l[@cb]                ;   vtoc size
                shrd    ebx, eax, 9                             ;   in bytes
                mov     mvl_vtocsize[@cb], ebx                  ;   and n. of
                shrd    ebx, eax, 4                             ;   slots in
                mov     mvl_vtocslots[@cb], ebx                 ;   the vtoc
; --- Set registers to returned values and leave. ------------
mvl_report:     mov     eax, mvl_dev_size_m[@cb]
                mov     ebx, mvl_dev_size_l[@cb]
                mov     ecx, mvl_dev_host[@cb]
                mov     edx, mvl_dev_type[@cb]
                mov     esi, mvl_dev_path[@cb]
                dmon____clear
                end_____fn                                      ; ok, mounted!
; --- Error: unsupported drive type or no media in drive. --------
mvl_unsupported:mov     byte ptr DIBr[@cb + edi + 00], -1
                mov     eax, mvl_dev_size_m[@cb]
                mov     ebx, mvl_dev_size_l[@cb]
                mov     ecx, mvl_dev_host[@cb]
                mov     edx, mvl_dev_type[@cb]
                mov     esi, mvl_dev_path[@cb]
                dmon____clear
                fail____fn                                      ; not mounted!
; VolumeIO 
; > Input:      EAX:EBX - Progressive Byte Number (PBN) based at zero.
;                       * EAX - most significant dword of the PBN;
;                       * EBX - least significant dword of the PBN.
;               ECX - number of BYTES to read/write (upto 4 Gb per call)
;                   * a value of zero for ECX returns successfully,
;                     but no data is transferred to/from volume.
;               EDX - target volume's drive ID string.
;               ESI - operation code.
;                   * must be set to either 'READ' or 'WRIT';
;                   * 'WRIT' sent to ROM or locked drives will fail;
;                   * invalid and null codes cause errors.
;               EDI - FLAT address of source/destination memory areas.
;                   * EDI points to source area if ESI='WRIT'
;                   * EDI points to destination area if ESI='READ'
; > Output: (nothing)
; > Faults:     A - invalid operation code
;               B - invalid drive ID
;               C - volume not mounted
;               D - write request for locked volume
;               E - internal problem
;               F - disk changed (please re-mount)
;               G - unrecoverable disk error (oh no!)
;               H - drive recalibration failure
COMMENT 

                        Byte-oriented Volume I/O scheme

           PBN
            
             (BTR - Bytes This Run) Ŀ
     
        |      |      |      |      |      |
        12Ĵ      |      |      34Ĵ
        |      |      |      |      |      |
     
        SECTOR Nsectr n+1sectr n+2sectr n+3SECTOR M
  STT = BPS+BPS+BPS+BPS+BPS / BPS

        1 - This amount of bytes is called "drift";
        2 - This is the "fraction";
        3 - This is the "remainder" of the disk transfer;
        4 - This is the "exclusion";
        N - Number of first sector needed;
        M - Number of last sector needed.

        PBN - Progressive byte number
        BTR - Bytes This Run, amount of bytes to read or write
        STT - Sectors To Transfer
        BPS - Bytes Per Sector
        mod - Arithmetics modulus, remainder of integer division

    * At this point, the following equates and pseudo-code are used
      to clip the sectors so that we can avoid sectors wraparound:

        N               = PBN / BPS
        M               = (PBN + BTR) / BPS
        drift           = (PBN) mod (BPS)
        fraction        = (BPS - drift) mod (BPS)
        remainder       = (PBN + BTR) mod (BPS)
        exclusion       = (BPS - remainder) mod (BPS)

        STT             = (BTR - fraction - remainder) / BPS
                        if (M > N) then
                                if (fraction > 0) then STT++
                                if (remainder > 0) then STT++
                        else
                                STT = 1
                        fi

    * Note that due to those calculations:
        When <drift> equals zero, <fraction> ALSO equals zero;
        when <remainder> equals zero, <exclusion> ALSO equals zero.

    * Here's the pseudo-code performing a byte-oriented "read" operation
      (steps in this algorythm must be performed in this exact order):

        1) * set current sector number (S) = N;
           * set remaining sectors count = STT;
        2) * subtract <chunk sectors> from remaining sectors count;
           * read <chunk sectors> from sector S;
        3) if (STT == 1)
                * copy <BTR> bytes from <buffer ptr + drift>;
                * STOP;
           fi
        4) if (fraction > 0)
                * copy <chunksize - fraction> bytes from <buffer ptr + drift>;
                * add <chunksize - fraction> to destination pointer;
                * add <chunk sectors> to S;
                * set fraction = 0;
                * GO TO (2);
           fi
        5) if (remaining sectors == 0)
                * copy <chunksize - exclusion> bytes from <buffer ptr>;
                * STOP;
           fi
        6) * copy <chunksize> bytes from <buffer ptr>;
           * add <chunksize> to destination pointer;
           * add <chunk sectors> to S;
        7) GO TO (2).

    * For a "write" operation, the initial "read" must be performed anyway,
      to initialize the I/O buffer with all of the possibly unmodified bytes,
      unless the conditions described in the code below are met (that's to
      optimize transfers of large amounts of data) - Then, all data transfers
      (whenever it says "copy") are reversed and go from the destination
      pointer (being the source pointer now) to the I/O buffer - Finally,
      after each data transfer, the sectors intially read are written back
      to the disk. For better clearance, here's the code. Pay attention to
      the conditions that force a "disk read", described at step 2.

        1) * set current sector number (S) = N;
           * set remaining sectors count = STT;
        2) * subtract <chunk sectors> from remaining sectors count;
           if (fraction != 0) or...
              [(remaining sectors == 0) and (remainder != 0)] then...
                * read <chunk sectors> from sector S;
           fi
        3) if (STT == 1)
                * copy <BTR> bytes to <buffer ptr + drift>;
                * write data chunk back;
                * STOP;
           fi
        4) if (fraction > 0)
                * copy <chunksize - fraction> bytes to <buffer ptr + drift>;
                * write data chunk back;
                * add <chunksize - fraction> to source pointer;
                * add <chunk sectors> to S;
                * set fraction = 0;
                * GO TO (2);
           fi
        5) if (remaining sectors == 0)
                * copy <chunksize - exclusion> bytes to <buffer ptr>;
                * write data chunk back;
                * STOP;
        6) * copy <chunksize> bytes to <buffer ptr>;
           * write data chunk back;
           * add <chunksize> to source pointer;
           * add <chunk sectors> to S;
        7) GO TO (2).


; --- Buffers relating VolumeIO and generic drive access --------
                                ; Inline Drive Info Block, initialized by
                                ; "VolumeIO" before performing reads/writes.
                                ; -------------------------------------------
drive_id        db      00h     ; Current drive ID (bios code).
nroftracks      dd      01h     ;    -      -   number of tracks.
nrofheads       dd      01h     ;    -      -   number of heads (or surfaces).
secspertrack    dd      01h     ;    -      -   sectors/track (single head).
secspercall     db      01h     ; Max. sectors to read at once (upto 64Kb).
bytespersec     db      01h     ;    -      -   bytes/sector (in 256b units).
driveflags      db  000101b     ; Drive status bitmask (see top of file).
                                ; -------------------------------------------
                                ; Inline Drive Info Block extension.
                                ; -------------------------------------------
lbytespersec    db      08h     ; log2(bytespersec): bitwise multiplier
secspercyl      dd      01h     ; nroftracks * nrofheads
mediasize       dq      256     ; nroftracks * secspercyl * bytes per sector
                                ; -------------------------------------------
                                ; Target and operation informations.
                                ; -------------------------------------------
iosectors       dd      0       ; last disk i-o operation: number of sectors.
iobytes         dd      0       ;   -    -          -      number of bytes.
pbn             label   qword   ; PBN code
pbn?l           dd      0       ; Progressive Byte Number, less significant DW
pbn?m           dd      0       ; Progressive Byte Number, most significant DW
vio_drive_id    dd      ?       ; temporarily holds source/target drive id
vio_attempts    dd      2       ; number of rd/wr attempts before giving up
; --- Internal function to recalibrate a drive, providing it's mounted. -
internal_RecalibrateVolume:
                pushad
                call    internal_TranslateDriveID               ;  translate
                jc      vrc_error                               ;  invalid ID
                cmp     DIBr[@cb + edi + 00], byte ptr 0FFh     ;  check DIB
                je      vrc_error                               ;  n/mounted
vrc_retry:      mov     dl, DIBr[@cb + edi + 00]                ;  BIOS ID
                mov     ah, 00h                                 ; =recalibrate
                int     13h                                     ;  execute
                cmp     ah, 00h                                 ; confirm code
                je      vrc_ok                                  ; success
                jmp     vrc_error                               ;unrecoverable
vrc_ok:         popad
                clc                                             ; ok, return
                ret
vrc_error:      popad
                stc                                             ; failed
                ret
                db      'VolumeIO', 8           ; Global Object IDentifier
; --- Check opcode validity. ----------------------
VolumeIO:       status  'i'
                status  'o'
                strup   esi
                cmp     esi, 'READ'
                je      vio_opc_valid
                cmp     esi, 'WRIT'
                je      vio_opc_valid
                mov     endstatus[@cb], 'A'                     ;  FAULT 
                jmp     vio_exit
; --- Preserve parameters. -----------------------
vio_opc_valid:  mov     pbn?m[@cb], eax                         ; PBN - MSD
                mov     pbn?l[@cb], ebx                         ; PBN - LSD
                mov     aux?0[@cb], ecx                         ; bytes
                mov     aux?8[@cb], esi                         ; operation
                mov     aux?3[@cb], edi                         ; dest. ptr.
                mov     vio_drive_id[@cb], edx                  ; drive ID
                idup    edx
                dmon    edx, dl
; --- Checks if user provided valid drive ID string, and translate it. -
                mov     eax, edx                                ;  pass ID
                call    internal_TranslateDriveID               ;  translate
                mov     endstatus[@cb], 'B'                     ;  FAULT 
                jc      vio_exit                                ;  invalid ID
                lea     ebp, DIBr[@cb + edi + 00]               ;  DIBpointer
; --- Copy drive info block within extended in-line area. --------
                cmp     DIBr[@cb + edi + 00], byte ptr 0FFh     ;  check DIB
                mov     endstatus[@cb], 'C'                     ;  FAULT 
                je      vio_exit                                ;  n/mounted
                lea     esi, DIBr[@cb + edi]
                lea     edi, drive_id[@cb]
                mov     ecx, di_size
                rep     movsb                                   ; copy DIB...
; --- Checks if drive info block states drive is writable (if needed). -
                test    driveflags[@cb], df_write_enable
                jnz     vio_ext_maths
                cmp     aux?8[@cb], 'READ'
                je      vio_ext_maths
                mov     endstatus[@cb], 'D'                     ;  FAULT 
                jmp     vio_exit
; --- Compute in-line DIB extended data. ----------------
;  Calculates log2 of (Bytes Per Sector). BPS are already in units of 256b;
;  Calculates Sectors Per Cylinder (being Sectors Per Track * Nr. of Heads);
;  Calculates media size (being Cyls * Sectors Per Cyl * BPS).
vio_ext_maths:  movzx   eax, bytespersec[@cb]
                shl     eax, 8
                mov     ebx, 1
                mov     ecx, 0
vio_findlog:    cmp     eax, ebx
                je      vio_foundlog
                shl     ebx, 1
                inc     ecx
                jmp     vio_findlog
vio_foundlog:   mov     lbytespersec[@cb], cl
                mov     eax, secspertrack[@cb]
                mov     edx, nrofheads[@cb]
                mul     edx
                mov     secspercyl[@cb], eax
                mov     edx, nroftracks[@cb]
                mul     edx
                mov     ebx, eax
                xor     eax, eax
                shld    eax, ebx, cl
                shl     ebx, cl
                mov     dword ptr mediasize[@cb], ebx
                mov     dword ptr mediasize[@cb + 4], eax
; --- Before you begin, check if this will be a null read or write. ---
                cmp     aux?0[@cb], 0                           ; zero bytes?
                jz      vio_done                                ; (yes, done)
; --- Adjust number of needed sectors to approximate nr. of bytes. ---
;        As from the byte-oriented R/W schemes, here's a quick reference:
;       * PBN                                                   [pbn?m/pbn?l]
;       * BTR                                                   [AUX?0]
;       ----------------------------------------------------------------------
;       * Drift     = (PBN) mod (BPS)                           [AUX?7]
;       * Fraction  = BPS - Drift                               [AUX?A]
;       * Remainder = (PBN + BTR) mod (BPS)                     [AUX?E]
;       * Exclusion = BPS - Remainder                           [AUX?2]
;       * N = LBA   = PBN / BPS                                 [AUX?1]
;       * M         = (PBN + BTR) / BPS                         [AUX?B]
;       * STT       = (BTR - FRACTION - REMAINDER) / BPS        [AUX?C]
;       * RSC       = STT                                       [AUX?D]
;       ----------------------------------------------------------------------
vio_maths:      mov     cl, lbytespersec[@cb]                   ;  CL = LBPS
                mov     ebx, 0FFFFFFFFh                         ;  EBX = MASK
                shl     ebx, cl
                not     ebx
                mov     eax, pbn?l[@cb]                         ;  drift
                and     eax, ebx                                ;  = aux?7
                mov     aux?7[@cb], eax
                mov     eax, 1                                  ;  fraction
                shl     eax, cl                                 ;  = aux?a
                sub     eax, aux?7[@cb]
                and     eax, ebx
                mov     aux?a[@cb], eax
                mov     eax, aux?0[@cb]                         ;  remainder
                add     eax, pbn?l[@cb]                         ;  = aux?e
                and     eax, ebx
                mov     aux?e[@cb], eax
                mov     eax, 1                                  ;  exclusion
                shl     eax, cl                                 ;  = aux?2
                sub     eax, aux?e[@cb]
                and     eax, ebx
                mov     aux?2[@cb], eax
                mov     eax, pbn?m[@cb]                         ;  N (LBA)
                mov     edx, pbn?l[@cb]                         ;  = aux?1
                shrd    edx, eax, cl
                mov     aux?1[@cb], edx
                mov     eax, pbn?m[@cb]                         ;  M
                mov     edx, pbn?l[@cb]                         ;  = aux?b
                add     edx, aux?0[@cb]
                adc     eax, 0
                shrd    edx, eax, cl
                mov     aux?b[@cb], edx
                mov     aux?c[@cb], 1                           ;  basic STT
                mov     eax, aux?b[@cb]                         ;  real need
                cmp     eax, aux?1[@cb]                         ;   calc. STT?
                je      vio_STT_all_ok
                mov     eax, aux?0[@cb]                         ;  STT
                sub     eax, aux?a[@cb]                         ;  = aux?c
                sub     eax, aux?e[@cb]
                shr     eax, cl
                mov     aux?c[@cb], eax
                cmp     aux?a[@cb], 0                           ;  STT adj #1
                je      vio_STT_fr_ok
                inc     aux?c[@cb]
vio_STT_fr_ok:  cmp     aux?e[@cb], 0                           ;  STT adj #2
                je      vio_STT_all_ok
                inc     aux?c[@cb]
vio_STT_all_ok: push    aux?c[@cb]                              ;  set RSC
                pop     aux?d[@cb]                              ;  = aux?d
; --- Select proper drive I-O buffer within real-mode ----------
vio_initbuffer: mov     word ptr ds:[@rmes], @seg_io
; --- Calculates drive coordinates of sector to read/write. -------
;        These data will be used to access all non-EDD, non-LBA drives.
;        LBA "logical block address" .. held in "aux?1" on entry;
;        Cylinder no. (AUX?4)         = LBA / (secspercyl);
;        Sector within cylinder (SWC) = remainder of [LBA / (secspercyl)];
;        Head within cylinder (AUX?6) = sector within cyl. / secspertrack;
;        Physical sector nr.  (AUX?5) = remainder of above (aux?6) division;
;        EDD drives ignore these calculations, and access the drive
;         throught the "vio_DAP" Disk Address Packet mantained below,
;         for which "aux?1" mantains the effective LBA, well, except
;         for the drive number, which still needs to be passed within DX,
;         and so the section flushing temporary variable DX2 must be present.
                movzx   ax, drive_id[@cb]                       ;  drive
                mov     dx2[@cb], ax                            ; > flush DX
                test    driveflags[@cb], df_enhanced            ;  stop here
                jnz     vio_set_chunk                           ; if enhanced
                mov     ecx, secspercyl[@cb]
                mov     eax, aux?1[@cb]
                xor     edx, edx
                div     ecx
                mov     aux?4[@cb], eax                         ; track nr.
                mov     eax, edx                                ; sec. of tk.
                mov     ecx, secspertrack[@cb]
                xor     edx, edx
                div     ecx
                mov     aux?6[@cb], eax                         ; head nr.
                inc     edx
                mov     aux?5[@cb], edx                         ; sector nr.
;  type-specific section 
;        this section must transform the drive coordinates in such a way
;         that the legacy ISA BIOS drive services will understand.
;        EDD drives ignore these calculations, and access the drive
;         throught the "vio_DAP" Disk Address Packet mantained below...
                mov     al, drive_id[@cb]                       ;  drive
                mov     ah, byte ptr aux?6[@cb]                 ;  head
                mov     dx2[@cb], ax                            ; > flush DX
                mov     al, byte ptr aux?5[@cb]                 ;  sector
                mov     ah, byte ptr aux?4[@cb]                 ;  track
                mov     bx, word ptr aux?4[@cb]
                shl     bh, 6
                or      al, bh                                  ; TK (8, 9)
                mov     cx2[@cb], ax                            ; > flush CX
;  end type-specific section 
; --- Check DMA overruns for attempted transfers beyond 64Kb per call. -
;        Also check if drive can read/write in cross-track mode: this is
;         always true for EDD-LBA/PIO/ATAPI drives, while is being cleared
;         for standard IDE/EIDE drives (floppies are usually unable to read
;         and/or write in cross-track mode so they should be always clipped).
vio_set_chunk:  mov     eax, aux?d[@cb]                         ;-n of sectors
                movzx   ebx, byte ptr secspercall[@cb]          ; to read for
                cmp     eax, ebx                                ; this call
                jbe     vio_bounds_ok                           ;-check limit
                mov     eax, ebx                                ; secspercall
vio_bounds_ok:  test driveflags[@cb],df_enhanced+df_crosstracker;-is capable
                jnz     vio_aligned                             ; of cross-tk?
                mov     ebx, secspertrack[@cb]                  ;-if no, cut
                sub     ebx, aux?5[@cb]                         ; sect.no. to
                inc     ebx                                     ; SPT bounds
                sub     ebx, eax                                ; (if drive
                jnc     vio_aligned                             ; can't read
                add     eax, ebx                                ; cross-track)
vio_aligned:    mov     iosectors[@cb], eax                     ;-effective
                mov     cl, lbytespersec[@cb]                   ; sectors and
                mov     iobytes[@cb], eax                       ; bytes rd/wn
                shl     iobytes[@cb], cl                        ; this time
; --- Subtract chunk size, in sectors, from remaining sectors count. --
                mov     eax, iosectors[@cb]
                sub     aux?d[@cb], eax
;  type-specific section 
;        READ data chunk:
;        This section will perform a call to the BIOS drive services
;         after having setup needed registers and parameters; it's also
;         responsible for interpreting errors and must react to them.
                test    driveflags[@cb], df_enhanced            ;  EDD drive?
                jz      vio_ISAdrive                            ;  (if not)
                mov     ax2[@cb], 4200h                         ;  set opcode
                mov     vio_wbackcode[@cb], 4300h               ;  write-back
                mov     al, byte ptr iosectors[@cb]             ; 
                mov     byte ptr vio_DAP[@cb + 2], al           ; > DAP count
                mov     eax, aux?1[@cb]                         ; 
                mov     dword ptr vio_DAP[@cb + 8], eax         ; > DAP's LBA
                jmp     vio_EDDdrive                            ; * (complete)
vio_ISAdrive:   mov     bx2[@cb], 0000h                         ; > flush BX
                mov     ah, 02h                                 ;  set opcode
                mov     al, byte ptr iosectors[@cb]             ;  set count
                mov     ax2[@cb], ax                            ; > flush AX
                mov     ah, 03h                                 ;  opcode for
                mov     vio_wbackcode[@cb], ax                  ;  write-back
vio_EDDdrive:   cmp     aux?8[@cb], 'READ'                      ;  check if
                je      vio_preload                             ;  reading
                cmp     aux?a[@cb], 0                           ;  is really
                jne     vio_preload                             ;  needed
                cmp     aux?e[@cb], 0                           ; 
                je      vio_success                             ; 
                cmp     aux?d[@cb], 0                           ; 
                je      vio_preload                             ; 
                jmp     vio_success                             ; 
vio_preload:    push    vio_attempts[@cb]                       ;  attempts
                pop     aux?f[@cb]                              ;  to make
vio_retry:      mov     ax, ax2[@cb]                            ;  prepare
                mov     bx, bx2[@cb]                            ; 
                mov     cx, cx2[@cb]                            ; 
                mov     dx, dx2[@cb]                            ; 
                lea     si, vio_DAP[@cb]                        ; 
                int     13h                                     ;  perform
                cmp     ah, 00h                                 ; 
                je      vio_success                             ;successful...
                cmp     ah, 11h                                 ; 
                je      vio_success                             ;corrected ECC
                cmp     ah, 06h                                 ; 
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                je      vio_demount                             ;disk changed!
                dec     aux?f[@cb]                              ; 
                mov     endstatus[@cb], 'G'                     ;  FAULT 
                jz      vio_exit                                ;unrecoverable
                mov     eax, vio_drive_id[@cb]                  ; 
                call    internal_RecalibrateVolume              ; 
                mov     endstatus[@cb], 'H'                     ;  FAULT 
                jc      vio_exit                                ; 
                jmp     vio_retry                               ;now retry
                ;       ------------------------------------------------------
vio_DAP         label   byte            ; Disk Address Packet for EDD drives
                db      16              ; size of Disk Address Packet
                db      0               ; reserved and must be zero, word-ext?
                db      ?               ; number of sectors to transfer (<128)
                db      0               ; reserved and must be zero, word-ext?
                dw      0               ;seen as a dword 16:16 real-mode ptr
                dw      @seg_io         ;equivalent to @seg_io:0000h by EDD
                dq      0               ; LBA dword, plus four padding zeroes
                dq      @io             ; EDD3.0 FLAT address
                ;       ------------------------------------------------------
                ;       The flat address to the transfer buffer supported by
                ;       EDD 3.0 and above specifications, unfortunately cannot
                ;       be used by VolumeIO, because of the byte-oriented
                ;       data pointers and counters (not sector-oriented):
                ;       if we let the BIOS directly access user memory then,
                ;       for instance, a read of 10 bytes would load with at
                ;       least N bytes with N = bytes per sector. The same
                ;       thing would happen even for writes, with even more
                ;       catastrophic effects. It's a pity, but there's nothing
                ;       to do: on the other side, byte-oriented disk buffering
                ;       allows very accurate mass memory management, you see.
                ;       ------------------------------------------------------
;  end type-specific section 
; --- Now copy from buffer to destination ----------------
;        Four cases are considered:
;            1) this iteration accessed a number of sectors which is
;               the beginning of the area to be transferred (fraction > 0);
;            2) this iteration accessed the last part of the area to be
;               transferred (remainder > 0);
;            3) this iteration accessed an area which is fully contained
;               into the area to be transferred (fraction = remainder = 0);
;            4) this iteration accessed the whole area to be transferred,
;               and no other iterations will be needed (STT <= iosectors).
vio_success:    mov     eax, iosectors[@cb]                     ;  STT >
                cmp     aux?c[@cb], eax                         ;   iosectors?
                ja      vio_morethisrun                         ;  (if yes)
                mov     esi, @io                                ;  only one
                add     esi, aux?7[@cb]                         ;   iteration:
                mov     edi, aux?3[@cb]                         ;   copy that
                mov     ecx, aux?0[@cb]                         ;   and STOP
                call    vio_fastcopy
                jc      vio_exit
                jmp     vio_done
vio_morethisrun:cmp     aux?a[@cb], 0                           ; fraction > 0
                je      vio_nofraction
                mov     esi, @io                                ;  copy the
                add     esi, aux?7[@cb]                         ;   fraction
                mov     edi, aux?3[@cb]                         ;   and go on
                mov     ecx, iobytes[@cb]
                sub     ecx, aux?a[@cb]
                call    vio_fastcopy
                jc      vio_exit
                add     aux?3[@cb], ecx
                mov     aux?a[@cb], 0
                jmp     vio_nextchunk
vio_nofraction: cmp     aux?d[@cb], 0                           ; remainder>0
                jne     vio_noremainder
                mov     esi, @io                                ;  copy all
                mov     edi, aux?3[@cb]                         ;   the other
                mov     ecx, iobytes[@cb]                       ;   bytes but
                sub     ecx, aux?2[@cb]                         ;  <exclusion>
                call    vio_fastcopy                            ;   and STOP
                jc      vio_exit
                jmp     vio_done
vio_noremainder:mov     esi, @io                                ;  copy all
                mov     edi, aux?3[@cb]                         ;   bytes and
                mov     ecx, iobytes[@cb]                       ;   go on...
                call    vio_fastcopy                            ; (frac=rem=0)
                jc      vio_exit
                add     aux?3[@cb], ecx
; --- There are still some sectors to read or write, so do another run. -
vio_nextchunk:  mov     eax, iosectors[@cb]                     ;  else go to
                add     aux?1[@cb], eax                         ;   ..next LBA
                jmp     vio_initbuffer                          ;  loop again
; --- If no bytes remaining, return successful. -------------
vio_done:       dmon____clear
                end_____fn
; --- On error set carry and return. ------------------
;        vio_demount invalidates drive data after the first operation
;         encountering a <06> error code (disk change line is active).
vio_demount:    mov     byte ptr [ebp], 0FFh
vio_exit:       dmon____clear
                fail____fn
; --- Internal VolumeIO routine to link RM/FM buffers. ---------
;        It basically transfers <ECX> bytes from ESI to EDI or vice-versa:
;         If the code mantained by <AUX?8> is READ, transfers from ESI to EDI;
;         If the code mantained by <AUX?8> is WRIT, transfers from EDI to ESI.
;        Also, when <AUX?8> says 'WRIT', this routine will write the contents
;         of the drive cache back to disk, assuming AX2/BX2/CX2/DX2 control
;         variables have already been initialized for a disk read. When this
;         routine fails to write back, returns with the carry flag set.
vio_fastcopy:   cmp     aux?8[@cb], 'WRIT'
                jne     vio_fc_normal
                xchg    esi, edi
vio_fc_normal:  test    ecx, ecx                                ; null count?
                jnz     vio_fc_start
                clc
                ret
vio_fc_start:   pushad
                push    ecx
                shr     ecx, 2
                jcxz    vio_fc_nodwords
                rep     movsd                                   ; copy dwords
vio_fc_nodwords:pop     ecx
                and     ecx, 3
                jcxz    vio_fc_done
                rep     movsb                                   ; copy bytes
vio_fc_done:    cmp     aux?8[@cb], 'WRIT'
                jne     vio_fc_nowrback
;  type-specific section 
vio_fc_wb:      push    vio_attempts[@cb]                       ;  attempts
                pop     aux?f[@cb]                              ;  to make
vio_fc_retry:   mov     ax, vio_wbackcode[@cb]                  ;  prepare
                mov     bx, bx2[@cb]                            ; 
                mov     cx, cx2[@cb]                            ; 
                mov     dx, dx2[@cb]                            ; 
                lea     si, vio_DAP[@cb]                        ; 
                int     13h                                     ;  write back
                cmp     ah, 00h                                 ; 
                je      vio_fc_nowrback                         ;successful...
                cmp     ah, 11h                                 ; 
                je      vio_fc_nowrback                         ;corrected ECC
                cmp     ah, 06h                                 ; 
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                je      vio_fc_demount                          ;disk changed!
                dec     aux?f[@cb]                              ; 
                mov     endstatus[@cb], 'G'                     ;  FAULT 
                jz      vio_fc_diskerr                          ;unrecoverable
                mov     eax, vio_drive_id[@cb]                  ; 
                call    internal_RecalibrateVolume              ; 
                jc      vio_fc_diskerr                          ; 
                jmp     vio_fc_retry                            ;now retry
vio_wbackcode   dw      0000h
;  end type-specific section 
vio_fc_nowrback:popad
                clc
                ret
vio_fc_demount: mov     byte ptr [ebp], 0FFh
vio_fc_diskerr: popad
                mov     endstatus[@cb], 'H'                     ;  FAULT 
                stc
                ret
; FormatVolume 
; > Input:      EAX - system drive ID string (RD01..RD04, MD01..MD08)
;               EBX:ECX - size of largest file slot (bytes)
;                         note: this 64-bit value given by EBX:ECX (EBX is the
;                         most significant doubleword) cannot drop below 1 Kb
;                         (1024 bytes), otherwise FormatVolume will fail with
;                         status H. Anyway, file slots below 1 Kb will never
;                         be allocated (too unefficient).
;               EDX - pointer to volume label string
;                     note: the volume label string is 64 bytes in size,
;                     and will be blank-padded if shorter; the string
;                     pointed by EDX may be either null-terminated, or
;                     already padded with blanks (ASCII code 32).
; > Output:     EAX - wasted file slots in VTOC
;                     (a value of zero reflects optimal format)
;               EBX:ECX - wasted volume space due to insufficient file slots
;                         (a value of zero reflects optimal format)
;               EDX:ESI - allocated volume space
;                         (usable space)
; > Notes:      The output values EAX and EBX:ECX depend on the input value
;               of EBX:ECX. Volumes are initially formated with a default,
;               general-purpose configuration. It is convenient to choose a
;               value for the largest file slot allowing these two conditions:
;                * zero waste of space;
;                * minimal waste of file slots.
;              The higher the value for EBX:ECX on entry, the smaller the
;               number of wasted file slots.
;              The smaller the value for EBX:ECX on entry, the higher the
;               amount of volume space that may be wasted because there are
;               too few file slots.
; > Faults:     A - invalid drive ID
;               B - volume not mounted
;               C - volume larger than 2 Tb and thus not supported
;               D - format request for locked volume
;               E - kernel memory image corrupted *** checks to be implemented
;               F - drive access failure
;               G - device reported invalid geometry
;               H - invalid value for largest file slot (less than 1024)
v_did           dd      ?
v_lab           dd      ?
lfs_l           dd      ?
lfs_m           dd      ?
fvl_datastart   dd      ?
fvl_vtocblock   dd      ?
fvl_fpos_l      dd      ?
fvl_fpos_m      dd      ?
fvl_wastedslots dd      ?
fvl_bouquet_c   dd      ?
fvl_bouquet_n   dd      ?
                db      'FormatVolume', 12      ; Global Object IDentifier
; --- Check drive ID validity. ---------------------
FormatVolume:   status  'f'
                status  'v'
                idup    eax
                mov     v_did[@cb], eax                         ; 
                mov     lfs_m[@cb], ebx                         ;  preserve
                mov     lfs_l[@cb], ecx                         ;  parameters
                mov     v_lab[@cb], edx                         ; 
                call    internal_TranslateDriveID               ;  translate
                mov     endstatus[@cb], 'A'                     ;  FAULT 
                jc      fvl_exit                                ;  invalid ID
                cmp     DIBr[@cb + edi + 00], byte ptr 0FFh     ;  check DIB
                mov     endstatus[@cb], 'B'                     ;  FAULT 
                je      fvl_exit                                ;  n/mounted
                test    DIBr[@cb + edi + 15], df_write_enable   ;  check WREN
                mov     endstatus[@cb], 'D'                     ;  FAULT 
                jz      fvl_exit                                ;  read-only
                call    sub_extend_DIB                          ;  extend DIB
                jz      fvl_exit                                ;  C / G 
                cmp     lfs_m[@cb], 0                           ;  
                jne     fvl_lfs_ok                              ;  check lfs
                cmp     lfs_l[@cb], 1024                        ;  validity
                jnb     fvl_lfs_ok                              ; 
                mov     endstatus[@cb], 'H'                     ;  FAULT 
                jmp     fvl_exit                                ;  lfs < 1024
fvl_lfs_ok:     mov     al, byte ptr DIBr[@cb + edi + 00]       ;  update
                mov     byte ptr ds:[@krnl + 02h], al           ;  BIOS ID
                mov     eax, dword ptr DIBr[@cb + edi + 01]     ;  update
                mov     dword ptr ds:[@krnl + 06h], eax         ;  n. tracks
                mov     eax, dword ptr DIBr[@cb + edi + 05]     ;  update
                mov     dword ptr ds:[@krnl + 0Ah], eax         ;  n. heads
                mov     eax, dword ptr DIBr[@cb + edi + 09]     ;  update
                mov     dword ptr ds:[@krnl + 0Eh], eax         ;  secs/track
                mov     al, byte ptr mvl_bytespersec[@cb]       ;  update
                mov     byte ptr ds:[@krnl + 12h], al           ;  log2 b/sec
                mov     eax, dword ptr ds:[@krnl + 14h]         ;  update
                mov     mvl_vtocbase[@cb], eax                  ;  VTOC base
                shr     edi, di_sizeshift - vi_sizeshift        ;  and VIBr
                mov     dword ptr VIBr[@cb + edi + 00], eax     ; (EDI destr)
                mov     esi, v_lab[@cb]                         ; 
                mov     edi, @krnl + 28h                        ; 
                mov     ecx, 64                                 ; 
fvl_vlabcopy:   cmp     byte ptr [esi], 0                       ; 
                je      fvl_vlabzero                            ; 
                mov     al, [esi]                               ; 
                inc     esi                                     ; 
                mov     [edi], al                               ;  update
                inc     edi                                     ;  volume
                dec     ecx                                     ;  label
                jnz     fvl_vlabcopy                            ; 
fvl_vlabzero:   test    ecx, ecx                                ; 
                jz      fvl_vlabdone                            ; 
fvl_vlabfill:   mov     byte ptr [edi], 32                      ; 
                inc     edi                                     ; 
                dec     ecx                                     ; 
                jnz     fvl_vlabfill                            ; 
fvl_vlabdone:   mov     eax, mvl_vtocbase[@cb]
                add     eax, mvl_vtocsize[@cb]
                mov     fvl_datastart[@cb], eax
                m2m     fvl_vtocblock[@cb], mvl_vtocbase[@cb]
                m2m     fvl_wastedslots[@cb], mvl_vtocslots[@cb]
                m2m     fvl_fpos_m[@cb], mvl_dev_size_m[@cb]
                m2m     fvl_fpos_l[@cb], mvl_dev_size_l[@cb]
                mov     fvl_bouquet_c[@cb], 2
                mov     fvl_bouquet_n[@cb], 2
fvl_bufferloop: mov     esi, @vtoc
                mov     edi, @vtoc + @vtsize
fvl_formatloop: mov     eax, fvl_fpos_m[@cb]
                mov     ebx, fvl_fpos_l[@cb]
                sub     ebx, lfs_l[@cb]
                sbb     eax, lfs_m[@cb]
                test    eax, eax
                jz      fvl_stillspace
                cmp     ebx, fvl_datastart[@cb]
                jnb     fvl_stillspace
                jmp     fvl_end
fvl_stillspace: mov     fvl_fpos_m[@cb], eax
                mov     fvl_fpos_l[@cb], ebx
                mov     dword ptr [esi + 0], 0
                mov     dword ptr [esi + 4], 0
                mov     dword ptr [esi + 8], ebx
                mov     dword ptr [esi + 12], eax
                add     esi, 16
                dec     fvl_wastedslots[@cb]
                jz      fvl_end
                cmp     lfs_m[@cb], 0
                jne     fvl_lfsgt1024
                cmp     lfs_l[@cb], 1024
                ja      fvl_lfsgt1024
                jmp     fvl_lfslt1024
fvl_lfsgt1024:  dec     fvl_bouquet_c[@cb]
                jnz     fvl_lfslt1024
                mov     eax, lfs_m[@cb]
                mov     ebx, lfs_l[@cb]
                shrd    ebx, eax, 1
                mov     lfs_m[@cb], eax
                mov     lfs_l[@cb], ebx
                shl     fvl_bouquet_n[@cb], 1
                m2m     fvl_bouquet_c[@cb], fvl_bouquet_n[@cb]
fvl_lfslt1024:  cmp     esi, edi
                jb      fvl_formatloop
                pushad                                          ; 
                mov     quietmode[@cb], 1                       ; 
                xor     eax, eax                                ; 
                mov     ebx, fvl_vtocblock[@cb]                 ; 
                mov     ecx, @vtsize                            ; 
                mov     edx, v_did[@cb]                         ; 
                mov     esi, 'WRIT'                             ; 
                mov     edi, @vtoc                              ; 
                call    VolumeIO                                ; 
                popad                                           ; 
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                jc      fvl_exit                                ;  drive err
                add     fvl_vtocblock[@cb], @vtsize
                jmp     fvl_bufferloop
fvl_end:        mov     quietmode[@cb], 1                       ; 
                xor     eax, eax                                ; 
                mov     ebx, fvl_vtocblock[@cb]                 ; 
                mov     ecx, esi                                ; 
                sub     ecx, @vtoc                              ; 
                mov     edx, v_did[@cb]                         ; 
                mov     esi, 'WRIT'                             ; 
                mov     edi, @vtoc                              ; 
                call    VolumeIO                                ; 
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                jc      fvl_exit                                ;  drive err
                mov     quietmode[@cb], 1                       ; 
                xor     eax, eax                                ; 
                xor     ebx, ebx                                ; 
                mov     ecx, 7E00h                              ; 
                mov     edx, v_did[@cb]                         ; 
                mov     esi, 'WRIT'                             ; 
                mov     edi, @krnl                              ; 
                call    VolumeIO                                ; 
                mov     endstatus[@cb], 'F'                     ;  FAULT 
                jc      fvl_exit                                ;  drive err
                mov     eax, v_did[@cb]                         ;  
                call    internal_TranslateDriveID               ;  now it's a
                or      DIBr[@cb + edi + 15], df_systemdisk     ;  systemdisk
                mov     eax, fvl_wastedslots[@cb]               ; 
                mov     ebx, fvl_fpos_m[@cb]                    ; 
                mov     ecx, fvl_fpos_l[@cb]                    ;  wasted
                sub     ecx, fvl_datastart[@cb]                 ;  space
                sbb     ebx, 0                                  ; 
                mov     edx, mvl_dev_size_m[@cb]                ; 
                mov     esi, mvl_dev_size_l[@cb]                ; 
                sub     esi, fvl_fpos_l[@cb]                    ;  allocated
                sbb     edx, fvl_fpos_m[@cb]                    ;  space
                sub     esi, fvl_datastart[@cb]                 ; 
                sbb     edx, 0                                  ; 
                end_____fn
fvl_exit:       fail____fn
System32        ends

; 
;  END SYSTEM CODE                                                          
; 

end
